#!/usr/bin/env python3
"""
Compare Branch to Metropolis
by Javantea
Aug 12, 2016

This is an important part of the Small Wide World paper. You can test any algorithm you wish, but Branch is the first to get its test.

python3 branch_metro1.py --seed 1 >branch_metro1a.txt &    
python3 branch_metro1.py --seed 2 >branch_metro1b.txt &    
python3 branch_metro1.py --seed 3 >branch_metro1c.txt &    
python3 branch_metro1.py --seed 4 >branch_metro1d.txt &    

python3 parsegridm1.py <(cat branch_metro1?.txt) >branch_metro1abcd.txt

import numpy as np
a = np.loadtxt('branch_metro1abcd.txt')
cols = [[c[i] for c in a] for i in range(len(a[0]))]

print("Average Branch dt: %3.3f" % (sum(cols[3])/len(cols[3])))
print("Average Metropolis dt: %3.3f" % (sum(cols[5])/len(cols[5])))

# Potential ratio
pot_ratio = np.array(cols[4])/np.array(cols[6])
print('Branch better than Metropolis %3.3f%%' % (100*len([x for x in pot_ratio if x < 1])/len(pot_ratio)))
print('Branch better than Metropolis %3.3f%%' % (100*len([x for x in pot_ratio if x < (1/2)])/len(pot_ratio)))

print('Metropolis better than Branch %3.3f%%' % (100*len([x for x in pot_ratio if x >= 1])/len(pot_ratio)))
print('Metropolis 2x better than Branch %3.3f%%' % (100*len([x for x in pot_ratio if x >= (2)])/len(pot_ratio)))
print('Metropolis 3x better than Branch %3.3f%%' % (100*len([x for x in pot_ratio if x >= (3)])/len(pot_ratio)))
print('Metropolis 4x better than Branch %3.3f%%' % (100*len([x for x in pot_ratio if x >= (4)])/len(pot_ratio)))
print('Metropolis 5x better than Branch %3.3f%%' % (100*len([x for x in pot_ratio if x >= (5)])/len(pot_ratio)))
print('Metropolis 6x better than Branch %3.3f%%' % (100*len([x for x in pot_ratio if x >= (6)])/len(pot_ratio)))
print('Metropolis 10x better than Branch %3.3f%%' % (100*len([x for x in pot_ratio if x >= (10)])/len(pot_ratio)))
print('Metropolis 100x better than Branch %3.3f%%' % (100*len([x for x in pot_ratio if x >= (100)])/len(pot_ratio)))
print('Metropolis 200x better than Branch %3.3f%%' % (100*len([x for x in pot_ratio if x >= (200)])/len(pot_ratio)))
print('Metropolis 1000x better than Branch %3.3f%%' % (100*len([x for x in pot_ratio if x >= (1000)])/len(pot_ratio)))
print('Metropolis 10000x better than Branch %3.3f%%' % (100*len([x for x in pot_ratio if x >= (10000)])/len(pot_ratio)))
print('Metropolis 100000x better than Branch %3.3f%%' % (100*len([x for x in pot_ratio if x >= (100000)])/len(pot_ratio)))
print('Metropolis 1000000x better than Branch %3.3f%%' % (100*len([x for x in pot_ratio if x >= (1000000)])/len(pot_ratio)))
print('Metropolis 10000000x better than Branch %3.3f%%' % (100*len([x for x in pot_ratio if x >= (10000000)])/len(pot_ratio)))
print('Metropolis 100000000x better than Branch %3.3f%%' % (100*len([x for x in pot_ratio if x >= (100000000)])/len(pot_ratio)))

"""
from __future__ import print_function
import os
from math import pi
import random
import time
import string
import numpy as np
from scipy.optimize import minimize
from sww_optimize1 import LJ_Potential, V_SmallWideWorld, getNonConns, createMap4, ConnList
import mapmake1
import metropolis1
tau = 2*pi
trueRand = random.SystemRandom()
Vector2 = mapmake1.Vector2
gridToPos = mapmake1.gridToPos

def testBranchMetropolis(seed=9999, start=0, end=10000):
	"""
	Check important assumptions about Branch vs Metropolis.
	"""
	lj_conf = LJ_Potential()
	b0 = 40
	rn = trueRand.randint(0, 0xffffffff)
	for i in range(0, end):
		print("seed:", seed)
		nodes = mapmake1.randint(12, 50)
		conns = mapmake1.randint(1, 10)
		branches = mapmake1.randint(0, 20)
		print("Map: %i %i %i" % (nodes, conns, branches))
		map1 = createMap4(nodes, conns, branches, seed, False)
		# Use a different seed to get different maps. Deterministic.
		seed = mapmake1.randint(0, 0xffffffff)
		# If we don't do createMap4, the seed will be different, expensive but not as expensive as sorting/optimizing by a long shot.
		if i < start: continue
		#map1.sortLinear()
		#map1.sortLoop()
		#map1.centerBound()
		map1.desc = "SWW Optimize map with {} nodes and {} conns [results]".format(nodes, conns)
		# FIXME: rare overwrites
		base_filename = 'grid1_%i_%i.json' % (rn, i)
		x0, conns = map1.getX0Conns(True)
		conns_cpp = ConnList(conns)
		#print(x0)
		#if verbose:
		#	print("atom conns:", conns)
		#end if
		nonConns = ConnList(getNonConns(conns, True))
		start_t = time.time()
		map1.sortBranch(b0)
		# Reduce the quality slightly, but reduce the overlaps significantly.
		map1.randomize(3)
		branch_dt = (time.time() - start_t)
		x_out = map1.getX0Conns()[0]
		V_branch = V_SmallWideWorld(x_out, conns_cpp, b0, lj_conf, nonConns)
		print("Branch dt: %3.3f seconds" % branch_dt)

		start2_t = time.time()
		# Remove almost any benefit branch sort gave us.
		map1.randomize(300)
		metropolis1.Metropolis(map1, time_limit=10., temperature=300, b0=40, dv=40)
		x_out = map1.getX0Conns()[0]
		V_metropolis = V_SmallWideWorld(x_out, conns_cpp, b0, lj_conf, nonConns)
		metropolis_dt = (time.time() - start2_t)
		print("Branch Potential:", V_branch)
		print("Metropolis dt: %3.3f seconds" % metropolis_dt)
		print("Metropolis Potential:", V_metropolis)
	#next i
#end def testBranchMetropolis(seed=9999, start=0, end=10000)

def testBranchWithMetropolis(seed=9999, start=0, end=10000):
	"""
	Check important assumptions about Branch+Metropolis vs Metropolis.
	"""
	lj_conf = LJ_Potential()
	b0 = 40
	rn = trueRand.randint(0, 0xffffffff)
	for i in range(0, end):
		print("seed:", seed)
		nodes = mapmake1.randint(12, 50)
		conns = mapmake1.randint(1, 10)
		branches = mapmake1.randint(0, 20)
		print("Map: %i %i %i" % (nodes, conns, branches))
		map1 = createMap4(nodes, conns, branches, seed, False)
		# Use a different seed to get different maps. Deterministic.
		seed = mapmake1.randint(0, 0xffffffff)
		# If we don't do createMap4, the seed will be different, expensive but not as expensive as sorting/optimizing by a long shot.
		if i < start: continue
		#map1.sortLinear()
		#map1.sortLoop()
		#map1.centerBound()
		map1.desc = "SWW Optimize map with {} nodes and {} conns [results]".format(nodes, conns)
		# FIXME: rare overwrites
		base_filename = 'grid1_%i_%i.json' % (rn, i)
		x0, conns = map1.getX0Conns(True)
		conns_cpp = ConnList(conns)
		#print(x0)
		#if verbose:
		#	print("atom conns:", conns)
		#end if
		nonConns = ConnList(getNonConns(conns, True))
		start_t = time.time()
		map1.sortBranch(b0)
		# Reduce the quality slightly, but reduce the overlaps significantly.
		map1.randomize(3)
		metropolis1.Metropolis(map1, time_limit=0.2, temperature=300, b0=40, dv=40)
		branch_dt = (time.time() - start_t)
		x_out = map1.getX0Conns()[0]
		V_branch = V_SmallWideWorld(x_out, conns_cpp, b0, lj_conf, nonConns)
		print("Branch dt: %3.3f seconds" % branch_dt)

		start2_t = time.time()
		# Remove almost any benefit branch sort gave us.
		map1.randomize(300)
		metropolis1.Metropolis(map1, time_limit=10., temperature=300, b0=40, dv=40)
		x_out = map1.getX0Conns()[0]
		V_metropolis = V_SmallWideWorld(x_out, conns_cpp, b0, lj_conf, nonConns)
		metropolis_dt = (time.time() - start2_t)
		print("Branch Potential:", V_branch)
		print("Metropolis dt: %3.3f seconds" % metropolis_dt)
		print("Metropolis Potential:", V_metropolis)
	#next i
#end def testBranchWithMetropolis(seed=9999, start=0, end=10000)


def testMetropolisMetropolis(seed=9999, start=0, end=10000):
	"""
	Check important assumptions about Branch+Metropolis vs Metropolis.
	"""
	lj_conf = LJ_Potential()
	b0 = 40
	rn = trueRand.randint(0, 0xffffffff)
	for i in range(0, end):
		print("seed:", seed)
		nodes = mapmake1.randint(12, 50)
		conns = mapmake1.randint(1, 10)
		branches = mapmake1.randint(0, 20)
		print("Map: %i %i %i" % (nodes, conns, branches))
		map1 = createMap4(nodes, conns, branches, seed, False)
		# Use a different seed to get different maps. Deterministic.
		seed = mapmake1.randint(0, 0xffffffff)
		# If we don't do createMap4, the seed will be different, expensive but not as expensive as sorting/optimizing by a long shot.
		if i < start: continue
		#map1.sortLinear()
		#map1.sortLoop()
		#map1.centerBound()
		map1.desc = "SWW Optimize map with {} nodes and {} conns [results]".format(nodes, conns)
		# FIXME: rare overwrites
		base_filename = 'grid1_%i_%i.json' % (rn, i)
		x0, conns = map1.getX0Conns(True)
		conns_cpp = ConnList(conns)
		#print(x0)
		#if verbose:
		#	print("atom conns:", conns)
		#end if
		nonConns = ConnList(getNonConns(conns, True))
		start_t = time.time()
		map1.sortBranch(b0)
		# Remove almost any benefit branch sort gave us.
		map1.randomize(300)
		metropolis1.Metropolis(map1, time_limit=0.2, temperature=300, b0=40, dv=40)
		branch_dt = (time.time() - start_t)
		x_out = map1.getX0Conns()[0]
		V_branch = V_SmallWideWorld(x_out, conns_cpp, b0, lj_conf, nonConns)
		print("Metropolis(0.2s) dt: %3.3f seconds" % branch_dt)

		start2_t = time.time()
		# Remove almost any benefit branch sort gave us.
		map1.randomize(300)
		metropolis1.Metropolis(map1, time_limit=10., temperature=300, b0=40, dv=40)
		x_out = map1.getX0Conns()[0]
		V_metropolis = V_SmallWideWorld(x_out, conns_cpp, b0, lj_conf, nonConns)
		metropolis_dt = (time.time() - start2_t)
		print("Metropolis(0.2s) Potential:", V_branch)
		print("Metropolis dt: %3.3f seconds" % metropolis_dt)
		print("Metropolis Potential:", V_metropolis)
	#next i
#end def testMetropolisMetropolis(seed=9999, start=0, end=10000)

def testBranchRandom(seed=9999, start=0, end=10000):
	"""
	Check important assumptions about Branch vs Random.
	"""
	lj_conf = LJ_Potential()
	b0 = 40
	rn = trueRand.randint(0, 0xffffffff)
	for i in range(0, end):
		print("seed:", seed)
		nodes = mapmake1.randint(12, 50)
		conns = mapmake1.randint(1, 10)
		branches = mapmake1.randint(0, 20)
		print("Map: %i %i %i" % (nodes, conns, branches))
		map1 = createMap4(nodes, conns, branches, seed, False)
		# Use a different seed to get different maps. Deterministic.
		seed = mapmake1.randint(0, 0xffffffff)
		# If we don't do createMap4, the seed will be different, expensive but not as expensive as sorting/optimizing by a long shot.
		if i < start: continue
		#map1.sortLinear()
		#map1.sortLoop()
		#map1.centerBound()
		map1.desc = "SWW Optimize map with {} nodes and {} conns [results]".format(nodes, conns)
		# FIXME: rare overwrites
		base_filename = 'grid1_%i_%i.json' % (rn, i)
		x0, conns = map1.getX0Conns(True)
		conns_cpp = ConnList(conns)
		#print(x0)
		#if verbose:
		#	print("atom conns:", conns)
		#end if
		nonConns = ConnList(getNonConns(conns, True))
		start_t = time.time()
		map1.sortBranch(b0)
		# Reduce the quality slightly, but reduce the overlaps significantly.
		map1.randomize(3)
		branch_dt = (time.time() - start_t)
		x_out = map1.getX0Conns()[0]
		V_branch = V_SmallWideWorld(x_out, conns_cpp, b0, lj_conf, nonConns)
		print("Branch dt: %3.3f seconds" % branch_dt)

		start2_t = time.time()
		# Remove almost any benefit branch sort gave us.
		map1.randomize(300)
		V_random = V_SmallWideWorld(x_out, conns_cpp, b0, lj_conf, nonConns)
		random_dt = (time.time() - start2_t)
		print("Branch Potential:", V_branch)
		print("Random dt: %3.3f seconds" % random_dt)
		print("Random Potential:", V_random)
	#next i
#end def testBranchRandom(seed=9999, start=0, end=10000)

def main():
	"""Usage: python3 grid1.py [-v] [-h] [-d] [-t] [filename.json] [...]
	Runs AltSci Grid Global Minimization Algorithm on a JSON file.
	-v | --verbose Print the name combinations that produce the dr
	-h | --help    Print this help.
	-d | --data    Write a dat file as output instead of a array. Dat files 
	               are very easy to parse with numpy or gnuplot.
	-t | --test    Run a quick test to ensure that Grid sort always beats Branch sort.
	-n | --nodes   How many nodes in the backbone [default 20]
	-c | --conns   How many connections between the backbone [default 4]
	-b | --branches How many branches to add to the backbone [default 0]
	-r | --random  Use a random seed instead of a deterministic seed id
	-x | --seed    Deterministic random number generator seed [default 9999]
	"""
	import sys
	import getopt
	
	# Deterministic
	mapmake1.seed(12345)
	
	# Do not switch to scientific notation even though you enjoy it.
	# We're trying to do science here!
	np.set_printoptions(suppress=True)

	filenames = ['']
	verbose = False
	datOutput = False
	#if len(sys.argv) > 1:
	#	filename = sys.argv[1]
	#end if
	start = 0
	end = None
	actualRandom = False
	seed = 9999
	nodes = 20
	conns = 4
	branches = 0
	branchWithMetro = False
	control = False
	experimentRandom = False
	funcName = ''
	#steps = 20
	gtol = 1e-2
	try:
		flags, extra_args = getopt.getopt(sys.argv[1:], 'vhmcs:e:rug:x:f:y', 
		['verbose', 'help', 'metro', 'control', 'funcname=', 'start=', 
		'end=', 'random', 'shuffle', 'gtol=', 'seed=', 'experiment'])
	except getopt.GetoptError as goe:
		print(goe)
		return 1
	#end try
	for flag in flags:
		if flag[0] == '-v' or flag[0] == '--verbose':
			verbose = True
		elif flag[0] == '-m' or flag[0] == '--metro':
			branchWithMetro = True
		elif flag[0] == '-c' or flag[0] == '--control':
			control = True
		elif flag[0] == '-h' or flag[0] == '--help':
			usage()
			return 1
		elif flag[0] == '-s' or flag[0] == '--start':
			start = int(flag[1])
		elif flag[0] == '-e' or flag[0] == '--end':
			end = int(flag[1])
		elif flag[0] == '-r' or flag[0] == '--random':
			actualRandom = True
		elif flag[0] == '-u' or flag[0] == '--shuffle':
			shuffle = True
		elif flag[0] == '-g' or flag[0] == '--gtol':
			gtol = float(flag[1])
		elif flag[0] == '-x' or flag[0] == '--seed':
			seed = int(flag[1])
		#end if
	#next flag
	if end == None:
		end = start + 10000
	#end if
	
	if extra_args != []:
		filenames = extra_args
	#end if
	
	lj_conf = LJ_Potential()
	
	# Necessary for the functions we call, optional for mapmake1.
	mapmake1.loadPolar()
	b0 = 40
	
	if branchWithMetro:
		return testBranchWithMetropolis(seed, start, end)
	elif control:
		return testMetropolisMetropolis(seed, start, end)
	elif experimentRandom:
		return testBranchRandom(seed, start, end)
	else:
		return testBranchMetropolis(seed, start, end)
	#end if
#end def main()

if __name__ == '__main__':
	main()
#end if
