/*
Small Wide World Optimize C++
by Javantea
Feb 18, 2016
Based on SWW Minimize 1
by Javantea
Sept 30, 2015
Based on N Atom molecule using Basin-Hopping Method (aka L-BFGS-B)
by Javantea
July 23, 2014
Based on metropolis_min1.py by Javantea, July 17-22, 2014
Based on s_4atom.py by Javantea, June 24, 2014

*/
#include <cmath>
#include "sww.h"
#include <algorithm>
#include <vector>
#include <iostream>
#include <unistd.h>

class LJ_Potential {
	//public:
	float A;
	float B;
	public:
	float r0;
	float scale_factor;
	LJ_Potential(float r0_init=40, float scale_factor_init=0.055);
	~LJ_Potential();
	float V_LJ(float r);
	float V_LJ_fast(Vector2 &r);
};

/*"""
Proper Lennard-Jones 6 12 Potential.
"""*/
LJ_Potential::LJ_Potential(float r0_init, float scale_factor_init)
{
	r0 = r0_init;
	scale_factor = scale_factor_init;
	A = 2 * powf(r0, 6) * scale_factor;
	B = powf(r0, 12) * scale_factor;
} //#end def LJ_Potential(r0=40, scale_factor=0.055)

LJ_Potential::~LJ_Potential()
{
}

float LJ_Potential::V_LJ(float r)
{
	/*"""
	Attempt at a Lennard-Jones 6 12 potential
	The potential is very high when the two nodes are close.
	The potential is low when two nodes are far away.
	The potential is lowest when two nodes are at the r0 distance, not too 
	close, not too far away.
	Non-Connected nodes should have a LJ potential.
	r0 is the value for which this LJ potential is optimized.
	"""*/
	//# This is a divide by zero protection.
	if (r < 1e-3) {
		return V_LJ(1e-3);
	} //#end if
	float r6 = 1. / powf(r, 6);
	return ((-A*r6) + (B*r6*r6));
} //#end def V_LJ(r)
//#end class LJ_Potential

float LJ_Potential::V_LJ_fast(Vector2 &r)
{
	/*
	Equivalent to V_LJ(r.length())
	*/
	float lsq = r.lensq();
	// This is a divide by zero protection.
	if (lsq < 1e-6) {
		return V_LJ(1e-3);
	}
	float r6 = 1. / powf(lsq, 3);
	return ((-A*r6) + (B*r6*r6));
} //#end def V_LJ_fast(r)
//#end class LJ_Potential

float V_Harmonic(float r, float b0)
{
	/*"""
	Attempt at a Harmonic potential
	Connected nodes should have a harmonic potential.
	b0 is the value for which this Harmonic potential is optimized.
	A is the scale of potential.
	"""*/
	float A = 1.;
	return A * powf(r - b0, 2);
} //#end def V_Harmonic(r, b0)

float V_Harmonic_Angle(float theta, float theta0)
{
	/*"""
	Attempt at a Harmonic potential for the angle between two connected items
	Connected nodes should have a harmonic potential.
	theta0 is the value for which this Harmonic potential is optimized.
	A is the scale of potential.
	"""*/
	float A = 2.;
	return A * powf(theta - theta0, 2);
} //#end def V_Harmonic_Angle(theta, theta0)

float V_SmallWideWorld(float *x, ConnList *conns, float b0, LJ_Potential *lj_conf, ConnList *nonConns)
{
	/*"""
	Just like a molecule but in 2D.
	"""*/
	size_t n = (*conns).size();
	//std::cout << "V_SWW(" << n << ")" << std::endl;
	float Potential = 0.0;
	size_t i, j, k, m;
	//dps_all = [[]] * len(conns)
	std::vector<std::vector<Vector2>> dps_all;
	for(i = 0; i < n; i++)
	{
		dps_all.push_back(std::vector<Vector2>());
	}
	//# Generate conns list for each node
	for(i = 0; i < n; i++)
	{
		Vector2 pt_i(x[2*i], x[2*i+1]);
		for (m = 0; m < ((*conns)[i].indices).size(); m++) {
			j = (*conns)[i].indices[m];
			Vector2 pt_j(x[2*j], x[2*j+1]);
			Vector2 dp = pt_i - pt_j;
			dps_all[i].push_back(dp);
			dps_all[j].push_back(dp * -1.);
		} //#next j
	} //#next i
	//# Harmonic potential with length b0.
	for(i = 0; i < n; i++)
	{
		Vector2 pt_i(x[2*i], x[2*i+1]);
		//std::cout << "pt_i: " << pt_i.x << ", " << pt_i.y << std::endl;
		//printf("pt_i[%lu]: %3.9f, %3.9f\n", i, pt_i.x, pt_i.y);
		std::vector<Vector2> dps = dps_all[i];
		for (m = 0; m < ((*conns)[i].indices).size(); m++) {
			j = (*conns)[i].indices[m];
			Vector2 pt_j(x[2*j], x[2*j+1]);
			Vector2 dp = pt_i - pt_j;
			float dp_len = dp.length();
			Potential += V_Harmonic(dp_len, b0);
			//std::cout << "P1: " << Potential << std::endl;
		} //#next j
		if (dps.size() >= 2) {
			//# Angular Harmonic calculation
			float theta0 = 2 * PI / dps.size();
			//# theta is computed for axb, bxc, cxa
			std::vector<float> thetas;
			for (k = 0; k < dps.size(); k++) {
				Vector2 ax = dps[k];
				thetas.push_back(fmod((float)atan2(ax.y, ax.x), (2*PI)));
			} //#next k
			sort(thetas.begin(), thetas.end());
			for (k = 0; k < thetas.size(); k++) {
				int b = (k + 1) % thetas.size();
				float theta = fmod(((2*PI) + thetas[b] - thetas[k]), (2*PI));
				Potential += V_Harmonic_Angle(theta, theta0);
				//std::cout << "P2: " << Potential << " " << theta <<  " " << theta0 << ";ffs" << std::endl;
			} //#next k
		} //#end if
	} //#next i
	//# Lennard-Jones for non-connected nodes.
	for(i = 0; i < n; i++) {
		Vector2 pt_i(x[2*i], x[2*i+1]);
		for (m = 0; m < ((*nonConns)[i].indices).size(); m++) {
			j = (*nonConns)[i].indices[m];
			Vector2 pt_j(x[2*j], x[2*j+1]);
			Vector2 dp = pt_i - pt_j;
			Potential += lj_conf->V_LJ_fast(dp);
			//std::cout << "P3: " << Potential << std::endl;
		} //#next j
	} //#next i
	return Potential;
} //#end def V_SmallWideWorld(x, conns, b0, lj_conf)

ConnList *getNonConns(ConnList *conns, bool i_lt_j=true)
{
	/*"""
	Inverse of conns. If a is not connected to b, b should be in conns[a].
	"""*/
	ConnList *nonConns = new ConnList();
	int n = (*conns).size();
	int i, j;
	std::vector<int>::iterator p;
	for (i = 0; i < n; i++)
	{
		Conns nc;
		for (j = 0; j < n; j++)
		{
			//# Don't check against ourself.
			if (i == j) continue;
			//# If we are connected, not not connected
			p = std::find((*conns)[i].indices.begin(), (*conns)[i].indices.end(), j);
			if (p != (*conns)[i].indices.end()) continue;
			//std::cout << i << " nonconn " << j << std::endl;
			nc.indices.push_back(j);
		} //#next j
		// i_lt_j filter here
		if(i_lt_j)
		{
			//std::remove_if(nc.begin(), nc.end(), );
			for(p = nc.indices.begin(); p != nc.indices.end();) {
				if((*p) <= i)
				{ nc.indices.erase(p); }
				else
				{ ++p; }
			}
		}
		(*nonConns).push_back(nc);
	} //#next i
	return nonConns;
} //#end def getNonConns(conns, bool i_lt_j)

#ifdef TEST

int main(int argc, char **argv)
{
	LJ_Potential lj_conf;
	std::cout << "lj_conf defaults: " << lj_conf.r0 << ", " << lj_conf.scale_factor << std::endl;
	// TODO: map1.getX0Conns()
	ConnList conns;
	Conns conns0;
	conns0.indices.push_back(1); // 0 conn 1,2
	conns0.indices.push_back(2);
	conns.push_back(conns0);
	Conns conns1;
	conns1.indices.push_back(0); // 1 conn 0
	conns.push_back(conns1);
	Conns conns2;
	conns2.indices.push_back(0); // 2 conn 0
	conns.push_back(conns2);
	// Non-conns should be 1 nonconn 2.
	ConnList *nonConns = getNonConns(&conns, true);
	std::cout << "NonConns: " << (*nonConns).size() << std::endl;
	size_t n = conns.size();
	size_t i, j;
	for (i = 0; i < n; i++)
	{
		for (j = 0; j < (*nonConns)[i].indices.size(); j++)
		{
			std::cout << "Disconnect: " << i << " " << (*nonConns)[i].indices[j] << std::endl;
		}
	}
	delete nonConns;
	nonConns = getNonConns(&conns, false);
	std::cout << "NonConns: " << (*nonConns).size() << std::endl;
	for (i = 0; i < n; i++)
	{
		for (j = 0; j < (*nonConns)[i].indices.size(); j++)
		{
			std::cout << "Disconnect: " << i << " " << (*nonConns)[i].indices[j] << std::endl;
		}
	}
	delete nonConns;
	float b0 = 40;
	unsigned long seed = time(0) ^ (getpid() << 8);
	std::cout << "Seed: " << seed << std::endl;
	srand(seed);
	Map map1;
	map1.addPoint(Point("a", 0, 0));
	map1.addPoint(Point("b", 0, 0));
	map1.addPoint(Point("c", 0, 0));
	map1.points[0].addConn(&map1.points[1]);
	map1.points[1].addConn(&map1.points[2]);
	map1.randomize();
	float *x0 = map1.getX0();
	ConnList *connsp = map1.getConns();
	nonConns = getNonConns(connsp, true);
	float Potential = V_SmallWideWorld(x0, connsp, b0, &lj_conf, nonConns);
	std::cout << "Potential: " << Potential << std::endl;
	delete nonConns;
}

#endif // TEST
