
/*
 * MetropolisAccept(V_test, V_min, beta)
	Based on code from SciPy basinhopping.
	V_test is the new potential.
	V_min is the previous minimum potential.
	Beta is 1 / temperature
	w is between 0 and 1, as V_test becomes much larger than V_min, w decreases.
	When V_test <= V_min, w == 1
	However, pretty large values of V_test are accepted at a very high 
	probability when beta is small (temperature is high).
	Therefore, temperature should be low for good results.
 */
function MetropolisAccept(V_test, V_min, beta)
{
	var w = Math.min(1.0, Math.exp(-(V_test - V_min) * beta));
	var rand = Math.random();
	return w >= rand;
} //#end def MetropolisAccept(energy_new, energy_old, beta)

/**
 * MetropolisObject
 * Used by Basinhopping, but not Metropolis
 **/
function MetropolisObject(T) {
    this.beta = 1 / T;
}
new MetropolisObject(1);

/**
 * NOTE: Unlike python, no keyword arguments. You have to figure out the correct order!
 */
MetropolisObject.prototype.call = function(energy_new, x_new, energy_old, x_old) {
    return MetropolisAccept(energy_new, energy_old, this.beta);
}

/*
 * Metropolis(map1, time_limit=10., temperature=300, b0=40, dv=40)
	Metropolis Monte-Carlo
 */
function Metropolis(map1, time_limit, temperature, b0, dv)
{
	if (time_limit == undefined) { time_limit = 10.; }
	if (temperature == undefined) { temperature = 300; }
	if (b0 == undefined) { b0 = 40; }
	if (dv == undefined) { dv = 40; }
	var lj_conf = new LJ_Potential();
	//# The Metropolis algorithm in a nutshell.
	var x0conns = map1.getX0Conns(true);
	var x0 = x0conns[0];
	var conns = x0conns[1];
	var nonConns = getNonConns(conns, true);
	var V0 = V_SmallWideWorld(x0, conns, b0, lj_conf, nonConns);
	//debug("got past V_SmallWideWorld 1");
	var x_min = x0;
	var V_min = V0;
	var x_in = x0;
	var beta = 1 / temperature;
	var start = Date.now() * 0.001;
	var dt = 0;
	var i;
	while (dt < time_limit) {
		var x_t = Array.from(x_in);
		for (i = 0; i < x_t.length; i++) {
			x_t[i] += (Math.random() - 0.5) * dv;
		} //#next i
		V_t = V_SmallWideWorld(x_t, conns, b0, lj_conf, nonConns);
		dt = (Date.now() * 0.001) - start;
		//debug("V_SmallWideWorld at " + dt + " " + V_t);
		var resetPos = true;
		if (V_t < V_min) {
			V_min = V_t;
			x_min = x_t;
			debug("Min found at t=" + dt + " " + V_t);
			resetPos = false;
		} else if (MetropolisAccept(V_t, V_min, beta)) {
			//# Keep this x_t if random chance tells us to.
			resetPos = false;
		} //#end if
		if (resetPos) {
			//# Reset to x_min
			x_in = x_min;
		} //#end if
	} //#loop
	
	//# End of Metropolis algorithm
	//m = np.matrixlib.matrix(x_min)
	//m = m.reshape(len(x_min)//2, 2)
	var x_min_shaped = [];
	for (i = 0; i < x_min.length/2; i++) {
		x_min_shaped.push(x_min.slice(2*i, 2*i+2));
	}
	map1.setPos(x_min_shaped);
} //#end def Metropolis(map1, time_limit=10., temperature=300, b0=40, dv=40)
