#!/usr/bin/env python3
"""
Metropolis
by Javantea
Jan 24, 2016
Based on Loop Sort
by Javantea
Dec 31, 2015

Similar to sww_optimize but with a different (worse) algorithm. Designed to prove the validity of claim that branch sort and loop sort are superior to naive algorithms in global optimization.
"""
from __future__ import print_function
import os
from sww_optimize1 import createMap4, LJ_Potential, getNonConns, V_SmallWideWorld, ConnList
import mapmake1
import time
import numpy as np

def usage():
	"Print the usage info."
	print(main.__doc__)
	return 1
#end def usage()

def MetropolisAccept(V_test, V_min, beta):
	"""
	Based on code from SciPy basinhopping.
	V_test is the new potential.
	V_min is the previous minimum potential.
	Beta is 1 / temperature
	w is between 0 and 1, as V_test becomes much larger than V_min, w decreases.
	When V_test <= V_min, w == 1
	However, pretty large values of V_test are accepted at a very high 
	probability when beta is small (temperature is high).
	Therefore, temperature should be low for good results.
	"""
	w = min(1.0, np.exp(-(V_test - V_min) * beta))
	rand = np.random.rand()
	return w >= rand
#end def MetropolisAccept(energy_new, energy_old, beta)

def Metropolis(map1, time_limit=10., temperature=300, b0=40, dv=40):
	"""
	Metropolis Monte-Carlo
	"""
	lj_conf = LJ_Potential()
	# The Metropolis algorithm in a nutshell.
	x0, conns = map1.getX0Conns(True)
	conns_cpp = ConnList(conns)
	nonConns = ConnList(getNonConns(conns, True))
	V0 = V_SmallWideWorld(x0, conns_cpp, b0, lj_conf, nonConns)
	x_min = x0
	V_min = V0
	x_in = x0
	beta = 1 / temperature
	start = time.time()
	dt = 0
	j = 0
	while dt < time_limit:
		x_t = np.array(x_in)
		for i in range(len(x_t)):
			x_t[i] += (np.random.rand() - 0.5) * dv
		#next i
		V_t = V_SmallWideWorld(x_t, conns_cpp, b0, lj_conf, nonConns)
		dt = time.time() - start
		resetPos = True
		if V_t < V_min:
			V_min = V_t
			x_min = x_t
			print("Min found at t=%3.6f %3.6f" % (dt, V_t))
			resetPos = False
		elif MetropolisAccept(V_t, V_min, beta):
			# Keep this x_t if random chance tells us to.
			resetPos = False
		#end if
		if resetPos:
			# Reset to x_min
			x_in = x_min
		#end if
		j += 1
	#loop
	print(j, 'rounds')
	
	# End of Metropolis algorithm
	m = np.matrixlib.matrix(x_min)
	m = m.reshape(len(x_min)//2, 2)
	map1.setPos([np.array(xa)[0] for xa in m])
#end def Metropolis(map1, time_limit=10., temperature=300, b0=40, dv=40)

def main():
	"""
	Small Wide World Loop Sort Interface
This program organizes a graph by loop detection algorithm.
Usage:
loopsort1.py [-v] [-h] [-n nodes] [-c conns] [-b branches] [filename]
-v       | verbose
-h       | This usage.
-n nodes | How many nodes in the backbone [default 20]
-c conns | How many connections between the backbone [default 1]
-b branches | How many branches to add to the backbone [default 0]
-l       | Output the linear sort instead of loop sort.
filename | JSON file optional.
	"""
	import sys
	import getopt
	nodes = 20
	conns = 1
	branches = 0
	seed = 9999
	verbose = False
	linear = False
	fix = False
	time_limit = 10.
	temperature = 100
	b0 = 40
	try:
		flags, extra_args = getopt.getopt(sys.argv[1:], 'vhdn:c:b:x:t:w:lf', ['verbose', 'help', 'nodes=', 'conns=', 'branches=', 'seed=', 'time=', 'temperature=', 'linear', 'fix'])
	except getopt.GetoptError as goe:
		print(goe)
		return 1
	#end try
	for flag in flags:
		if flag[0] == '-v' or flag[0] == '--verbose':
			verbose = True
		elif flag[0] == '-h' or flag[0] == '--help':
			usage()
			return 1
		elif flag[0] == '-n' or flag[0] == '--nodes':
			nodes = int(flag[1])
		elif flag[0] == '-c' or flag[0] == '--conns':
			conns = int(flag[1])
		elif flag[0] == '-b' or flag[0] == '--branches':
			branches = int(flag[1])
		elif flag[0] == '-x' or flag[0] == '--seed':
			seed = int(flag[1])
		elif flag[0] == '-l' or flag[0] == '--linear':
			linear = True
		elif flag[0] == '-f' or flag[0] == '--fix':
			fix = True
		elif flag[0] == '-t' or flag[0] == '--time':
			time_limit = float(flag[1])
		elif flag[0] == '-w' or flag[0] == '--temperature':
			temperature = float(flag[1])
		#end if
	#next flag
	filename = None
	if len(extra_args) > 0:
		filename = extra_args[0]
	#end if
	mapmake1.seed(seed)
	#end if
	if filename:
		map1 = mapmake1.Map()
		map1.loadJson(open(filename, 'r').read())
		nodes = len(map1.points)
		# If we need conns at some point:
		# All connections minus the backbone connections (nodes - 1)
		#conns = sum([len(n.conns) for n in map1.points]) - (nodes - 1)
		map1.desc = "SWW Loop Sort map with {} nodes from {} [results]".format(nodes, filename)
	else:
		# Don't sort please, faster.
		map1 = createMap4(nodes, conns, branches, seed, False)
		map1.desc = ("SWW Loop Sort map with {} nodes, {} conns, and " +
			   "{} branches, seed: {} [results]").format(nodes, conns, branches, seed)
	#end if
	
	Metropolis(map1, time_limit, temperature, b0)
	map1.centerBound()

	if verbose:
		for p in map1.points:
			for q in p.conns:
				dr = p - q
				print("%s -- %s distance: %3.3f" % (p.name, q.name, dr.length()))
			#next q
		#next p
	#end if
	file_num = 0
	while os.path.exists('metropolis1_%i.json' % file_num):
		file_num += 1
	#loop
	open('metropolis1_%i.json' % file_num, 'w').write(map1.saveJson())
	print("Wrote metropolis1_%i.json" % file_num)
#end def main()

if __name__ == '__main__':
	main()
#end if
