#!/usr/bin/env python3
"""
Small Wide World Potential Energy Function Printer
by Javantea
Jul 3, 2016

Based on sww_energy1.py aka. Small Wide World Potential Energy Calculator
by Javantea
Jan 2, 2016
Based on sww_optimize1.py aka. SWW Minimize 1
by Javantea
Sept 30, 2015
Based on N Atom molecule using Basin-Hopping Method (aka L-BFGS-B)
by Javantea
July 23, 2014
Based on metropolis_min1.py by Javantea, July 17-22, 2014
Based on s_4atom.py by Javantea, June 24, 2014

"""
from __future__ import print_function
import math
import sww_optimize1
import mapmake1
Vector2 = mapmake1.Vector2
ConnList = sww_optimize1.ConnList

def V_SmallWideWorld_string(map1, x, conns, b0, lj_conf, nonConns):
	"""
	Just like a molecule but in 2D.
	TODO: Harmonic_Angle and Lennard-Jones
	"""
	Potential_function = "V_SmallWideWorld() = "
	#Potential = 0.0
	dps_all = []
	for i in range(len(conns)):
		dps_all.append([])
	#next i
	# Generate conns list for each node
	for i in range(len(conns)):
		pt_i = Vector2(*x[2*i:2*i+2])
		for j in conns[i]:
			pt_j = Vector2(*x[2*j:2*j+2])
			dp = pt_i - pt_j
			dps_all[i].append(dp)
			dps_all[i].append(dp * -1.)
		#next j
	#next i
	# Harmonic potential with length b0.
	for i in range(len(conns)):
		pt_i = Vector2(*x[2*i:2*i+2])
		dps = dps_all[i]
		for j in conns[i]:
			pt_j = Vector2(*x[2*j:2*j+2])
			dp = pt_i - pt_j
			dp_len = dp.length()
			#Potential += V_Harmonic(dp_len, b0)
			Potential_function += " + A*(sqrt((" + map1.points[i].name + ".x - " + map1.points[j].name + ".x)**2 - (" + map1.points[i].name + ".y - " + map1.points[j].name + ".y)**2) - b0)**2"
		#next j
		if len(dps) >= 2:
			# Angular Harmonic calculation
			theta0 = 2 * math.pi / len(dps)
			# theta is computed for axb, bxc, cxa
			thetas = []
			for k in range(len(dps)):
				ax = dps[k]
				thetas.append(math.atan2(ax.y, ax.x) % (2*math.pi))
			#next k
			thetas.sort()
			for k in range(len(thetas)):
				b = (k + 1) % len(thetas)
				theta = (thetas[b] - thetas[k]) % (2*math.pi)
				#Potential += V_Harmonic_Angle(theta, theta0)
			#next k
		#end if
	#next i
	# Lennard-Jones for non-connected nodes.
	for i in range(len(nonConns)):
		pt_i = Vector2(*x[2*i:2*i+2])
		for j in nonConns[i]:
			pt_j = Vector2(*x[2*j:2*j+2])
			dp = pt_i - pt_j
			dp_len = dp.length()
			#Potential += lj_conf.V_LJ(dp_len)
		#next j
	#next i
	return Potential_function
#end def V_SmallWideWorld_string(map1, x, conns, b0, lj_conf, nonConns)

def V_Harmonic_gradient(dp, r, b0):
	"""
	Harmonic potential gradient with respect to pt_i
	Connected nodes should have a harmonic potential.
	b0 is the value for which this Harmonic potential is optimized.
	A is the scale of potential.
	Returns tuple of two values, dx and dy.
	"""
	if r < 1e-6:
		return V_Harmonic_gradient(dp, 1e-6, b0)
	#end if
	A = 1
	d = 2 * A * (r - b0) * (1/r)
	dVdx = d * (dp.x)
	dVdy = d * (dp.y)
	return (dVdx, dVdy)
#end def V_Harmonic_gradient(dp, r, b0)

class LJ_Potential_g(sww_optimize1.LJ_Potential):
	"""
	Proper Lennard-Jones 6 12 Potential with gradient.
	"""
	def V_LJ_gradient(self, dp, r_sq):
		"""
		Lennard-Jones 6 12 potential gradient
		The potential is very high when the two nodes are close.
		The potential is low when two nodes are far away.
		The potential is lowest when two nodes are at the r0 distance, not too 
		close, not too far away.
		Non-Connected nodes should have a LJ potential.
		r0 is the value for which this LJ potential is optimized.
		dV_LJ(x)/dx = 6*(A / (r_sq**4) - 2*B / (r_sq**7)) * (x - d.x)
		"""
		# This is a divide by zero protection.
		if r_sq < 1e-3:
			return self.V_LJ_gradient(dp, 1e-3)
		#end if
		r8 = 1/(r_sq ** 4)
		r14 = 1/(r_sq ** 7)
		d = 6*((self.A*r8) - (self.B*r14))
		return (d*dp.x, d*dp.y)
	#end def V_LJ(r)
#end class LJ_Potential_g

def V_Harmonic_Angle_gradient(dp, theta, theta0):
	"""
	
	"""
	if theta < 1e-6:
		return V_Harmonic_Angle_gradient(dp, 1e-6, b0)
	#end if
	A = 2
	d = 2 * (theta - theta0) * (1/theta)
	dp_lensq = dp.lensq()
	dtheta = (-dp.y / dp_lensq, dp.x / dp_lensq)
	dVdx = d * dtheta[0]
	dVdy = d * dtheta[1]
	return (dVdx, dVdy)
#end def V_Harmonic_Angle_gradient(dp, theta, theta0)

def V_SmallWideWorld_gradient(x, conns, b0, lj_conf, nonConns):
	"""
	Just like a molecule but in 2D.
	"""
	Potential_gradient = [0.0] * (len(x))
	dps_all = []
	for i in range(len(conns)):
		dps_all.append([])
	#next i
	# Generate conns list for each node
	for i in range(len(conns)):
		pt_i = Vector2(*x[2*i:2*i+2])
		for j in conns[i]:
			pt_j = Vector2(*x[2*j:2*j+2])
			dp = pt_i - pt_j
			dps_all[i].append(dp)
			dps_all[j].append(dp * -1.)
		#next j
	#next i
	# Harmonic potential with length b0.
	for i in range(len(conns)):
		pt_i = Vector2(*x[2*i:2*i+2])
		dps = dps_all[i]
		for j in conns[i]:
			pt_j = Vector2(*x[2*j:2*j+2])
			dp = pt_i - pt_j
			dp_len = dp.length()
			grad = V_Harmonic_gradient(dp, dp_len, b0)
			Potential_gradient[2*i] += grad[0]
			Potential_gradient[2*i + 1] += grad[1]
		#next j
		if len(dps) >= 2:
			# Angular Harmonic calculation
			theta0 = 2 * math.pi / len(dps)
			# theta is computed for axb, bxc, cxa
			thetas = []
			for k in range(len(dps)):
				ax = dps[k]
				thetas.append((math.atan2(ax.y, ax.x) % (2*math.pi), k))
			#next k
			thetas.sort()
			for k in range(len(thetas)):
				b = (k + 1) % len(thetas)
				theta = (thetas[b][0] - thetas[k][0]) % (2*math.pi)
				grad = V_Harmonic_Angle_gradient(dps[thetas[b][1]], theta, theta0)
				Potential_gradient[2*i] += grad[0]
				Potential_gradient[2*i+1] += grad[1]
			#next k
		#end if
	#next i
	# Lennard-Jones for non-connected nodes.
	for i in range(len(nonConns)):
		pt_i = Vector2(*x[2*i:2*i+2])
		for j in nonConns[i]:
			pt_j = Vector2(*x[2*j:2*j+2])
			dp = pt_i - pt_j
			dp_lensq = dp.lensq()
			grad = lj_conf.V_LJ_gradient(dp, dp_lensq)
			Potential_gradient[2*i] += grad[0]
			Potential_gradient[2*i+1] += grad[1]
		#next j
	#next i
	return Potential_gradient
#end def V_SmallWideWorld_gradient(x, conns, b0, lj_conf)


def V_SmallWideWorld_gradient_verbose(x, conns, b0, lj_conf, nonConns):
	"""
	Just like a molecule but in 2D.
	"""
	Potential_gradient = [0.0] * (len(x))
	dps_all = []
	for i in range(len(conns)):
		dps_all.append([])
	#next i
	# Generate conns list for each node
	for i in range(len(conns)):
		pt_i = Vector2(*x[2*i:2*i+2])
		for j in conns[i]:
			pt_j = Vector2(*x[2*j:2*j+2])
			dp = pt_i - pt_j
			dps_all[i].append(dp)
			dps_all[j].append(dp * -1.)
		#next j
	#next i
	# Harmonic potential with length b0.
	for i in range(len(conns)):
		pt_i = Vector2(*x[2*i:2*i+2])
		dps = dps_all[i]
		for j in conns[i]:
			pt_j = Vector2(*x[2*j:2*j+2])
			dp = pt_i - pt_j
			dp_len = dp.length()
			grad = V_Harmonic_gradient(dp, dp_len, b0)
			Potential_gradient[2*i] += grad[0]
			Potential_gradient[2*i + 1] += grad[1]
			print('i:', i, 'dp:', dp, 'r:', dp_len, 'grad:', grad)
		#next j
		if len(dps) >= 2:
			# Angular Harmonic calculation
			theta0 = 2 * math.pi / len(dps)
			# theta is computed for axb, bxc, cxa
			thetas = []
			for k in range(len(dps)):
				ax = dps[k]
				thetas.append((math.atan2(ax.y, ax.x) % (2*math.pi), k))
			#next k
			thetas.sort()
			print('thetas:', thetas)
			for k in range(len(thetas)):
				b = (k + 1) % len(thetas)
				theta = (thetas[b][0] - thetas[k][0]) % (2*math.pi)
				grad = V_Harmonic_Angle_gradient(dps[thetas[b][1]], theta, theta0)
				Potential_gradient[2*i] += grad[0]
				Potential_gradient[2*i+1] += grad[1]
				print('i:', i, 'dp:', dps[thetas[b][1]], 'theta:', theta, 'theta0:', theta0, 'grad:', grad)
			#next k
		#end if
	#next i
	# Lennard-Jones for non-connected nodes.
	for i in range(len(nonConns)):
		pt_i = Vector2(*x[2*i:2*i+2])
		for j in nonConns[i]:
			pt_j = Vector2(*x[2*j:2*j+2])
			dp = pt_i - pt_j
			dp_lensq = dp.lensq()
			grad = lj_conf.V_LJ_gradient(dp, dp_lensq)
			Potential_gradient[2*i] += grad[0]
			Potential_gradient[2*i+1] += grad[1]
			# TODO: Reverse?
			print('i:', i, 'dp:', dp, 'grad:', grad)
		#next j
	#next i
	return Potential_gradient
#end def V_SmallWideWorld_gradient_verbose(x, conns, b0, lj_conf)


def usage():
	print(main.__doc__)
	return 1
#end def usage()

def main():
	"""Small Wide World Potential Energy Function Printer
This program computes the energy of a graph in terms of variables.
Usage:
sww_energy1.py [-h] [-v] filename
-h       | This usage.
-v       | Verbose output.
filename | JSON file.

Example output:
Conns: 4
V_SmallWideWorld() =  + A*(sqrt((a.x - b.x)**2 - (a.y - b.y)**2) - b0)**2 + A*(sqrt((a.x - c.x)**2 - (a.y - c.y)**2) - b0)**2 + A*(sqrt((b.x - c.x)**2 - (b.y - c.y)**2) - b0)**2 + A*(sqrt((d.x - c.x)**2 - (d.y - c.y)**2) - b0)**2

"""
	lj_conf = LJ_Potential_g()
	import sys
	
	filename = None
	V_SmallWideWorld = sww_optimize1.V_SmallWideWorld
	if '-h' in sys.argv:
		return usage()
	#end if
	V_SmallWideWorld_gradient_f = V_SmallWideWorld_gradient
	if '-v' in sys.argv:
		V_SmallWideWorld_gradient_f = V_SmallWideWorld_gradient_verbose
		sys.argv.remove('-v')
	#end if
	if len(sys.argv) > 1:
		filename = sys.argv[1]
	#end if
	
	if filename is None:
		return usage()
	#end if
	map1 = mapmake1.Map()
	map1.loadJson(open(filename, 'r').read())
	nodes = len(map1.points)
	# If we need conns at some point:
	# All connections minus the backbone connections (nodes - 1)
	#conns = sum([len(n.conns) for n in map1.points]) - (nodes - 1)
	map1.desc = "SWW Optimize map with {} nodes from {} [results]".format(nodes, filename)
	x0, conns = map1.getX0Conns(True)
	print("Conns:", len(conns))
	conns_cpp = ConnList(conns)
	non_conns = sww_optimize1.getNonConns(conns, True)
	non_conns_cpp = ConnList(non_conns)
	#print("Configuration:")
	potential_function = V_SmallWideWorld_string(map1, x0, conns, 40, lj_conf, non_conns)
	print(potential_function)
	gradient = V_SmallWideWorld_gradient_f(x0, conns, 40, lj_conf, non_conns)
	gradient_compare = []
	epsilon = 1e-4
	for i in range(len(x0)):
		x_delta = x0[:]
		x_delta[i] += epsilon
		gradient_compare.append((V_SmallWideWorld(x_delta, conns_cpp, 40, lj_conf, non_conns_cpp) - V_SmallWideWorld(x0, conns_cpp, 40, lj_conf, non_conns_cpp))/epsilon)
	#next i
	print(gradient)
	print(gradient_compare)
#end def main()

if __name__ == '__main__':
	main()
#end if
