#!/usr/bin/env python3
"""
Regular Graph Generator
by Javantea
Jan 6, 2016

A regular graph is a graph where every node has the same number of neighbors. For example,

 o   single node network.

o--o two-node network.

  o  triangle
 / \
o---o
o--o square
|  |
o--o
are all regular graphs.

for i in $(seq 2 50); do python3 regular1.py "$i"; done

https://en.wikipedia.org/wiki/Regular_graph
"""
from __future__ import print_function
import os
import mapmake1

def usage():
	"Print the usage info."
	print(main.__doc__)
	return 1
#end def usage()

def createRegularGraph(nodes=6, conns=1, seed_id=3243, runSorts=True):
	"""
	The 0 conns is easy.
	1 conns means that we add connections on unconnected nodes until each node has only one.
	This only works for even number of nodes.
	2 conns means we have a loop.
	3 conns means we have a loop with each node connected to a third node not adjacent.
	TODO: 1, 3, and so on.
	Start with 2
	FIXME: There's a bug in regular1.py that treats the --conns argument as boolean instead of as an integer.
	"""
	if conns >= nodes: raise ValueError("Too many connections, conns must be less than nodes")
	mapmake1.seed(seed_id)
	a = mapmake1.Map()
	b = []
	for i in range(nodes):
		b.append(mapmake1.Point(nameIncremental2(i)))
	#next i
	if conns > 0:
		# Connect them all end to end
		for i in range(nodes-1):
			b[i].addConn(b[i+1])
		#next i
		# Connect first to last to get a loop
		if nodes > 2:
			b[0].addConn(b[-1])
		#end if
	#end if
	a.addPoints(b)
	if runSorts:
		# Sort it in the way of the branch.
		a.sortBranch()
		# Only 2 pixels of randomization
		a.randomize(2)
		a.center()
	#end if
	return a
#end def createRegularGraph([nodes], [conns], [seed_id], runSorts=True)

def nameIncremental2(i):
	"""
	a b c ... aa ab ac ad ae ... ba bb bc bd ... ca cb cc ...
	aaa aab aac aad ... baa bab bac bad ...
	aaaa aaab aaac aaad ... baaa baab baac baad ....
	"""
	if i == 0: return 'a'
	alphabet = 'abcdefghijklmnopqrstuvwxyz'
	output = ''
	#verbose = False
	#if i >= 26*26: verbose = True
	p = i
	digit = 0
	while i > 0:
		#if verbose: print('i in', i)
		i, p = divmod(i, 26)
		if digit > 0: p -= 1
		#if verbose: print('i, p out', i, p)
		output = alphabet[p] + output
		if p < 0:
			if i == 1: break
			i -= 1
		#end if
		digit += 1
	#loop
	#output = alphabet[p] + output
	#if verbose: print('output', output)
	return output
#end def nameIncremental2(i)

def main():
	"""
	Small Wide World Regular Graph Generator
This program organizes a graph by loop detection algorithm.
Usage:
regular1.py [-v] [-h] [-n nodes] [-c conns]
-v       | verbose
-h       | This usage.
-n nodes | How many nodes in the backbone [default 20]
-c conns | How many connections between the backbone [default 1]
	"""
	import sys
	import getopt
	nodes = 20
	conns = 1
	seed = 9999
	b0 = 40
	verbose = False
	try:
		flags, extra_args = getopt.getopt(sys.argv[1:], 'vhdn:c:b:x:', ['verbose', 'help', 'nodes=', 'conns=', 'seed='])
	except getopt.GetoptError as goe:
		print(goe)
		return 1
	#end try
	for flag in flags:
		if flag[0] == '-v' or flag[0] == '--verbose':
			verbose = True
		elif flag[0] == '-h' or flag[0] == '--help':
			usage()
			return 1
		elif flag[0] == '-n' or flag[0] == '--nodes':
			nodes = int(flag[1])
		elif flag[0] == '-c' or flag[0] == '--conns':
			conns = int(flag[1])
		elif flag[0] == '-x' or flag[0] == '--seed':
			seed = int(flag[1])
		#end if
	#next flag
	mapmake1.seed(seed)
	#end if
	# Don't sort please, faster.
	map1 = createRegularGraph(nodes, conns, seed, False)
	map1.desc = ("SWW Regular map with {} nodes, {} conns, and " +
			"seed: {} [results]").format(nodes, conns, seed)
	#end if
	# FIXME: Should definitely sort loop before connecting fully.
	map1.sortLinear(b0)
	# A horrible fix for low nodes
	map1.points[-1].x -= 2 * b0
	map1.sortLoop(b0)
	map1.centerBound()
	file_num = 0
	while os.path.exists('regular1_%i.json' % file_num):
		file_num += 1
	#loop
	open('regular1_%i.json' % file_num, 'w').write(map1.saveJson())
	print("Wrote regular1_%i.json" % file_num)
#end def main()

if __name__ == '__main__':
	main()
#end if
