/*
         o     o         o
         |     |        /
         |     |       /
  o======o=====o======o=======o
        / \            \
       /   \            \
      o     o            o

Small Wide World == Open Source Graph Optimization with an editor
Copyright (C) 2015 Javantea
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
var tau = 2*Math.PI;

/*
Map(points, width, height, line_width, line_color)
Base class for all maps (friend, object, route).
Write the hard math and physics here once and use it by calling once.
*/
function Map(points, width, height, line_width, line_color) {
	if(points === undefined) { points = []; }
	if(width === undefined) { width = 640; }
	if(height === undefined) { height = 480; }
	if(line_width === undefined) { line_width = 2; }
	if(line_color === undefined) { line_color = '#000000'; }
	this.points = points;
	this.width  = width;
	this.height = height;
	this.line_width = line_width;
	this.line_color = line_color;
	this.selectedPoint = null;
	this.title = 'Small Wide World Map';
	this.desc = 'Test of drawing a network map with svg';
	this.prettying = false;
}

// Make it a class.
new Map();

Map.prototype.addPoint = function(point) {
	this.points.push(point);
}

Map.prototype.addPoints = function(points) {
	var args = [this.points.length, 0]
	args = args.concat(points);
	this.points.splice.apply(this.points, args);
}

/*if (debug === undefined)
{
	// Silent but deadly.
	debug = function() {}
}*/

/*
 * This randint doesn't care if you get the params backwards.
 * Therefore it functions same in all sane calls, functions better in insane 
 * calls.
 * Similar to python random.randint(x, y)
 */
function randint(x, y) {
	var dx = Math.abs(y - x) + 1;
	var minXY = Math.min(x, y);
	return Number((minXY + (Math.random() * (dx-1))).toFixed());
}

Map.prototype.randomize = function(amount) {
	// Deal with optional parameters.
	if(amount === undefined) { amount = 180; }
	var i;
	for (i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		point.x += randint(-amount, amount)
		point.y += randint(-amount, amount)
	} //#next point
} //#end Map.prototype.randomize

Map.prototype.drag = function(x, y) {
	if (this.selectedPoint === null) {
		return false;
	}
	var r = new Vector2(x, y);
	var d_len = (r.sub(this.selectedPoint)).lensq();
	if (d_len > 100) {
		this.selectedPoint.x = x;
		this.selectedPoint.y = y;
		return true;
	}
	return false;
} //#end Map.prototype.drag

Map.prototype.click = function(x, y) {
	var cp = new Vector2(x, y);
	var min_dr = -1.0;
	var min_dr_p = null;
	var i;
	for (i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		dr = cp.sub(point);
		dr_len = dr.lensq();
		if ((min_dr < 0) || (dr_len < min_dr)) {
			min_dr = dr_len;
			min_dr_p = point;
		}
	} //#next point
	if (min_dr_p != null) {
		//print("Setting color!")
		min_dr_p.setColor('#0000cc');
		if (this.selectedPoint != null) {
			this.selectedPoint.setColor('#ccc');
		}
		this.selectedPoint = min_dr_p;
	}
} //#end Map.prototype.click

/*
 * Map.getNamed
 * Only available in JavaScript for now.
 * Aides Conn('a', 'b')
 */
Map.prototype.getNamed = function(s) {
	var i;
	for(i = 0; i < this.points.length; i++)
	{
		if(this.points[i].name == s)
			return this.points[i];
	}
	return null;
}

/*
 * Map.sort1()
 * Bring all points together.
 */
Map.prototype.sort1 = function() {
	var i;
	var j;
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		for (j = 0; j < point.conns.length; j++) {
			var conn = point.conns[j].b;
			if (conn == point) { conn = point.conns[j].a; }
			var dr = conn.sub(point);
			//#(dx, dy) = (conn.x - point.x, conn.y - point.y)
			point.x += dr.x * 0.4;
			point.y += dr.y * 0.4;
		} //#next conn
	} //#next point
	this.prettying = false;
} //#end Map.prototype.sort1

/*
 * Map.avgLength()
 * Expensive function tot find average length.
 * O(n**2)
 */
Map.prototype.avgLength = function() {
	var avgLen = 0;
	var conns = 0;
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		for (j = 0; j < point.conns.length; j++) {
			var conn = point.conns[j].b;
			if (conn == point) { conn = point.conns[j].a; }
			//#(dx, dy) = (conn.x - point.x, conn.y - point.y)
			var dr = conn.sub(point);
			//#dr_len = sqrt(dx*dx + dy*dy)
			var dr_len = dr.length();
			avgLen += dr_len;
			conns++;
		}
	}
	//debug("Average Length: " + (avgLen / conns));
	if (conns == 0)
		return 80;
	return avgLen / conns;
}

/*
 * Map.sort2(dist1, f)
 * Use springs to decide the position of each point.
 * Moves the whole thing to the lower right due to a bug.
 */
Map.prototype.sort2 = function(dist1, f) {
	// Deal with optional parameters.
	if(dist1 === undefined) { dist1 = this.avgLength(); }
	if(f === undefined) { f = 0.5; }
	var i;
	var j;
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		for (j = 0; j < point.conns.length; j++) {
			var conn = point.conns[j].b;
			if (conn == point) { conn = point.conns[j].a; }
			//#(dx, dy) = (conn.x - point.x, conn.y - point.y)
			var dr = conn.sub(point);
			//#dr_len = sqrt(dx*dx + dy*dy)
			var dr_len = dr.length();
			//#print '	dr:', dr
			if (dr_len == 0.0) {
				point.x += 5.0;
				point.y += 5.0;
				dr_len = Math.sqrt(5*5 + 5*5);
				//#print 'zero dr is now:', dr
			}
			//#print 'non-zero dr:', dr
			var m = 1;
			var a_len = (f/m) * (dr_len - dist1);
			//#dr_hat = (dx/dr, dy/dr)
			var dr_hat = dr.div(dr_len);
			//#a = (a_len*dr_hat[0], a_len*dr_hat[1]/m)
			var a = dr_hat.mul(a_len / m);
			dt = 0.6;
			//#dp = (0.5*dt*dt*a[0], 0.5*dt*dt*a[1])
			var dp = a.mul(0.5 * dt * dt);
			point.x += dp.x;
			point.y += dp.y;
			//debug(i + " " + a_len + " " + dr_hat.toString());
		} //#next conn
	} //#next point
	this.prettying = false;
}//#end Map.prototype.sort2

/*
 * Map.sort3
 * Pull all nodes to the center of their connections.
 */
Map.prototype.sort3 = function() {
	var i;
	var j;
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		if (point.conns.length <= 1) continue;
		//#avgx = 0
		//#avgy = 0
		var avg = new Vector2(0, 0);
		for (j = 0; j < point.conns.length; j++) {
			var conn = point.conns[j].b;
			if (conn == point) { conn = point.conns[j].a; }
			//#avgx += conn.x
			//#avgy += conn.y
			avg = avg.add(conn);
		} //#next conn
		//#point.x = avgx / point.conns.length
		//#point.y = avgy / point.conns.length
		var p = avg.div(point.conns.length);
		point.x = p.x;
		point.y = p.y;
	} //#next point
	this.prettying = false;
}//#end Map.prototype.sort3

/**
 * checkPointInConns(point, conns)
 * FIXME: This should only be used when getNonConns is slower (never).
 * 
 **/
function checkPointInConns(point, conns) {
	var i;
	for(i = 0; i < conns.length; i++) {
		if (conns[i].b == point)
			return true;
	}
	return false;
}

/*
 * Map.sort4()
 * Use a style of reverse attack on non-conns that are close.
 * FIXME: broken
 * Also should use getNonConns
 */
Map.prototype.sort4 = function() {
	var i;
	var j;
	var dr, dr_len;
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		for (j = 0; j < this.points.length; j++) {
			var nonconn = this.points[j];
			if (checkPointInConns(nonconn, point.conns)) {
				continue;
			}
			dr = point.sub(nonconn);
			dr_len = dr.lensq();
			//Console.log("dr_len = " + dr_len);
			//#print 'dr_len =', dr_len
			if (dr_len < 2500) {
				if (dr_len == 0) {
					dr = new Vector2(1, 1);
					dr_len = Math.sqrt(2);
				}
				dr = dr.mul(4.0 / dr_len);
				point.x += dr.x;
				point.y += dr.y;
			}
		} //#next nonconn
	} //#next point
	this.prettying = false;
}//#end Map.prototype.sort4

/*
 * Map.sort5()
 * This is obviously the best algorithm ever. Hooray for the power 
 * of awesomeness.
 * FIXME: broken
 */
Map.prototype.sort5 = function() {
	var i;
	var j;
	var dr, dr_len;
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		var enemies_pos = new Vector2(0, 0);
		var enemies_len = 0;
		var friends_pos = new Vector2(0, 0);
		for (j = 0; j < this.points.length; j++) {
			var nonconn = this.points[j];
			if (checkPointInConns(nonconn, point.conns)) {
				friends_pos = friends_pos.add(nonconn);
				continue;
			}
			dr = point.sub(nonconn);
			dr_len = dr.lensq();
			//#print 'dr_len =', dr_len
			if (dr_len < 2500) {
				enemies_pos = enemies_pos.add(nonconn);
				enemies_len++;
			}
		} //#next nonconn
		enemies_pos = enemies_pos.div(enemies_len);
		friends_pos = friends_pos.div(point.conns.length);
		dr = friends_pos.mul(2).sub(enemies_pos);
		//# Our position looks like this: enemies ---- friends ---- me
		point.x = dr.x;
		point.y = dr.y;
		//debug(point.toString());
		//debug(' &nbsp; &nbsp;' + '(' + (enemies_len) + ')', enemies_pos);
		//debug(' &nbsp; &nbsp;' + '(' + (point.conns.length) + ')', friends_pos);
	} //#next point
	this.prettying = false;
}//#end Map.prototype.sort5

/*
 * Map.sort6()
 * A new algorithm based on sort5 but with added validity information.
 * FIXME: broken
 */
Map.prototype.sort6 = function() {
	var i;
	var j;
	var dr;
	var validity = new Array(this.points.length);
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		var enemies_pos = new Vector2(0, 0);
		var enemies_len = 0;
		var friends_pos = new Vector2(0, 0);
		var friends_len = 0;
		for(j = 0; j < this.points.length; j++) {
			var nonconn = this.points[j];
			if (!validity[j]) {
				continue;
			}
			if (checkPointInConns(nonconn, point.conns)) {
				friends_pos = friends_pos.add(nonconn);
				friends_len++;
				continue;
			}
			/*
			dr = point.sub(nonconn)
			dr_len = dr.length()
			print 'dr_len =', dr_len
			if (dr_len < 150) {
			*/
			enemies_pos = enemies_pos.add(nonconn);
			enemies_len++;
			//}
		} //#next nonconn
		if (enemies_len > 1) enemies_pos = enemies_pos.div(enemies_len);
		if (friends_len > 1) friends_pos = friends_pos.div(friends_len);
		if ((enemies_len > 0) || (friends_len > 0)) {
			dr = friends_pos.mul(1.7).sub(enemies_pos);
			if ((enemies_len == 0) || (dr.sub(friends_pos).lensq() < 1) || (dr.sub(enemies_pos).lensq() < 1)) {
				//# Fix anything by using a random angle.
				var theta = randint(0, 360) * (Math.PI / 180.0);
				dr = new Vector2(dr.x + 30 * Math.cos(theta), dr.y + 30 * Math.sin(theta));
			}
		} else {
			dr = new Vector2(0, 0);
		}
		//# Our position looks like this: enemies ---- friends ---- me
		point.x = dr.x + (randint(0, 100) / 4.0);
		point.y = dr.y + (randint(0, 100) / 4.0);
		//#print point
		//#print '\t' + '(' + str(enemies_len) + ')', enemies_pos
		//#print '\t' + '(' + str(friends_len) + ')', friends_pos
		validity[i] = true;
	} //#next point
	this.prettying = false;
} //#end Map.prototype.sort6

/*
 * Map.center()
 * Centers the graph on the screen.
 */
Map.prototype.center = function() {
	if (this.points.length == 0) {
		return;
	}
	var mean = new Vector2(0, 0);
	var i;
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		mean = mean.add(point);
	} //#next point
	mean = mean.div(1.0 * this.points.length);
	var goto = new Vector2(this.width/2.0, this.height/2.0);
	var diff = goto.sub(mean);
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		point.x += diff.x;
		point.y += diff.y;
	} //#next point
}; //#end Map.prototype.center

/*
 * Map.centerBound()
 * Centers the graph on the screen based on bounds, better in most ways to center().
 */
Map.prototype.centerBound = function() {
	if (this.points.length == 0) {
		return;
	}//#end if
	var bounds = this.bounds();
	var mean = (new Vector2(bounds[0], bounds[1]).add(new Vector2(bounds[2], bounds[3]))).div(2.0);
	var goto = new Vector2(this.width/2.0, this.height/2.0);
	var diff = goto.sub(mean);
	var i;
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		point.x += diff.x;
		point.y += diff.y;
	} //#next point
}; //#end Map.prototype.center

/*
 * Map.bounds()
 * Find the maximum extents.
 * returns [x_min, y_min, x_max, y_max]
 */
Map.prototype.bounds = function() {
	if (this.points.length == 0) {
		return [0, 0, 0, 0];
	}//#end if
	//# x_min, y_min, x_max, y_max
	var minp = new Vector2(this.points[0].x, this.points[0].y);
	var maxp = minp.add(new Vector2(0, 0));
	var i;
	//for point in self.points:
	for (i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		if (point.x == NaN) point.x = 0;
		if (point.y == NaN) point.y = 0;
		if (point.x < minp.x) minp.x = point.x;
		if (point.y < minp.y) minp.y = point.y;
		if (point.x > maxp.x) maxp.x = point.x;
		if (point.y > maxp.y) maxp.y = point.y;
	}//#next point
	return [minp.x, minp.y, maxp.x, maxp.y];
}//#end def bounds()

function uniq(a) {
	return a.sort().filter(function(item, pos, ary) {
		return !pos || item != ary[pos - 1];
	});
}

/*
 * Map.split()
 * Return a list of independent graphs.
 * 
 */
Map.prototype.split = function() {
	var output = new Array(this.points.length).fill(null);
	var revMap = {};
	var i, j, k;
	for (i = 0; i < this.points.length; i++) {
		revMap[this.points[i]] = i;
		// Can't do fill because JavaScript. Sets don't have concat except 
		// new Set([...set1, ...set2]) which makes me worry about the sanity of 
		// the language.
		output[i] = [];
	} //#next i
	//# First step: merge all connections together.
	for (i = 0; i < this.points.length; i++) {
		var ourSlot = null;
		var conns = [];
		//for conn in self.points[i].conns:
		for(k = 0; k < this.points[i].conns.length; k++) {
			var conn = this.points[i].conns[k];
			if (conn.a == this.points[i]) {
				conn = conn.b;
			} else {
				conn = conn.a;
			}
			var j = revMap[conn];
			if (output[j].length > 0) {
				if (ourSlot != null) {
					//# Incorporate their useless map.
					output[ourSlot] = output[ourSlot].concat(output[j]);
					output[j] = [];
				} else {
					ourSlot = j;
				} //#end if
				//#print(self.points[i].name, 'going to', self.points[j].name)
			} //#end if
			conns.push(conn);
		} //#next conn
		//debug("i: " + i + ", ourSlot: " + ourSlot + ", conns: " + conns.length + ", " + this.points[i].conns.length);
		if (ourSlot == null) {
			output[i].push(this.points[i]);
			output[i] = output[i].concat(conns);
		} else {
			output[ourSlot].push(this.points[i]);
			output[ourSlot] = output[ourSlot].concat(conns);
		} //#end if
	} //#next i
	// Dedupe output.
	for (i = 0; i < output.length; i++) {
		output[i] = uniq(output[i]);
	}
	//# this leaves us with at most a handful of groups that need to be merged.
	//# Look for things to merge.
	var m;
	for (i = 0; i < this.points.length; i++) {
		if (output[i].length <= 0) { continue; }
		var toMerge = [];
		//for p in output[i]:
		for(j = 0; j < output[i].length; j++) {
			var p = output[i][j];
			//#j = revMap[p]
			for (k = 0; k < this.points.length; k++) {
				//# Don't try to merge with yourself.
				if (i == k) continue;
				if (toMerge.indexOf(k) >= 0) continue;
				if (output[k].indexOf(p) >= 0) {
					//# Incorporate their useless map.
					toMerge.push(k);
				} //#end if
			} //#next k
		} //#next p
		//for k in toMerge:
		for (m = 0; m < toMerge.length; m++) {
			k = toMerge[m];
			output[i] = output[i].concat(output[k]);
			output[k] = [];
		} //#next k
	} //#next i
	// A simple function to mimick output.indexOf([]) which doesn't work in JavaScript.
	output.emptyPos = function() {
		var n;
		for(n = 0; n < this.length; n++) {
			if (this[n].length == 0) return n;
		}
		return -1;
	}
	while (output.emptyPos() >= 0) {
		output.splice(output.emptyPos(), 1);
	} //#loop
	return output;
}; //#end def split()

/*
 * Map.sortBranch(r0=40)
 * A brilliant sorting method that works well with branches but not loops.
 * r0 is the length of the bonds set by the algorithm.
 * roundTry is necessary only for loops.
 */
Map.prototype.sortBranch = function(r0, knownPos, debugSort) {
	if (r0 === undefined) { r0 = 40; }
	if (debugSort === undefined) { debugSort = false; }
	var nodes = this.points.length;
	if (knownPos === undefined || knownPos == null) { knownPos = new Array(nodes).fill(false); }
	//debug("sortBranch(" + nodes + ")");
	if (nodes < 1) {
		//# Congrats, you're guaranteed to be sorted.
		return;
	}
	var checkEverything = true;
	var i;
	//# This is our algorithm... Written on the whiteboard last night. Understood many years ago but unable to write the code. Until now!
	if (knownPos.indexOf(true) < 0) {
		knownPos[0] = true;
		checkEverything = false;
		this.points[0].x = this.width / 2;
		this.points[0].y = this.height / 2;
	}
	//# v1 in spherical coordinates.
	//#v1 = (r1, theta1) == (r1*cos(theta1), r1*sin(theta1))
	var v1 = [r0, 0];
	//# -v1 in spherical coordinates
	//var neg_v1 = [r0, v1[1] + Math.PI];
	// This solution is suboptimal, but I really like sortBranch now.
	var knownDone;
	var fillReady = false;
	
	for (roundTry = 0; roundTry < nodes; roundTry++) {
		if (debugSort) debug('fillReady? ' + fillReady);
		knownDone = true;
		for(i = 0; i < nodes; i++) {
			//#print(this.points[i].name, this.points[i].conns)
			givens = [];
			unknowns = [];
			//# Set the position of its connections.
			// for conn in this.points[i].conns:
			for (k = 0; k < this.points[i].conns.length; k++) {
				var conn = this.points[i].conns[k].b;
				if(conn == this.points[i]) {
					conn = this.points[i].conns[k].a;
				}
				var j = this.points.indexOf(conn);
				if(knownPos[j]) {
					givens.push(j);
					continue;
				} //#end if
				unknowns.push(j);
			} //#next conn
			if (knownPos[i]) {
				//# First solve the easy cases, then work your way up.
				if(unknowns.length == 0) {
					//# Everything is known, we're good.
					if (debugSort) debug('Yay ' + this.points[i].x + " " + this.points[i].y + " " + this.points[i].name)
					continue;
				} //#end if
				if(givens.length == 0) {
					//# All unknowns, so use the simplex method.
					var n_conns = unknowns.length;
					var theta0 = tau / n_conns; //# tau = 2*pi
					var x = 0;
					// for j in unknowns:
					for(k = 0; k < unknowns.length; k++) {
						var j = unknowns[k];
						//# Spherical coordinates
						var theta_j = v1[1] + (theta0 * (x+1));
						//#v_j = (r0, theta_j)
						this.points[j].x = this.points[i].x + (r0*Math.cos(theta_j)) 
						this.points[j].y = this.points[i].y + (r0*Math.sin(theta_j))
						knownPos[j] = true;
						knownDone = false;
						if (debugSort) debug('Simple ' + x + " " + j + " " + this.points[j].x + " " + this.points[j].y + " " + theta_j + " " + this.points[j].name)
						x++;
					} //#next j
					continue;
				} //#end if
				if(givens.length == 1) {
					//# Very easy, create a v1 and do the same as above
					var v1i_cartesian = this.points[i].sub(this.points[givens[0]]);
					var v1i_theta = Math.atan2(v1i_cartesian.y, v1i_cartesian.x);
					//#v1i_spherical = (v1i_cartesian.length(), v1i_theta)
					//#neg_v1i_spherical = (v1i_cartesian.length(), v1i_theta+pi)
					var n_conns = unknowns.length + 1;
					var theta0 = tau / n_conns; //# tau = 2*pi
					var x = 0;
					// for j in unknowns:
					for(k = 0; k < unknowns.length; k++) {
						var j = unknowns[k];
						//# Spherical coordinates
						if(n_conns == 2) {
							theta_j = v1i_theta;
						} else {
							theta_j = v1i_theta+Math.PI + (theta0 * (x+1));
						} //#end if
						//#v_j = (r0, theta_j)
						this.points[j].x = this.points[i].x + (r0*Math.cos(theta_j))
						this.points[j].y = this.points[i].y + (r0*Math.sin(theta_j))
						knownPos[j] = true;
						knownDone = false;
						if (debugSort) debug('Simple1 ' + x + " " + j + " " + this.points[j].x + " " + this.points[j].y + " " + theta_j + " " + 'v1it' + " " + v1i_theta + " " + this.points[j].name)
						x++;
					} //#next j
				} //#end if
				if (debugSort) debug("Not sure what to do for: given: " + givens.length + ", unknown: " + unknowns.length + "")
				//# Take the list of givens and create a perfect placement for our node
			} else {
				//# Don't use v1 until all known nodes are figured out.
				//# Fixes sortBranch_fail1.json
				if (!fillReady) {
					continue;
				}
				//debug("Not known pos " + i + " " + knownPos[i])
				if(givens.length == 0) {
					//# We don't know any of our neighbors yet, nope.
					continue;
				} //#end if
				if(givens.length == 1) {
					//# Use v1 to choose our position
					knownPos[i] = true;
					var j = givens[0];
					if (debugSort) debug(i + ' using v_' + j + " " + this.points[i].name)
					var theta_j = v1[1];
					this.points[i].x = this.points[j].x + (r0*Math.cos(theta_j));
					this.points[i].y = this.points[j].y + (r0*Math.sin(theta_j));
					if (checkEverything) {
						//# Expensive check for overlapping points, stops early when no overlap occurs.
						//# Worst case scenario, 3*known points
						var theta_j = [v1[1] + Math.PI, v1[1] + (Math.PI*0.5), v1[1] + (Math.PI*1.5)];
						for (theta_j_i = 0; theta_j_i < theta_j.length; theta_j_i++) {
							var theta_j_new = theta_j[theta_j_i];
							var isOverlapped = false;
							//k, p in enumerate(this.points):
							for (k = 0; k < this.points.length; k++) {
								if (k == i) continue;
								if (knownPos[k] == false) continue;
								var p = this.points[k];
								var dr = this.points[i].sub(p);
								if (dr.lensq() < SMALL) {
									//#print("Found overlap, it's time to fix.")
									//# We're on top of something known, move.
									this.points[i].x = this.points[j].x + (r0*Math.cos(theta_j_new));
									this.points[i].y = this.points[j].y + (r0*Math.sin(theta_j_new));
									isOverlapped = true;
									break;
								} //#end if
							} //#next k, p
							//# No overlap found, this is good enough.
							if (!isOverlapped) break;
						} //#next theta_j_new
					}
					continue;
				} //#end if
				if (debugSort) debug('avg');
				//# Average of its peers
				var totalPos = [0, 0];
				// for j in givens:
				for(k = 0; k < givens.length; k++) {
					var j = givens[k];
					totalPos[0] += this.points[j].x;
					totalPos[1] += this.points[j].y;
				} //#next j
				var avg = [totalPos[0] / givens.length, totalPos[1] / givens.length];
				this.points[i].x = avg[0];
				this.points[i].y = avg[1];
				knownPos[i] = true;
			} //#end if
			// if not false in knownPos:
			if (knownPos.indexOf(false) == -1) break;
		} //#next i
		// if not false in knownPos:
		if (knownPos.indexOf(false) == -1) {
			break;
		} //#end if
		// If knownDone is still true at this point, we're ready to fill in the rest of the values.
		if (knownDone) {
			fillReady = true;
		}
	} //#next roundTry
	if (debugSort) debug("Took " + (roundTry + 1) + " tries: " + knownPos.toString())
	this.prettying = false;
}; //#end def sortBranch(r0=40, knownPos=None, debugSort=false)

/*
 * Map.sortLinear(b0=40)
 * Draws the "backbone" in a line.
 * TODO: Test and Understand everything that you need to understand for backbone.
 * FIXME: python3 ../loopsort1.py -n 10 -c 2 puts c and d in the same place.
 */
Map.prototype.sortLinear = function(b0) {
	if (b0 === undefined) { b0 = +40; }
	// //# Polar coordinates to the rescue!
	//loadPolar();
	var v = [];
	var i;
	for (i = 0; i < this.points.length; i++) {
		v[i] = [b0, 0];
	}
	this.setCartesian(v);
	this.prettying = false;
}; //#end def sortLinear(b0=40)

/*
 * Map.sortCircle(b0=40)
Draws the "backbone" in a circle.

 */
Map.prototype.sortCircle = function(b0) {
	if (b0 === undefined) { b0 = +40; }
	// //# Polar coordinates to the rescue!
	//loadPolar();
	var theta0 = tau / 12; //# 30 degrees
	var x = this.getPolarConn();
	var rev = x[0];
	var atom_hash = x[1];
	var v = [];
	var i;
	var i_x;
	var j;
	//bonds = defaultdict(int)
	var bonds = new Array(rev.length).fill(0);
	// for i, i_x in rev:
	for (j = 0; j < rev.length; j++) {
		//i = rev[j][0];
		i_x = rev[j][1];
		v.push([b0, theta0 * (i_x + bonds[i_x])]);
		bonds[i_x]++;
	}
	this.setCartesian(v);
	this.prettying = false;
}; //#end def sortCircle(b0=40)

/*
 * Map.sortSpiral(b0=40)
Draws the "backbone" in a spiral.
Sort in a novel way. Works for backbone only and loops well, but not branches.
TODO: take branches into account.
TODO: Make rows and cols variable.
TODO: Sort based on backbone
 */
Map.prototype.sortSpiral = function(b0) {
	if (b0 === undefined) { b0 = +40; }
	var rows = this.points.length;
	var cols = rows;
	var grid = [];
	for (i = 0; i < rows; i++) {
		grid.push(new Array(cols).fill(null));
	} //#next i
	var pos = [Math.floor(cols / 2), Math.floor(rows / 2)];
	//#grid[pos[1]][pos[0]] = self.points[0]
	var directions = [[1, 0], [0, 1], [-1, 0], [0, -1]];
	//#dir_names = 'right down left up'.split(' ')
	var currDirection = 0;
	//for p in self.points:
	for (i = 0; i < this.points.length; i++) {
		var p = this.points[i];
		//#print(p.name)
		var foundSpot = false;
		while (!foundSpot) {
			if (grid[pos[1]][pos[0]] == null) {
				grid[pos[1]][pos[0]] = p;
				foundSpot = true;
				break;
			} //#end if
			//# Move in the current direction.
			//#print(dir_names[currDirection])
			pos = [pos[0] + directions[currDirection][0], pos[1] + directions[currDirection][1]];
		} //#loop
		//# Move in the current direction until we can turn.
		while(true) {
			pos = [pos[0] + directions[currDirection][0], pos[1] + directions[currDirection][1]];
			if (grid[pos[1]][pos[0]] == null) {
				break;
			} //#end if
			//#print(dir_names[currDirection], 'because', grid[pos[1]][pos[0]].name)
		} //#loop
		//# If we can turn, turn.
		var nextDirection = (currDirection + 1) % 4;
		var peekPos = [pos[0] + directions[nextDirection][0], pos[1] + directions[nextDirection][1]];
		if (grid[peekPos[1]][peekPos[0]] == null) {
			currDirection = nextDirection;
			//#print(dir_names[currDirection], 'turn')
		} //#end if
	} //#next i
	//# TODO: Shrink the grid if necessary.
	gridToPos(grid, b0);
	this.centerBound();
	this.prettying = false;
}

/*
 * Map.sortLattice(b0=40)
Sort almost perfectly for completely unconnected nodes.
TODO: take connections into account.
TODO: Make rows and cols variable.
TODO: Limited sortLattice using knownPos
 */
Map.prototype.sortLattice = function(b0) {
	if (b0 === undefined) { b0 = 40; }
	var nodes = this.points.length;
	var side_len = Math.ceil(Math.sqrt(nodes));
	if (side_len < 2) {
		//# We don't want your tiny 3 node graph at this point.
		return;
	}
	var row_pos = new Vector2(this.width / 2, this.height / 2);
	var i = 0;
	var row, col, q;
	var mul = -1;
	for (row = 0; row < side_len; row++) {
		q = side_len;
		if (i + q > nodes) {
			// Last row needs a fix
			q = (nodes - i);
		}
		if (mul == 1) {
			mul = -1;
			row_pos.x += (side_len - 1) * b0;
		} else {
			mul = 1;
			row_pos.x -= (side_len - 1) * b0;
		} //#end if
		for (col = 0; col < q; col++) {
			this.points[i].x = row_pos.x + (mul * col * b0);
			this.points[i].y = row_pos.y;
			i++;
		} //#next col
		row_pos.x += b0 * -0.5;
		row_pos.y += b0 * 0.8660254037844386;
	} //#next row
	this.prettying = false;
}//#end def sortLattice(b0=40)

/*
 * addNodeToGroup(p, left, right, groups, currGroup=null)
	Static utility function, finds the right group and adds to right and left bracket functions.
	You are part of a group if you have two connections to the group.
	This function doesn't cover the case of the left bracket. That's handled in the calling function.
	currGroup is an optional attribute designed to optimize which groups we have to check.
 */
function addNodeToGroup(p, left, right, groups, currGroup) {
	if (currGroup === undefined) { currGroup = null; }
	//# TODO: improve this
	var i;
	var j;
	var myGroups = new Set();
	var myGroup = null;
	var checkGroups;
	if (currGroup) {
		//debug('antg\t' + currGroup.points.map(function(x) { return x.name.toString(); }).join(', '));
		checkGroups = [currGroup];
	} else {
		checkGroups = groups;
	} //#end if
	// This function checks if the point's connections are in checkGroups.
	//for q in p.conns:
	for(i = 0; i < p.conns.length; i++) {
		var q = p.conns[i];
		if (q.a == p) {
			q = q.b;
		} else {
			q = q.a;
		}
		//debug('q: ' + q.name);
		//for group in checkGroups:
		for(j = 0; j < checkGroups.length; j++) {
			var group = checkGroups[j];
			//# Can't add anything to a leaf.
			if (group.finished) continue;
			if (group.points.indexOf(q) >= 0) {
				//# I may belong to this group I suppose...
				if (myGroups.has(group)) {
					//# We found our group.
					myGroup = group;
					break;
				} else {
					myGroups.add(group);
				} //#end if
			} //#end if
		} //#next group
		if (myGroup != null) break;
	} //#next q
	//debug("myGroup: " + myGroup);
	if (myGroup === null && myGroups.size != 0) {
		//# Assume we're in the first group. Not very good idea.
		myGroup = Array.from(myGroups)[0];
	} //#end if
	if (myGroup === null) {
		//# First member creates the group
		myGroup = new MapGroup();
		myGroup.addPoint(p);
		groups.push(myGroup);
		if (left.length == 0) {
			//# Likely Left bracket
			myGroup.left_bracket = p;
		} else {
			//# Likely Right bracket
			myGroup.right_bracket = p;
		} //#end if
	} else {
		//# Add to existing group
		myGroup.addPoint(p);
		if (left.length == 0 && myGroup.left_bracket === null) {
			//# Likely Left bracket
			myGroup.left_bracket = p;
		} else if (right.length == 0 && myGroup.right_bracket === null) {
			//# Likely Right bracket
			myGroup.right_bracket = p;
		} //#end if
	} //#end if
	return myGroup;
} //#end def addNodeToGroup(p, left, right, groups, currGroup)

/*
 * Map.groupNodes() 
	Requires sortLinear to be called first.
	Returns a list of group objects:
		a list of nodes,
		a left bracket,
		a right bracket,
		a finished boolean
	This allows the organization much easier using a minor modification of sortLoop.
	Look at me algorithming.
 */
Map.prototype.groupNodes = function() {
	var groups = [];
	var currGroup = null;
	var i;
	var j;
	
	//for p in this.points:
	for (i = 0; i < this.points.length; i++) {
		var p = this.points[i];
		var left = [];
		var right = [];
		//for q in p.conns:
		for(j = 0; j < p.conns.length; j++) {
			var q = p.conns[j];
			if (q.a == p) {
				q = q.b;
			} else {
				q = q.a;
			}
			if (q.x < p.x) {
				left.push(q);
			} else {
				right.push(q);
			} //#end if
		} //#next q
		if (left.length == 0 || right.length == 0) {
			if (left.length == 1 || right.length == 1) {
				//# This is a leaf
				//debug('leaf ' + p.name);
				var leafGroup = new MapGroup();
				leafGroup.leaf(p);
				groups.push(leafGroup);
			} else {
				//# This is a node from a group
				//debug('node from a group ' + p.name + ' ' + left.length + ' ' + right.length);
				currGroup = addNodeToGroup(p, left, right, groups, currGroup);
				//debug("cglb: " + currGroup.left_bracket + " rb: " + currGroup.right_bracket);
			} //#end if
		} else {
			//# left and right have something in them.
			if (left.length == 1 || right.length == 1) {
				//# We've got a bracket
				if (left.length == 1) {
					//# Possible left bracket
					//# Check to see if I'm the start of the new group.
					var lpoints = [];
					var rpoints = [];
					//for q in this.points:
					for (j = 0; j < this.points.length; j++) {
						var q = this.points[j];
						//# Don't count yourself
						if (q == p) continue;
						if (q.x < p.x) {
							lpoints.push(q);
						} else {
							rpoints.push(q);
						} //#end if
					} //#next q
					//# Expensive on large maps for for for for.
					var newGroup = true;
					if (currGroup != null) {
						//for q in lpoints:
						for (j = 0; j < lpoints.length; j++) {
							var q = lpoints[j];
							//for r in q.conns:
							for (k = 0; k < q.conns.length; k++) {
								var r = q.conns[k];
								if (r.a == q) {
									r = r.b;
								} else {
									r = r.a;
								}
								if (rpoints.indexOf(r) >= 0) {
									newGroup = false;
									break;
								} //#end if
							} //#next r
							if (newGroup == false) break;
						} //#next q
					}
					if (newGroup) {
						//debug('left bracket ' + p.name);
						newGroup = new MapGroup();
						newGroup.addPoint(p);
						groups.push(newGroup);
						currGroup = newGroup;
					} else {
						//debug('node from group 3 ' + p.name + " " + currGroup);
						currGroup = addNodeToGroup(p, left, right, groups, currGroup);
						//debug('currGroup\t' + currGroup.points.map(function(x) { return x.name.toString(); }).join(', '));
					} //#end if
				} else {
					//# Possible right bracket
					//debug('right bracket ' + p.name);
					currGroup = addNodeToGroup(p, left, right, groups, currGroup);
				} //#end if
			} else {
				//# It's a group node apparently
				//debug('node from group 2 ' + p.name);
				currGroup = addNodeToGroup(p, left, right, groups, currGroup);
			} //#end if
		} //#end if
	} //#next p
	
	// Fix the problem with node 0 in one group and the rest in another.
	var k, m;
	for (i = 0; i < groups.length; i++) {
		var group = groups[i];
		if(group.points.length > 2) { continue; }
		// One or two nodes in a group on their own.
		var foundConcat = null;
		for (j = 0; j < groups.length; j++) {
			if (i == j) continue;
			for (k = 0; k < group.points.length; k++) {
				var p = group.points[k];
				var conns_in_G = 0;
				//for q in p.conns:
				for(m = 0; m < p.conns.length; m++) {
					var q = p.conns[m];
					if (q.a == p) {
						q = q.b;
					} else {
						q = q.a;
					}
					if (groups[j].points.indexOf(q) >= 0) {
						conns_in_G++;
					} //#end if
				} //#next q
				//debug("conns in G: " + conns_in_G + " " + k);
				if (conns_in_G >= 2) {
					foundConcat = j;
					break;
				}//#end if
			} //#next p
			if (foundConcat !== null) break;
		}
		//debug("Concat " + foundConcat + " " + i);
		if (foundConcat !== null) {
			// Move all group members to the found group.
			for (k = 0; k < group.points.length; k++) {
				groups[foundConcat].addPoint(group.points[k]);
			}
			group.points = [];
		}
	}
	
	// Fix the problem of loop_bug3.json where 0 is conn to 5 6
	// Looking for group A is small and connected to group B and C.
	// Group B is small and connected to C also.
	var groupsConns = [];
	for (i = 0; i < groups.length; i++) {
		var group = groups[i];
		if(group.points.length > 2) { 
			// Have to put something here for consistency.
			groupsConns.push([]);
			continue; 
		}
		// One or two nodes in a group on their own.
		// TODO: We can skip half of these.
		var groupConn = new Array(groups.length).fill(false);
		for (j = 0; j < groups.length; j++) {
			if (i == j) continue;
			for (k = 0; k < group.points.length; k++) {
				var p = group.points[k];
				var conns_in_G = 0;
				//for q in p.conns:
				for(m = 0; m < p.conns.length; m++) {
					var q = p.conns[m];
					if (q.a == p) {
						q = q.b;
					} else {
						q = q.a;
					}
					if (groups[j].points.indexOf(q) >= 0) {
						conns_in_G++;
						break;
					} //#end if
				} //#next q
				//debug("conns in G: " + conns_in_G + " " + k);
				if (conns_in_G > 0) {
					groupConn[j] = true;
					break;
				}//#end if
			} //#next p
		}
		var foundConcat = null;
		var foundConcat2 = null;
		// look for A in groupConns.
		for (j = 0; j < groupsConns.length; j++) {
			// We are connected to them.
			if (groupConn[j]) {
				// Look for the group we are both connected to.
				for (k = 0; k < groups.length; k++) {
					if (groupConn[k] && groupsConns[j].length > k && groupsConns[j][k]) {
						// We have found a loop, add ourselves and them.
						foundConcat = k;
						foundConcat2 = j;
						break;
					}
				}
			}
		}
		//debug("Concat 2 " + foundConcat + " " + i);
		if (foundConcat !== null) {
			// Move all group members to the found group.
			for (k = 0; k < group.points.length; k++) {
				groups[foundConcat].addPoint(group.points[k]);
			}
			group.points = [];
			group = groups[foundConcat2];
			for (k = 0; k < group.points.length; k++) {
				groups[foundConcat].addPoint(group.points[k]);
			}
			group.points = [];
			// Have to put something here for consistency.
			groupsConns.push([]);
		} else {
			groupsConns.push(groupConn);
		}
	}
	
	//# FIXME: when a group consumes a second group, cull doesn't work.
	//# python3 ../loopsort1.py -x 26 -n 20 -c 2
	//# Fix the bug with groups consuming more nodes than just their group.
	//# python3 ../loopsort1.py -x 2 -n 20 -c 1
	var newGroups = [];
	//for group in groups:
	for (i = 0; i < groups.length; i++) {
		//debug('pre\t' + groups[i].points.map(function(x) { return x.name.toString(); }).join(', '));
		var nodesOrphaned = groups[i].cull();
		//debug('orphans\t' + nodesOrphaned.map(function(x) { return x.name.toString(); }).join(', '));
		newGroups = newGroups.concat(nodesOrphaned);
	} //#next group
	//debug("newGroups: " + JSON.stringify(newGroups));
	groups = groups.filter(function (x) { return x.points.length > 0; });
	/*for (i = 0; i < groups.length; i++) {
		debug('af\t' + groups[i].points.map(function(x) { return x.name.toString(); }).join(', '));
	}*/
	//for p in newGroups:
	for (i = 0; i < newGroups.length; i++) {
		var stemGroup = new MapGroup();
		stemGroup.addPoint(newGroups[i]);
		groups.push(stemGroup);
	} //#next p
	//# Sort by minimum x value.
	groups.sort(function(g, g2) { 
		var gmx = g.minX();
		var g2mx = g2.minX();
		if (gmx < g2mx) { return -1; }
		if (gmx > g2mx) { return 1; }
		return 0;
	});
	return groups;
} //#end def groupNodes()

/*
 * Map.pointToIndex()
 * Create a mapping from point to index.
 * 
 */
Map.prototype.pointToIndex = function() {
	//# Reverse point to index.
	mapPointToIndex = {};
	var i;
	for (i = 0; i < this.points.length; i++) {
		var p = this.points[i];
		mapPointToIndex[p.name] = i;
	} //#next i, p
	return mapPointToIndex;
} //#end def pointToIndex()

/*
 * Map.solveDoubleLoop(group, inBetweensSet, ungrouped, knownPos, level=0)
	Solve double diamond problem and so forth.
	a--b--c--d--e--f--g
	a--d
	d--g
	As you can see, there is just one group in this graph.
 */
Map.prototype.solveDoubleLoop = function(group, inBetweens, ungrouped, knownPos, level) {
	if (level === undefined) { level = 0; }
	//# If this turns up more groups than the original, this is a cull bug we can fix.
	//# This seems strange, but believe me it works in all known problem cases.
	var groups = this.groupNodes();
	var i;
	var j;
	//for group_new in groups:
	for (i = 0; i < groups.length; i++) {
		var group_new = groups[i];
		//debug(' &nbsp; &nbsp;' + group_new.points.map(function(x) { return x.name.toString(); }).join(', '));
	} //#next group
	var equal_groups = false;
	var group_picked = null;
	var first_ungrouped = ungrouped[0];
	//for group_new in groups:
	for (i = 0; i < groups.length; i++) {
		var group_new = groups[i];
		if (group_new.points.indexOf(first_ungrouped) >= 0) {
			//# This is our group, so check to see if it is equal to group.
			if (group_new.points.length == group.points.length) {
				//# Same size, must be same data.
				equal_groups = true;
				group_picked = group_new;
			} else {
				//# Different group.
				equal_groups = false;
				group_picked = group_new;
			} //#end if
			break;
		} //#end if
	} //#next group
	mapPointToIndex = this.pointToIndex();
	var ungrouped_nodes = 0;
	if (equal_groups) {
		//# double diamond problem
		//# given a b c, find d, then sortGroup?
		var connectors = new Set();
		//for x in ungrouped:
		for (i = 0; i < ungrouped.length; i++) {
			var x = ungrouped[i];
			//debug("x: " + x.name + ", ibs: " + inBetweens.length);
			//for v in inBetweensSet:
			for(j = 0; j < inBetweens.length; j++) {
				var v = inBetweens[j];
				if (checkPointInConns(v, x.conns)) {
					//# this is d.
					connectors.add(v);
				} //#end if
			} //#next v
		} //#next x
		debug('Found ' + connectors.size + ' connectors');
		if (connectors.size >= 0 && connectors.size <= 2) {
			//# This almost solves double_diamond.json, but the problem is overlap.
			//# This is a singly or doubly connected thing, easy, just sort group.
			var unGroupedGroup = new MapGroup();
			//# Use d as the known, this is bound to cause problems, but...
			//# This is a trick to get consistency for python3 ../loopsort1.py -x 31 -n 20 -c 2
			var connectors_sorted = Array.from(connectors);
			//connectors_sorted.sort(key=lambda x: x.x);
			connectors_sorted.sort(function (x, y) {
				if(x.x < y.x) {
					return -1;
				}
				if(x.x > y.x) {
					return 1;
				}
				return 0;
			});
			unGroupedGroup.points = connectors_sorted.concat(ungrouped);
			ungrouped_nodes = this.sortGroup(unGroupedGroup, knownPos, mapPointToIndex, 40, level + 1);
		} //#end if
	} else {
		//# We lucked out, it's a cull problem
		//# This is a very cheap algorithm I assume.
		//# python3 ../loopsort1.py -x 42 -n 20 -c 2
		var groups_solve = new Set();
		//for x in ungrouped:
		for (i = 0; i < ungrouped.length; i++) {
			var x = ungrouped[i];
			//for group_new in groups:
			for (j = 0; j < groups.length; j++) {
				var group_new = groups[j];
				if (group_new.points.indexOf(x) >= 0) {
					groups_solve.add(group_new);
				} //#end if
			} //#next group_new
		} //#next x
		//#print('To solve:')
		var groups_solve_list = Array.from(groups_solve);
		ungrouped_nodes = 0;
		//for group_new in groups_solve:
		for(i = 0; i < groups_solve_list.size; i++) {
			var group_new = groups_solve_list[i];
			//#print('\t', ', '.join([str(p.name) for p in group_new.points]))
			if (group_new.points.length == 1) {
				continue;
			} else if (group_new.points.length == 2) {
				continue;
			} //#end if
			ungrouped_nodes += this.sortGroup(group_new, knownPos, mapPointToIndex, 40, level + 1);
		} //#next group_new
	} //#end if
	
} //#end def solveDoubleLoop(group, inBetweensSet, ungrouped, level=0)

/*
 * Map.sortLoop(b0=40, level=0)
	Start small, more and more because it will improve the speed of algorithm and optimize.
	Assumes that sortLinear has already been called.
	Uses sortBranch in combination with sortLoop to fix branches.
 */
Map.prototype.sortLoop = function(b0, level) {
	if (b0 === undefined) { b0 = 40; }
	if (level === undefined) { level = 0; }
	var groups = this.groupNodes();
	var i;
	//for group in groups:
	for (i = 0; i < groups.length; i++) {
		var group = groups[i];
		debug('Group:\t' + group.points.map(function(x) { return x.name.toString(); }).join(', '));
	} //#next group
	
	//knownPos = [False for i in self.points]
	var knownPos = new Array(this.points.length).fill(false);
	//# Reverse point to index dictionary.
	var mapPointToIndex = this.pointToIndex();
	//#print([x.name for x in this.points[5].conns])
	//# TODO: optimize this double loop if it becomes an issue (so far not)
	//# FIXME: currX is not threadsafe but who would run two sorts at the same time?
	this.currX = null;
	//#addX = 0
	var ungrouped_nodes = 0;
	//for group in groups:
	for (i = 0; i < groups.length; i++) {
		var group = groups[i];
		if (group.points.length < 1) {
			continue;
		} else if (group.points.length == 1) {
			//# Single node groups are leaves or stems. For example: a--b--c--d--e
			p = group.points[0];
			if (this.currX === null) this.currX = p.x;
			//#addX += this.currX - p.x
			//#print('addX', p.name, addX)
			p.x = this.currX;
			//#print('this.currX', p.name, this.currX)
			this.currX += b0;
			//# TODO: May or may not be necessary for leaves:
			//#knownPos[mapPointToIndex[p]] = true
			continue;
		} else if (group.points.length == 2) {
			//# A two point group? That's an error.
			debug("Warning: Groups of two are pretty rare. One might say they shouldn't exist.");
			p = group.points[0];
			if (this.currX === null) this.currX = p.x;
			//#addX += this.currX - p.x
			p.x = this.currX;
			//#print('this.currX', p.name, this.currX)
			this.currX += b0;
			q = group.points[1];
			//#addX += this.currX - q.x
			q.x = this.currX;
			this.currX += b0;
			//# TODO: May or may not be necessary for leaves:
			//#knownPos[mapPointToIndex[p]] = true
			//#knownPos[mapPointToIndex[q]] = true
			continue;
		} //#end if
		ungrouped_nodes += this.sortGroup(group, knownPos, mapPointToIndex, b0);
	} //#next group
	//# If we know nothing, then just run sortBranch normally.
	//if knownPos == [false for i in this.points]:
	if(knownPos.indexOf(true) == -1) {
		knownPos = null;
	} //#end if
	this.sortBranch(b0, knownPos);
	if (ungrouped_nodes > 1 && level < 1) {
		//# One expensive not yet working solution is to repeat sort by loop.
		//#this.sortLoop(b0, level+1)
		//pass
	} //#end if
	this.prettying = false;
} //#end def sortLoop(b0=40, level=0)

/*
 * sortGroup(group, knownPos, mapPointToIndex, b0=40, level=0)
	Chooses the position of a group based on shape.
	returns the number of nodes in the group that were not positioned.
	//# Pentagon and up
	//# Pentagon: python3 loopsort1.py -x 8
	//# loopsort1_162.json
	//# Hexagon: python3 loopsort1.py -x 9
	//# loopsort1_163.json
	//# Septagon: python3 loopsort1.py -x 7
	//# loopsort1_164.json
	//# Octagon: python3 loopsort1.py -x 12
	//# loopsort1_167.json
	//# Octdecagon: python3 loopsort1.py -x 10
	//# loopsort1_165.json
	//# The first one goes to 180 - 360/shape. cos((pi-(2*pi/shape))), sin((pi-(2*pi/shape)))

	//# 108°
	theta1 = (pi-(2*pi/shape))
	pos1 = (cos(theta1), sin(theta1))
	//# 72°
	theta2a = pi - theta1
	//# 18°
	theta2b = (pi*0.5) - theta2a
	//# 108+18-90 = 36°, we could just remove pi*0.5 from theta2b if we wanted to.
	theta2b = theta2b + theta1 - (pi*0.5)
	pos2a = (cos(theta2b), sin(theta2b))
	pos2 = (pos1[0]+pos2a[0], pos1[0]+pos2a[0])
 */
Map.prototype.sortGroup = function(group, knownPos, mapPointToIndex, b0, level) {
	if (b0 === undefined) { b0 = 40; }
	if (level === undefined) { level = 0; }
	//# TODO: Ton of little stuff here..
	var p = null;
	var q = null;
	var i;
	var best_q;
	if (group.left_bracket && group.right_bracket) {
		debug('lb rb');
		p = group.left_bracket;
		q = group.right_bracket;
		//if q not in p.conns:
		if(!checkPointInConns(q, p.conns)) {
			//# Find the correct q in p.conns
			best_q = null;
			maxConns = 0;
			//for q_test in p.conns:
			for (qti = 0; qti < p.conns.length; qti++) {
				var q_test = p.conns[qti].a;
				if (q_test == p) q_test = p.conns[qti].b;
				if (q_test.x > p.x && q_test.conns.length > maxConns) {
					maxConns = q_test.conns.length;
					best_q = q_test;
				} //#end if
			} //#next q_test
			q = best_q;
			if (q === null) {
				debug("Error: Unable to find a valid q 1");
				q = group.right_bracket;
			} //#end if
		} //#end if
		debug('p=' + p.name + ', q=' + q.name);
	} //#end if

	//# Find p and q amongst the group. Naturally the furthest left.
	if (group.left_bracket) {
		debug('lb');
		p = group.left_bracket;
	} else {
		debug('*shrug*')
		p = group.points[0];
	} //#end if
	//#if q not in p.conns:
	//# Find the correct q in p.conns
	best_q = null;
	maxConns = 0;
	var qti;
	//for q_test in p.conns:
	for (qti = 0; qti < p.conns.length; qti++) {
		var q_test = p.conns[qti].a;
		if (q_test == p) q_test = p.conns[qti].b;
		//#print(q_test.name, 'has', q_test.conns.length, 'conns')
		//# We don't want any other groups in our group.
		if (group.points.indexOf(q_test) < 0) continue;
		if (q_test.conns.length > maxConns) {
			maxConns = q_test.conns.length;
			best_q = q_test;
		} //#end if
	} //#next q_test
	q = best_q;
	if (q === null) {
		debug("Error: Unable to find a valid q 2 " + group.right_bracket);
		q = group.right_bracket;
		// For now avoid error.
		if (q === null) {
			return 0;
		}
	} //#end if
	//#end if
	//debug('p=' + p.name + ', q=' + q.name);

	//# Find p, q, and inBetweens correctly!
	inBetweens = findRoute(p, q);
	
	if (inBetweens != null && inBetweens.length > group.points.length) {
		//# We're using parts of a connected group, verify that we're using 
		//# the least possible of it.
		var inBetweensSet = new Set(inBetweens);
		var groupSet = new Set(group.points);
		//diff = inBetweensSet.difference(groupSet)
		var diff = inBetweens.filter(function(item) {
			return !groupSet.has(item);
		});
		var found_unconn = [];
		//for v in diff:
		var j, k;
		for(j = 0; j < diff.length; j++) {
			var v = diff[j];
			var connected = false;
			// for q in groupSet:
			for (k = 0; k < groupSet.length; k++) {
				var q = groupSet[k];
				if (checkPointInConns(q, v.conns)) {
					connected = true;
					break;
				} //#end if
			} //#next q
			if (!connected) {
				found_unconn.push(v);
				//#break
			} //#end if
		} //#next v
		if (found_unconn.length > 0) {
			// [v.name for v in found_unconn]
			debug("Found unconnected nodes: " + found_unconn.map(function (v) { return v.name.toString(); }).toString());
			//# Find a route that avoids one of the unconnected.
			var inBetweens_test = findRoute(p, q, [found_unconn[0]]);
			if (inBetweens_test != null) {
				inBetweens_testSet = new Set(inBetweens_test);
				var diff_test = Array.from(inBetweens_testSet).filter(function(item) {
					return !groupSet.has(item);
				});
				//# Choose the shorter of the two:
				if (diff_test.length < diff.length) {
					debug("Found a shorter route: " + inBetweens.map(function (v) { return v.name.toString(); }).toString() + " " + inBetweens_test.map(function (v) { return v.name.toString(); }).toString());
					inBetweens = inBetweens_test;
				} //#end if
			} //#end if
		} //#end if
	} //#end if
	
	if (this.currX === null) { this.currX = p.x; }
	//#p.x = this.currX
	//#print('this.currX', p.name, this.currX)
	
	//#TODO: Fix addX here.
	//#addX += this.currX - p.x
	//# There's always two nodes inline, the rest are perpendicular.
	this.currX += b0 * 2;
	//# This can be fixed by setting the shape based on len of inBetweens, I suppose.
	//#if len(inBetweens) != (shape):
	//#	print("inBetweens are different size.", ', '.join([str(x) for x in inBetweens]))
	//#	continue
	//#end if
	if (inBetweens == null)
	{
		debug("findRoute bug 2 " + group.points.map(function (v) { return v.name.toString(); }).toString());
		return 0;
	}
	var shape = inBetweens.length;
	//# Simple bug to fix.
	if (shape == 1) {
		debug("findRoute bug " + group.points.map(function (v) { return v.name.toString(); }).toString());
		return 0;
	} //#end if
	debug("Picked ngon " + shape + " " + p.name + " " + inBetweens.map(function (v) { return v.name.toString(); }).toString() + " " + group.points.map(function (v) { return v.name.toString(); }).toString());
	//# Using this method make a loop over shape for each vertex
	var prevPos = [p.x, p.y];
	var pos = [];
	var theta_inner = (2*Math.PI/shape);
	//for i in range(1, shape):
	for(i = 1; i < shape; i++) {
		var theta_i = Math.PI-(theta_inner*i);
		var pos_i = [b0*Math.cos(theta_i), b0*Math.sin(theta_i)];
		pos.push([pos_i[0] + prevPos[0], pos_i[1] + prevPos[1]]);
		prevPos = pos[pos.length-1];
	} //#next i
	//# Skip p.
	//for i, x in enumerate(inBetweens[1:]):
	for (i = 0; i < inBetweens.length-1; i++) {
		//var x = inBetweens[i+1];
		inBetweens[i+1].x = pos[i][0];
		inBetweens[i+1].y = pos[i][1];
	} //#next i, x
	//# Move all the others left shape - 2.
	/*
	dx = b0 * (shape - 2)
	further = q
	if q.x < p.x: further = p
	for ttr in this.points:
		if ttr in inBetweens: continue
		if ttr.x >= further.x:
			ttr.x -= dx
		} //#end if
	} //#next ttr
	*/
	knownPos[mapPointToIndex[p.name]] = true;
	knownPos[mapPointToIndex[q.name]] = true;
	//for inBetween in inBetweens:
	for (i = 0; i < inBetweens.length; i++) {
		knownPos[mapPointToIndex[inBetweens[i].name]] = true;
		//# Solve the moving issue.
		//#inBetween.x += addX
	} //#next inBetween
	
	//# Detect multiple loops in a single group
	ungrouped_nodes = 0;
	//# level < 1 needed to avoid infinite recursion
	if (inBetweens.length < group.points.length && level < 1) {
		var inBetweensSet = new Set(inBetweens);
		var groupSet = new Set(group.points);
		//var ungrouped = groupSet.difference(inBetweensSet);
		var ungrouped = group.points.filter(function(item) {
			return !inBetweensSet.has(item);
		});
		//debug('Ungrouped: ' + ungrouped.length + " " + JSON.stringify(ungrouped.map(function (v) { return v.name.toString(); }).join(', ')));
		//debug('In Betweens: ' + inBetweens.length + " " + JSON.stringify(inBetweens.map(function (v) { return v.name.toString(); }).join(', ')));
		this.solveDoubleLoop(group, inBetweens, ungrouped, knownPos, level);
		ungrouped_nodes = ungrouped.length;
	} //#end if
	return ungrouped_nodes;
} //#end def sortGroup(group, knownPos, mapPointToIndex, b0=40, level=0)

/**
 * Map.toGrid(width=40)
 * 
 * Unlike sortGrid, toGrid uses input positions to decide which cell to place 
 * nodes in. It returns a grid (array of array of points and/or arrays).
 * TODO: Improve the quality of position picking.
 **/
Map.prototype.toGrid = function(width) {
	if (width === undefined) { width = 40; }
	var r = this.bounds();
	var dx = (+r[2]) - (+r[0]);
	var dy = (+r[3]) - (+r[1]);
	var cols = Math.ceil(2 * dx / width);
	var rows = Math.ceil(2 * dy / width);
	var nearby = [[0, 1], [1, 1], [1, 0], [0, -1], [-1, 0], [-1, 1], [-1, -1], [1, -1]];
	var grid = [];
	var i, j;
	// for i in range(rows):
	for (i = 0; i < rows; i++) {
		grid.push(new Array(cols).fill(null));
	} //#next i
	var min_x = r[0];
	var min_y = r[1];
	// for p in self.points:
	for (i = 0; i < this.points.length; i++) {
		var p = this.points[i];
		var row = Math.round((p.y - min_y) / width);
		var col = Math.round((p.x - min_x) / width);
		if (grid[row][col]) {
			// Check for a nearby empty spot.
			var found = false;
			for(j = 0; j < nearby.length; j++) {
				if (row+nearby[j][1] < 0) continue;
				if (col+nearby[j][0] < 0) continue;
				if(grid[row+nearby[j][1]][col+nearby[j][0]] == null) {
					grid[row+nearby[j][1]][col+nearby[j][0]] = p;
					found = true;
					break;
				}
			}
			if (!found) {
				if (Array.isArray(grid[row][col])) {
					grid[row][col].push(p);
				} else {
					grid[row][col] = [grid[row][col], p];
				} //#end if
			}
		} else {
			grid[row][col] = p;
		} //#end if
	} //#next p
	//# Should this be optional? Perhaps.
	gridToPos(grid, width);
	return grid;
} //#end def toGrid(width=40)


/**
 * Map.sortRightLoop(b0=40)
 * 
 * A simple algorithm using breadth first.
 **/
Map.prototype.sortRightLoop = function(b0)
{
	if (b0 === undefined) { b0 = 40; }
	var nodes = this.points.length;
	if (nodes < 1) return;
	this.points[0].setPos(new Vector2(0, 0));
	//# Guaranteed to be sorted if nodes == 0 or 1.
	if (nodes < 2) return;
	var knowns = [this.points[0]];
	var neighbors = [];
	var k;
	for(k = 0; k < this.points[0].conns.length; k++) {
		var conn = this.points[0].conns[k].b;
		if (conn == this.points[0]) {
			conn = this.points[0].conns[k].a;
		}
		neighbors.push(conn);
	}
	var x = b0;
	var roundTry;
	for(roundTry = 0; roundTry < nodes; roundTry++) {
		var newNeighbors = [];
		var y = (neighbors.length - 1)*-b0*0.5;
		var i;
		//for point in neighbors:
		for (i = 0; i < neighbors.length; i++) {
			var point = neighbors[i];
			//# Put point on a line to the right depending on how many neighbors.
			point.x = x;
			point.y = y;
			y += b0;
			var args = [newNeighbors.length, 0];
			var conns = [];
			var k;
			for(k = 0; k < point.conns.length; k++) {
				var conn = point.conns[k].b;
				if (conn == point) {
					conn = point.conns[k].a;
				}
				conns.push(conn);
			}
			args = args.concat(conns);
			newNeighbors.splice.apply(newNeighbors, args);
			knowns.push(point);
		} //#next point
		x += b0;
		// FIXME: Javascript doesn't have this.
		//neighbors = [p for p in newNeighbors if p not in knowns]
		neighbors = [];
		var j;
		for (j = 0; j < newNeighbors.length; j++) {
			if ((knowns.indexOf(newNeighbors[j]) < 0) && (neighbors.indexOf(newNeighbors[j]) < 0)) {
				neighbors.push(newNeighbors[j]);
			}
		}
		if (knowns.length >= nodes) break;
		/*if (neighbors.length == 0) {
			debug("No neighbors wth? k: " + knowns.length + ", n: " + nodes);
		}*/
	} //#next roundTry
	this.centerBound();
	this.prettying = false;
} //#end def sortRightLoop(b0=40)

/**
 * Map.sortBest(width=40, rearrange=True)
 * 
 * sortGrid, sortBranch, and sortLoop, pick whichever has lowest potential.
 **/
Map.prototype.sortBest = function(width, rearrange) {
	if (width === undefined) { width = 40; }
	if (rearrange === undefined) { rearrange = true; }
	var lj_conf = new LJ_Potential();
	this.sortBranch(width);
	var x0C = this.getX0Conns(true);
	var x_branch = x0C[0];
	var conns = x0C[1];
	var nonConns = getNonConns(conns, true);
	var V_branch = V_SmallWideWorld(x_branch, conns, width, lj_conf, nonConns);
	this.sortLinear(width);
	this.sortLoop(width);
	var x_loop = this.getX0Conns()[0];
	var V_loop = V_SmallWideWorld(x_loop, conns, width, lj_conf, nonConns);
	var V_best;
	var bestSort;
	if (V_loop > V_branch) {
		V_best = V_branch;
		bestSort = 'branch';
	} else {
		V_best = V_loop;
		bestSort = 'loop';
	}
	this.sortLattice(width);
	var x_lattice = this.getX0Conns()[0];
	var V_lattice = V_SmallWideWorld(x_lattice, conns, width, lj_conf, nonConns);
	if (V_best > V_lattice) {
		V_best = V_lattice;
		bestSort = 'lattice';
	}
	this.sortSpiral(width);
	var x_spiral = this.getX0Conns()[0];
	var V_spiral = V_SmallWideWorld(x_spiral, conns, width, lj_conf, nonConns);
	if (V_best > V_spiral) {
		V_best = V_spiral;
		bestSort = 'spiral';
	}
	this.sortRightLoop(width);
	var x_rightloop = this.getX0Conns()[0];
	var V_rightloop = V_SmallWideWorld(x_rightloop, conns, width, lj_conf, nonConns);
	if (V_best > V_rightloop) {
		V_best = V_rightloop;
		bestSort = 'rightloop';
	}

	this.sortGrid(width, rearrange);
	var x_grid = this.getX0Conns()[0];
	var V_grid = V_SmallWideWorld(x_grid, conns, width, lj_conf, nonConns);
	var V_min = V_grid;
	if (V_grid > V_best) {
		debug('lowest potential is ' + bestSort + ' ' + V_best);
		// sortBranch is quick enough to just rerun.
		if (bestSort == 'branch') {
			this.sortBranch(width);
		} else if (bestSort == 'lattice') {
			this.sortLattice(width);
		} else if (bestSort == 'spiral') {
			this.sortSpiral(width);
		} else if (bestSort == 'rightloop') {
			this.sortRightLoop(width);
		} else {
			this.sortLinear(width);
			this.sortLoop(width);
		}
		V_min = V_best;
		this.centerBound();
	}
	this.prettying = false;
	return V_min;
}

/**
 * Map.sortGrid(width=40, rearrange=True)
 * 
 * Create an NxM grid, place one node per cell. This simplifies the branch sort algorithm into a lattice instead of a trigonometry problem.
 **/
Map.prototype.sortGrid = function(width, rearrange) {
	if (width === undefined) { width = 40; }
	if (rearrange === undefined) { rearrange = true; }
	var nodes = this.points.length;
	if (nodes <= 1) {
		//# Congratulations, you're guaranteed to be sorted.
		return;
	} //#end if
	//# Allocating all at the start will become a mess with large values of num_points, so we make it extensible.
	var knownPos = new Array(nodes).fill(false);
	var grid_rows = 2;
	var grid_cols = 2;
	var grid = [];
	var i;
	for (i = 0; i < grid_rows; i++) {
		grid.push(new Array(grid_cols).fill(null));
	} //#next i
	//#debug(grid)
	grid[0][0] = this.points[0];
	knownPos[0] = true;
	//# This is bound to change because of extensible. That sucks.
	var pointPosMap = {0: [0, 0]};
	//# FIXME: Not sure what the right number is, need some cheap boolean detector.
	var roundTry;
	// for roundTry in range(nodes):
	for (roundTry = 0; roundTry < nodes; roundTry++) {
		//#debug("fillReady?", fillReady)
		var knownDone = true;
		//for i in range(nodes):
		for (i = 0; i < nodes; i++) {
			var givens = [];
			var unknowns = [];
			//# Set the position of its connections.
			//for conn in self.points[i].conns:
			var k;
			for(k = 0; k < this.points[i].conns.length; k++) {
				var conn = this.points[i].conns[k].b;
				if (conn == this.points[i]) {
					conn = this.points[i].conns[k].a;
				}
				var j = this.points.indexOf(conn);
				if (knownPos[j]) {
					givens.push(j);
					continue;
				} //#end if
				unknowns.push(j);
			} //#next conn
			if (knownPos[i]) {
				if (unknowns.length == 0) {
					//# Everything is known, we're good.
					//#debug('Yay', map1.points[i].x, map1.points[i].y)
					continue;
				} //#end if
				//#debug('sorting from', map1.points[i].name)
				var j = 0;
				//# Put the unknowns on the grid near the position of the point.
				var pos_i = pointPosMap[i];
				//#debug(pointPosMap, i)
				var left = [pos_i[0]-1, pos_i[1]];
				var right = [pos_i[0]+1, pos_i[1]];
				var up = [pos_i[0], pos_i[1]-1];
				var down = [pos_i[0], pos_i[1]+1];
				var ul = [pos_i[0]-1, pos_i[1]-1];
				var ur = [pos_i[0]+1, pos_i[1]-1];
				var dl = [pos_i[0]-1, pos_i[1]+1];
				var dr = [pos_i[0]+1, pos_i[1]+1];
				var change = [0, 0];
				var choices = [left, right, up, down, ul, ur, dl, dr];
				var m;
				//for p in [left, right, up, down, ul, ur, dl, dr]:
				for (m = 0; m < choices.length; m++) {
					var p = choices[m];
					//# Move the p where it should go based on additions.
					p[0] += change[0];
					p[1] += change[1];
					//#debug('try', p)
					var dp = null;
					if (p[0] + change[0] < 0) {
						//# available, but we will need to extend left, maybe up or down
						//# Extend left
						//#debug(map1.points[unknowns[j]].name, end=' ')
						//#debug('extend left', end='')
						//for k in range(grid_rows):
						for(k = 0; k < grid_rows; k++) {
							grid[k] = [null].concat(grid[k]);
						} //#next k
						grid_cols++;
						dp = [1, 0];
						change[0]++;
						if (p[1] < 0) {
							//# Extend up
							//#debug(', extend up', end='')
							grid.splice(0, 0, new Array(grid_cols).fill(null));
							grid_rows++;
							dp[1] = 1;
							change[1]++;
						} //#end if
						if (p[1] > grid_rows) {
							//# Extend down
							//#debug(', extend down', end='')
							grid.push(new Array(grid_cols).fill(null));
							grid_rows++;
						} //#end if
						//#debug()
						grid[p[1]+dp[1]][p[0]+dp[0]] = this.points[unknowns[j]];
						pointPosMap[unknowns[j]] = [p[0], p[1]];
						knownPos[unknowns[j]] = true;
						//#debug("set %s to %r" % (map1.points[unknowns[j]].name, p))
						j++;
						if (j >= unknowns.length) {
							break;
						} //#end if
					} else if (p[0] >= grid_cols) {
						//# available, but we will need to extend right
						//# Extend right
						//#debug(map1.points[unknowns[j]].name, end=' ')
						//#debug('extend right', end='')
						//for k in range(grid_rows):
						for(k = 0; k < grid_rows; k++) {
							grid[k].push(null);
						} //#next k
						grid_cols++;
						dp = [0, 0];
						if (p[1] < 0) {
							//# Extend up
							//#debug(', extend up', end='')
							grid.splice(0, 0, new Array(grid_cols).fill(null));
							grid_rows++;
							dp[1] = 1;
							change[1]++;
						} //#end if
						if (p[1] > grid_rows) {
							//# Extend down
							//#debug(', extend down', end='')
							grid.push(new Array(grid_cols).fill(null));
							grid_rows++;
						} //#end if
						//#debug()
						grid[p[1]+dp[1]][p[0]+dp[0]] = this.points[unknowns[j]];
						pointPosMap[unknowns[j]] = [p[0], p[1]];
						knownPos[unknowns[j]] = true;
						//#debug("set %s to %r" % (map1.points[unknowns[j]].name, p))
						j++;
						if (j >= unknowns.length) {
							break;
						} //#end if
					} else if (p[1] < 0) {
						//# available, but we will need to extend up
						//# Extend up
						//#debug(map1.points[unknowns[j]].name, end=' ')
						//#debug('extend up')
						grid.splice(0, 0, new Array(grid_cols).fill(null));
						grid_rows++;
						dp = [0, 1];
						change[1]++;
						grid[p[1]+dp[1]][p[0]+dp[0]] = this.points[unknowns[j]];
						pointPosMap[unknowns[j]] = [p[0], p[1]];
						knownPos[unknowns[j]] = true;
						//#debug("set %s to %r" % (map1.points[unknowns[j]].name, p))
						j++;
						if (j >= unknowns.length) {
							break;
						} //#end if
					} else if (p[1] >= grid_rows) {
						//# available, but we will need to extend down
						//# Extend down
						//#debug(map1.points[unknowns[j]].name, end=' ')
						//#debug('extend down')
						grid.push(new Array(grid_cols).fill(null));
						grid_rows++;
						grid[p[1]][p[0]] = this.points[unknowns[j]];
						pointPosMap[unknowns[j]] = [p[0], p[1]];
						knownPos[unknowns[j]] = true;
						//#debug("set %s to %r" % (map1.points[unknowns[j]].name, p))
						j++;
						if (j >= unknowns.length) {
							break;
						} //#end if
					} else if (grid[p[1]][p[0]] == null) {
						//#debug(map1.points[unknowns[j]].name, end=' ')
						//#debug('in spot')
						//# available, put it there.
						grid[p[1]][p[0]] = this.points[unknowns[j]];
						pointPosMap[unknowns[j]] = [p[0], p[1]];
						knownPos[unknowns[j]] = true;
						//#debug("set %s to %r" % (map1.points[unknowns[j]].name, p))
						j++;
						if (j >= unknowns.length) {
							break;
						} //#end if
					} //#end if
					//# Fix the pointPosMap using generated dp. 
					//# TODO: do this after the loop for optimization.
					if (dp && dp != [0, 0]) {
						//#debug("Fix ppm using", dp)
						//for v in pointPosMap:
						var v_i;
						var pointPosMapKeys = Object.keys(pointPosMap);
						for(v_i = 0; v_i < pointPosMapKeys.length; v_i++) {
							var v = pointPosMapKeys[v_i];
							pointPosMap[v][0] += dp[0];
							pointPosMap[v][1] += dp[1];
						} //#next v
						dp = null;
					} //#end if
				} //#next p
				//# Fix the pointPosMap using generated dp. 
				//# NOTE: this is a minor hack. See above the same code. This is here because of the breaks.
				if (dp && dp != [0, 0]) {
					//#debug("Fix ppm 2 using", dp)
					var v_i;
					var pointPosMapKeys = Object.keys(pointPosMap);
					for(v_i = 0; v_i < pointPosMapKeys.length; v_i++) {
						var v = pointPosMapKeys[v_i];
						pointPosMap[v][0] += dp[0];
						pointPosMap[v][1] += dp[1];
					} //#next v
				} //#end if
				//#debug(pointPosMap)
			} else {
				//# Deal with other things later
				//pass
			} //#end if
		} //#next i
		if (knownPos.indexOf(false) < 0) {
			break;
		} //#end if
	} //#next roundTry
	//debug((roundTry + 1) + " rounds");
	if (knownPos.indexOf(false) >= 0) {
		//debug(JSON.stringify(knownPos));
		//debug([this.points[i].name for i in range(len(knownPos)) if not knownPos[i]])
		//# TODO: Try to put the unset nodes in slots that are empty near their connections.
		//for i in range(len(knownPos)):
		for (i = 0; i < knownPos.length; i++) {
			if (knownPos[i]) { continue; }
			var foundPos = false;
			// for y in range(len(grid)):
			var y, x;
			for (y = 0; y < grid.length; y++) {
				//for x in range(len(grid[0])):
				for (x = 0; x < grid[0].length; x++) {
					if (grid[y][x] == null) {
						grid[y][x] = this.points[i];
						pointPosMap[i] = [x, y];
						//#debug("set %s to %r" % (map1.points[i].name, (x, y)))
						foundPos = true;
						break;
					} //#end if
				} //#next x
				if (foundPos) { break; }
			} //#next y
			if (!foundPos) {
				//# We need to extend, which is almost never. euler's identity repros this.
				grid.push(new Array(grid_cols).fill(null));
				grid_rows++;
				grid[grid.length - 1][0] = this.points[i];
				pointPosMap[i] = [grid_rows - 1, 0];
			} //#end if
		} //#next i
	} //#end if
	
	if (rearrange) {
		//# Improve the potential dramatically at a very high cost.
		//#for i in range(10):
		//#	if (not rearrangeGrid(map1, grid, pointPosMap, 40)) {
		//#		break
		//#	} //#end if
		//#next i
		this.rearrangeGrid(grid, pointPosMap, 40);
	} //#end if
	//# Then set the points' positions.
	gridToPos(grid, width);
	this.centerBound();
	this.prettying = false;
} //#end def sortGrid(width=40, rearrange=True)

/**
 * Map.sortGridNoRearrange(width=40)
 * Equivalent to sortGrid(width, False)
 * Faster because it doesn't rearrange Nodes, which is expensive.
 * We only need this function so we can do getattr(userdata) in sortOrder
 * 
 **/
Map.prototype.sortGridNoRearrange = function(width) {
	if (width == undefined) { width = 40; }
	return this.sortGrid(width, false);
} //#end def sortGridNoRearrange(width)

/**
 * Map.rearrangeGrid(grid, pointPosMap, width=40, nonConns=None)
 * Reduce potential using algorithm designed last night.
TODO: auto loop sort of groups on the grid
Let's solve grid1_128331410_1009.json
Repro:
python3 grid1.py -n 16 -c 2 -b 0 -x 3033139000
Returns True on improvement, False when there is no improvement.
You do not need to run gridToPos before or after rearrangeGrid, but if you 
don't and rearrangeGrid returns False, then you will have map1 back as it 
was, thus the code at the bottom of sortGrid is _correct_ so long as this 
is true.
The only bug to be fixed for this graph is adding another row when necessary and putting a node there.
This algorithm doesn't attempt that.
FIXME: This is slow even with caching. We must do whatever we can to 
improve the speed of this as well as deciding not to use it.
 **/
Map.prototype.rearrangeGrid = function(grid, pointPosMap, width, nonConns) {
	if (width === undefined) { width = 40; }
	if (nonConns === undefined) { nonConns = null; }
	//# TODO: a strange circular dependency, not good, eventually just incorporate.
	//global V_Harmonic, V_Harmonic_Angle
	//from sww_optimize1 import LJ_Potential, V_SmallWideWorld, V_Harmonic, V_Harmonic_Angle, getNonConns
	var b0 = 40;
	var lj_conf = new LJ_Potential();
	gridToPos(grid, width);
	var x0C = this.getX0Conns(true);
	var x0 = x0C[0];
	var conns = x0C[1];
	if (nonConns === null) {
		nonConns = getNonConns(conns, true);
	} //#end if
	var V_init = V_SmallWideWorld(x0, conns, b0, lj_conf, nonConns);
	//debug("V_init: " + V_init)
	var nodes = this.points.length;
	var minV = V_init;
	var x_out = this.getX0Conns()[0];
	var best_x = x_out;
	var ret = false;
	
	var roundTry;
	for (roundTry = 0; roundTry < 10; roundTry++) {
		madeImprovement = false;
		var i;
		for (i = 0; i < nodes; i++) {
			//# The cache is made up of the potential minus this node's contribution.
			// V_cache = PotentialCache(0, i)
			var V_cache = {PotentialA: 0, variant: i};
			var V_node_i = V_SmallWideWorld_Cache(x_out, conns, b0, lj_conf, nonConns, V_cache);
			V_cache.PotentialA = minV - V_node_i;
			var pos_i = pointPosMap[i];
			//for conn in self.points[i].conns:
			var k;
			for(k = 0; k < this.points[i].conns.length; k++) {
				var conn = this.points[i].conns[k].b;
				if(conn == this.points[i]) {
					conn = this.points[i].conns[k].a;
				}
				var j = this.points.indexOf(conn);
				var pos_j = pointPosMap[j];
				var dx = pos_i[0] - pos_j[0];
				var dy = pos_i[1] - pos_j[1];
				var dr_sq = (dx * dx) + (dy * dy);
				if (dr_sq <= 1) {
					//# Don't worry about this bond. It's fine.
					continue;
				} //#end if
				//# Get a list of positions we can pick for this one...
				var possiblePos = getOptionsList(pos_i, pos_j);
				//debug(JSON.stringify(pos_i) + ", " + JSON.stringify(pos_j) + ", " + JSON.stringify(possiblePos));
				
				var best_pos = null;
				//for pos in possiblePos:
				var m;
				for(m = 0; m < possiblePos.length; m++) {
					var pos = possiblePos[m];
                                        if ((pos[1] >= grid.length) || (pos[0] >= grid[0].length)) continue;
                                        if ((pos[1] < 0) || (pos[0] < 0)) continue;
					//# Don't put two in the same place.
					if (grid[pos[1]][pos[0]] != null) { continue; }
					x_out[2 * i] = width * pos[0];
					x_out[2 * i + 1] = width * pos[1];
					//#V_i = V_SmallWideWorld(x_out, conns, b0, lj_conf, nonConns)
					var V_i = V_SmallWideWorld_Cache(x_out, conns, b0, lj_conf, nonConns, V_cache);
					//debug("Potential: " + V_i);
					if (V_i < minV) {
						//#debug("Improved V_i: %f when %s moved to %r" % (V_i, map1.points[i].name, pos))
						minV = V_i;
						best_x = Array.from(x_out);
						best_pos = pos;
					} //#end if
				} //#next pos
				if (best_pos !== null) {
					//# Change the map
					var mat = reshape(Array.from(best_x), this.points.length, 2);
					this.setPos(mat);
					//# Change the grid
					//#debug("name: %s pos %r" % (map1.points[i].name, pointPosMap[i]))
					//#debug('removing %s from grid pos %r' % (grid[pointPosMap[i][1]][pointPosMap[i][0]].name, pointPosMap[i]))
					grid[pointPosMap[i][1]][pointPosMap[i][0]] = null;
					pointPosMap[i] = best_pos;
					grid[best_pos[1]][best_pos[0]] = this.points[i];
					//#debug('added %s from grid pos %r' % (grid[best_pos[1]][best_pos[0]].name, best_pos))
					ret = true;
					madeImprovement = true;
				} else {
					//# Reset x_out
					x_out = best_x;
				} //#end if
			} //#next conn
		} //#next i
		//#debug("mi", madeImprovement)
		if (madeImprovement == false) { break; }
	} //#next roundTry
	if (ret) { debug("Improvement: " + (V_init - minV)); }
	return ret;
} //#end def rearrangeGrid(grid, pointPosMap, width=40, nonConns=None)


/*
 * Map.sortFruchtermanReingold(k, t)
 * Given a position, attempt to improve the position.
 * Use strange type of springs to decide the position of each point.
 * Fruchterman, T. M. J., & Reingold, E. M. (1991). Graph Drawing by Force-Directed Placement. Software: Practice and Experience, 21(11).
 */
Map.prototype.sortFruchtermanReingold = function(k, t) {
	// Deal with optional parameters.
	if(k === undefined) { k = +24; }
	if(t === undefined) { t = +k * +0.15; }
	var i;
	var j;
	var v_disp = [];
	//debug("k: " + k);
	// Avoid divide by zero.
	if (+k < 1e-20) k = 1e-20;
	var k_sq = +k * +k;
	var inv_k = +1 / +k;
	var delta;
	// Calculate repulsive forces
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		var v_disp_i = new Vector2(+0, +0);
		for (j = 0; j < this.points.length; j++) {
			if (i == j) continue;
			// delta is shorthand for the difference.
			// vector between the positions of the two vertices.
			delta = point.sub(this.points[j]);
			// TODO: inverse sqrt.
			var delta_len_sq = delta.lensq();
			// If they are too close, limit the force to avoid divide by zero error.
			if (+delta_len_sq < +1e-20) delta_len_sq = +1e-20;
			var inv_delta_len_sq = +1 / +delta_len_sq;
			v_disp_i = v_disp_i.add(delta.mul(+inv_delta_len_sq * +k_sq));
		} //#next j
		v_disp.push(v_disp_i);
	} //#next point
	//debug('repulsion: ' + v_disp.map(function(x) { return '(' + x.x + ', ' + x.y + ')'; }).join(', '));
	
	// Calculate attractive forces
	for (i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		for(j = 0; j < point.conns.length; j++) {
			var conn = point.conns[j].b;
			if (conn == point) {
				conn = point.conns[j].a;
			}
			// NOTE: I could use getX0Conns instead if it's faster.
			var m = this.points.indexOf(conn);
			// Don't do both directions. Each edge should be counted once.
			if (i >= m) continue;
			//debug(point.name + " conn " + this.points[m]);
			delta = point.sub(conn);
			var delta_len = delta.length();
			// (delta / delta_len) * (delta_len * delta_len / k) == delta * delta_len / k
			var dr = delta.mul(delta_len * inv_k);
			v_disp[i] = v_disp[i].sub(dr);
			v_disp[m] = v_disp[m].add(dr);
		}
	}
	//debug('after attraction: ' + v_disp.map(function(x) { return '(' + x.x + ', ' + x.y + ')'; }).join(', '));
	
	// Limit the maximum displacement to the temperature t
	// and then prevent from being displaced outside frame.
	for (i = 0; i < this.points.length; i++) {
		var v_disp_i_len = v_disp[i].length();
		// This is silly but allows t to reduce the maximum displacement over iterations.
		this.points[i].setPos(this.points[i].add(v_disp[i].mul(Math.min(v_disp_i_len, t) / v_disp_i_len)));
	}
	
	// Reduce the temperature as the layout approaches a better configuration.
	// When calling sortFruchtermanReingold, reduce the t variable over iterations.
	this.prettying = false;
}//#end Map.prototype.sortFruchtermanReingold

/*
 * findRoute(p, q, avoid=None, level=0)
	Given two connected Points, find a route not directly connecting the two.
	Returns a list of Points or null in case of failure.
	A* would rock, but I'm just a hacker, not a computer scientist. j/k
*/
function findRoute(p, q, avoid, level) {
	if (avoid === undefined) { avoid = null; }
	if (level === undefined) { level = 0; }
	if (avoid === null) {
		avoid = [p];
	} else {
		avoid = [p].concat(avoid);
	} //#end if
	var i;
	var foundRoute = null;
	//for conn in p.conns:
	for(i = 0; i < p.conns.length; i++) {
		var conn = p.conns[i];
		if (conn.a == p) {
			conn = conn.b;
		} else {
			conn = conn.a;
		}
		//# We don't want to go backwards, infinite recursion.
		//if (conn in avoid) continue;
		if (avoid.indexOf(conn) >= 0) continue;
		if (conn != q) {
			var v = findRoute(conn, q, avoid, level + 1);
			if (v) { 
				//#print('found route', ', '.join([str(x) for x in v]))
				foundRoute = [p].concat(v);
				return foundRoute;
			} //#end if
		} else {
			//# We got a path, but we might not want it.
			//# Check if we just started.
			if (level > 0) {
				return [p, q];
			} else if (foundRoute === null) {
				//# TODO: Verify this somehow...
				foundRoute = [q];
			} //#end if
		} //#end if
	} //#next conn
	return foundRoute;
} //#end def findRoute(p, q, avoid=None)

/*
 * Map.getPolarConn(b0=40)
 * Same code as toPolar sadly can't fix that.
 * returns a list of index pairs [(i, j), ...] and an atom_name_hash object.
 */
Map.prototype.getPolarConn = function() {
	//var dr = [];
	var atom_name_hash = new Object(); //sha256();
	var atom_name_data = '';
	var i;
	var j;
	//# We need sorted connections to fix a strange bug related to export/import, so I choose name.
	//for p in this.points:
	for (i = 0; i < this.points.length; i++) {
		var p = this.points[i];
		// TEST: Need to sort by name
		// key=lambda x: x.name
		p.conns.sort(function(x, y) {
			var xc = x.a;
			if (xc == p) xc = x.b;
			var yc = y.a;
			if (yc == p) yc = y.b;
			if (xc.name < yc.name) {
				return -1;
			}
			if (xc.name > yc.name) {
				return 1;
			}
			return 0;
		});
	} //#next p
	//# Make sure they don't use the reverse like N3 -- CT, CT -- N3 because 
	//# that produces no useful data.
	var knownsNotIncremented = 0;
	var reverses = [[0, 0]];
	var known = [0];
	var unconns = [];
	for (round_n = 0; round_n < (this.points.length * 2); round_n++) {
		var knownsIncremented = false;
		//for i, point in enumerate(this.points):
		for (i = 0; i < this.points.length; i++) {
			if (known.indexOf(i) >= 0) continue;
			var point = this.points[i];
			//# This algorithm requires each molecule to order its atoms in 
			//# the same way. We hash the names to ensure that the data is 
			//# the same.
			var atom_name = point.name;
			var dr_x = null;
			var i_x = null;
			//for x in point.conns:
			for (j = 0; j < point.conns.length; j++) {
				var x = point.conns[j];
				if (x.a == point) {
					x = x.b;
				} else {
					x = x.a;
				}
				dr_x = true; //x.sub(point);
				var i_x_poss = this.points.indexOf(x);
				if (known.indexOf(i_x_poss) < 0) continue;
				i_x = i_x_poss;
				if (x.conns.length > 1) {
					if (reverses.indexOf([i_x, i]) < 0) {
						//# We found one I think...
						break;
					} //#end if
				} //#end if
			} //#next x
			if (dr_x === null) {
				//debug("Unconnected atom? " + point.conns.toString() + " " + i);
				unconns.push(point);
				continue;
			} //#end if
			if (i_x == null) continue;
			//dr.push(dr_x);
			var atom2_name = x.name;
			//atom_name_hash.update((atom_name + ('[%i]' % i) + "--" + atom2_name + ('[%i]' % i_x)).encode('utf-8'))
			atom_name_data += (atom_name + "--" + atom2_name) + '\n';
			//#reverses.append((point, x))
			reverses.push([i, i_x]);
			known.push(i);
			knownsIncremented = true;
		} //#next i, point
		//# Connect unconnect atoms in a row to the last known.
		if ((unconns.length > 0) && (known.length == 0)){
			known.push(this.points.indexOf(unconns[0]));
		}//#end if
		var unc_i;
		for(unc_i = 0; unc_i < unconns.length; unc_i++) {
			var unc = unconns[unc_i];
			i = this.points.indexOf(unc);
			i_x = known[-1];
			reverses.push([i, i_x]);
			known.push(i);
		} //#next unc
		//# We are done when all have connections.
		if (reverses.length >= this.points.length) break;
		
		//# We are in trouble if knowns didn't increase. That means two separate structures.
		if (!knownsIncremented) {
			//#print('not knownsIncremented on round', round_n)
			knownsNotIncremented += 1;
			if (knownsNotIncremented >= 2) {
				//#print('knownsNotIncremented >= 2 on round', round_n)
				for (i = 0; i < this.points.length; i++) {
					if (known.indexOf(i) < 0) {
						//# Connect it to the last known.
						reverses.push([known[known.length-1], i]);
						known.push(i);
						//# Reset the counter.
						knownsNotIncremented = 0;
						break;
					} //#end if
				} //#next i
				//#break
			} //#end if
		} else {
			knownsNotIncremented = 0;
		} //#end if
	} //#next round_n
	debug((round_n+1) + " rounds to do getPolarConn");
	//debug(JSON.stringify(reverses));
	if (reverses.length != this.points.length) {
		debug("Error: getPolarConn failed, expect failure elsewhere.")
		debug(reverses.toString());
		debug(atom_name_data);
	} //#end if
	return [reverses, atom_name_hash];
} //#end def getPolarConn()

/*
 * Map.setCartesian(dr)
 * Unlike the name suggests, we are not setting the atom's 
 * positions to a list of cartesian coordinates.
 * dr is a list of relative polar coordinates.
 */
Map.prototype.setCartesian = function(dr) {
	if (dr.length != this.points.length) {
		debug("Not right number of dr. " + dr.length + " " + this.points.length);
	}
	var i;
	var x = this.getPolarConn();
	var drs = x[0];
	var atom_hash = x[1];
	var knowns = [drs[0][0]];
	var j = 0;
	while (knowns.length < this.points.length) {
		//# i starts at 0 and goes to N like it should.
		//for i, spc in enumerate(drs):
		for (i = 0; i < drs.length; i++) {
			var spc = drs[i];
			if (knowns.indexOf(spc[0]) >= 0) {
				if (knowns.indexOf(spc[1]) >= 0) {
					//# We know both, so skip this dr...
					continue;
				}//#end if
				knowns.push(spc[1]);
				var polarCoord = polarToCartesian(dr[i]);
				this.points[spc[1]].setPos(this.points[spc[0]].add(
					new Vector2(polarCoord[0], polarCoord[1])));
			} else if (knowns.indexOf(spc[1]) >= 0) {
				//# spc[0] not in knowns
				knowns.push(spc[0]);
				var polarCoord = polarToCartesian(dr[i]);
				this.points[spc[0]].setPos(this.points[spc[1]].add(
					new Vector2(polarCoord[0], polarCoord[1])));
			}
		} //#next i, spc
		j++;
		if (j > 1000) {
			debug("rounds has passed 1000, we give up. " + knowns.length + " " + this.points.length + " " + knowns.toString());
			debug(drs.toString());
			break;
		} //#end if
	} //#loop
}

/*
 * Map.setPos(v)
	Take an x0 properly reshaped and load into points.
	Example:
	m = np.matrixlib.matrix(x0)
	m = m.reshape(len(x0)//2, 2)
	map1.setPos([np.array(xa)[0] for xa in m])
 */
Map.prototype.setPos = function(v) {
	var i;
	for (i = 0; i < v.length; i++) {
		var xa = v[i];
		this.points[i].x = xa[0];
		this.points[i].y = xa[1];
	} //#next i, xa
} //#end def setPos(v)

/*
 * Map.getX0Conns()
	Returns (X0, conns)
	x0: the 1d array of point [x, y, x, y ...] and the 
	conns: the 2d array of connections [[1], [0, 2], ...]
	x0 is a numpy array, conns is a Python array.
	This is used in sww_optimize1, sortLoop and other things.
 */
Map.prototype.getX0Conns = function(i_lt_j) {
	if (i_lt_j === undefined) { i_lt_j = false; }
	//# Get the xy values into a 1D array.
	var i;
	var x0 = [];
	for (i = 0; i < this.points.length; i++) {
		x0.push(this.points[i].x);
		x0.push(this.points[i].y);
	}
	var conns = [];
	//for p in this.points:
	for (i = 0; i < this.points.length; i++) {
		var p = this.points[i];
		var pConns = [];
		//for conn in p.conns:
		for(j = 0; j < p.conns.length; j++) {
			var q = p.conns[j];
			if (q.a == p) {
				q = q.b;
			} else {
				q = q.a;
			}
			pConns.push(this.points.indexOf(q));
		} //#next conn
		// Ensure that we only have i < j
		if (i_lt_j) {
			pConns = pConns.filter(function (j) { return i < j; });
		}
		conns.push(pConns);
	} //#next p
	return [x0, conns];
} //#end def getX0Conns()

/*
 * Map.squash()
 * A very strange algorithm, not sure why it was written.
 * FIXME: Broken
 */
Map.prototype.squash = function() {
	var mean = new Vector2(0, 0);
	for(i = 0; i < this.points.length; i++) {
		//#mean[0] += point.x
		//#mean[1] += point.y
		mean = mean.add(point);
	} //#next point
	//#mean[0] /= (1.0 * this.points.length)
	//#mean[1] /= (1.0 * this.points.length)
	mean = mean.div(1.0 * this.points.length);
	//#goto = [this.width/2.0, this.height/2.0]
	var goto = new Vector2(this.width/2.0, this.height/2.0);
	//#diff = [goto[0]-mean[0], goto[1]-mean[1]]
	var diff = goto.sub(mean);
	var i;
	var j;
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		//#point.x += diff[0]
		//#point.y += diff[1]
		for(j = 0; j < p.conns.length; j++) {
			var conn = p.conns[i].b;
			if (conn == point) {
				conn = p.conns[i].a;
			}
			point.x += diff.x;
			point.y += diff.y;
		} //#next conn
	} //#next point
}; //#end Map.prototype.squash

/*
 * Map.toString()
 * Returns an SVG of the Map.
 */
Map.prototype.toString = function() {
	return this.drawSvg();
}; //#end Map.prototype.__str__

/*
 * Map.drawSvg()
 * Returns an SVG of the Map.
 */
Map.prototype.drawSvg = function() {
	var output = '<?xml version="1.0" standalone="no"?>\n<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN" "http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">';
	output += '\n<svg width="' + str(this.width) + '" height="' + str(this.height) + '" viewBox="0 0 ' + str(this.width) + ' ' + str(this.height) + '" xmlns="http://www.w3.org/2000/svg" version="1.1" xmlns:xlink="http://www.w3.org/1999/xlink">\n';
	output += '<title>' + this.title + '</title>\n';
	output += '<desc>' + this.desc + '</desc>\n';
	output += '<rect x="1" y="1" width="' + str(this.width-2) + '" height="' + str(this.height-2) + '" fill="none" stroke="blue" />\n';
	
	var i;
	var j;
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		var tpos = point.add(new Vector2(8, 8));
		//# Draw each point
		//#output += "<!--<text x='" + ('%3.3f' % tpos.x) + "' y='" + ('%3.3f' % tpos.y) + "'>" + point.name + "</text>-->\n"
		output += "<circle cx='" + (point.x.toFixed(3)) + "' cy='" + (point.y.toFixed(3)) + "' r='" + ('%3.1f' % point.size) + "' fill='" + point.color + "' stroke='grey' stroke-width='1' />\n";
		//# Draw each conn
		for (j = 0; j < point.conns.length; j++) {
			var conn = point.conns[j].b;
			if (conn == point) conn = point.conns[j].a;
			//#output += '<!-- ' + str(point) + ' to ' + str(conn) + '-->\n'
			output += "<path d='M " + (point.x.toFixed(3)) + " " + (point.y.toFixed(3)) + " L " + (conn.x.toFixed(3)) + " " + (conn.y.toFixed(3)) + "' fill='none' stroke='" + str(this.line_color) + "' stroke-width='" + str(this.line_width) + "' />\n";
		} //#next conn
	} //#next point
	//# A layer of text above the connections.
	output += '<g>';
	for(i = 0; i < this.points.length; i++) {
		var tpos = point.add(new Vector2(8, 8));
		if (point.url != '') {
			output += "<a xlink:href='" + point.url.replace('&', '&amp;') + "'>";
		}
		output += "<text x='" + (tpos.x.toFixed(3)) + "' y='" + (tpos.y.toFixed(3)) + "' font-size='11'>" + point.name.replace('&', '&amp;') + "</text>\n";
		if (point.url != '') {
			output += "</a>\n";
		}
	}//#next point
	output += '</g>';
	output += '</svg>';
	return output;
}; //#end Map.prototype.drawSvg

/*
 * Map.saveJson()
 * Returns a JSON string of the Map.
 * To load, use loadJson(JSON.parse(s))
 */
Map.prototype.saveJson = function() {
	var output = '{"nodes":[';
	var i;
	var conns = [];
	for(i = 0; i < this.points.length; i++)
	{
		output += JSON.stringify([this.points[i].name, Number(this.points[i].x.toFixed(2)), Number(this.points[i].y.toFixed(2))]) + ', ';
		var j;
		for(j = 0; j < this.points[i].conns.length; j++)
		{
			conns.push(this.points[i].conns[j]);
		}
	}
	// Strip comma.
	if(this.points.length > 0)
	{
		output = output.substr(0, output.length - 2);
	}
	// Add end bracket, add conns attribute.
	output += '], "conns":[';
	for(j = 0; j < conns.length; j++)
	{
		if(this.points.indexOf(conns[j].a) > this.points.indexOf(conns[j].b)) continue;
		output += JSON.stringify([conns[j].a.name, conns[j].b.name]) + ', ';
	}
	// Strip comma, add end bracket.
	if(conns.length > 0)
	{
		output = output.substr(0, output.length - 2);
	}
	output += ']}';
	return output;
}

/*
 * Map.loadJson(jsobj)
 * Loads points and connections from a JSON object.
 * Object jsobj looks like: 
 * {"nodes":[["node 0", 233, 224], ["node 1", 407, 118]], "conns":[["node 0", "node 1"]]}
 * To generate one from a Map, use JSON.parse(map.saveJson())
 */
Map.prototype.loadJson = function(jsobj) {
	// Add the nodes.
	var i;
	for(i = 0; i < jsobj.nodes.length; i++)
	{
		this.addPoint(new Point2(jsobj.nodes[i][0], null, jsobj.nodes[i][1], jsobj.nodes[i][2]));
	}
	// Connect the connections.
	var j;
	var connObjs = [];
	for(i = 0; i < jsobj.conns.length; i++)
	{
		// We can resolve names now, so do it.
		connObjs.push(new Conn(jsobj.conns[i][0], jsobj.conns[i][1]));
		connObjs[i].resolve(this);
		for(j = 0; j < this.points.length; j++)
		{
			if(this.points[j] == connObjs[i].a)
			{
				this.points[j].addConn(connObjs[i].b);
				break;
			}
		}
	}
}

/*
 * Map.saveDot()
 * Returns a DOT string of the Map.
 * To convert to SVG with Graphviz, run dot -otest1.svg -Tsvg test1.dot
 */
Map.prototype.saveDot = function() {
	var output = 'graph {\n';
	var i;
	for(i = 0; i < this.points.length; i++)
	{
		var uniqueConns = 0;
		var connStr = '';
		var j;
		for(j = 0; j < this.points[i].conns.length; j++)
		{
			var k = this.points.indexOf(this.points[i].conns[j].b);
			if(i >= k) continue;
			uniqueConns++;
			connStr += JSON.stringify(this.points[i].conns[j].b.name) + " ";
		}
		var printed = false;
		if((uniqueConns > 0) || (this.points[i].conns.length == 0))
		{
			output += JSON.stringify(this.points[i].name);
			printed = true;
		}
		if(uniqueConns == 0)
		{
			if(printed) output += '\n';
			continue;
		}
		output += ' -- ';
		if(uniqueConns > 1)
		{
			output += '{';
		}
		output += connStr.substr(0, connStr.length - 1);
		if(uniqueConns > 1)
		{
			output += '}';
		}
		output += '\n';
	}
	for(i = 0; i < this.points.length; i++)
	{
		output += JSON.stringify(this.points[i].name) + '[ pos = "' + this.points[i].x + ',' + this.points[i].y + '!" ]\n';
	}
	output += '}';
	return output;
}

function shuffle(o) {
	for(var j, x, i = o.length; i; j = Math.floor(Math.random() * i), x = o[--i], o[i] = o[j], o[j] = x);
	return o;
}

function adder(previousValue, currentValue, currentIndex, array) {
	return previousValue + currentValue;
}

/**
 * Map.sortOrder(sortFunc = None, n = 100)
 * 
 **/
Map.prototype.sortOrder = function(sortFunc, n) {
	if (sortFunc == undefined || sortFunc == null) {
		sortFunc = 'sortGridNoRearrange';
	} //#end if
	if (n == undefined) { n = 100; }
	
	var checkPast = false;
	var currentV = 0;
	if (this.prettying) {
		// Save this value so that if it's good, we can go back to it.
		// This makes clicking Pretty multiple times in your best interest.
		var currentX0 = this.getX0Conns(true)[0];
		currentV = this.prettyV;
		checkPast = true;
		var currentPoints = this.points.slice();
	}
	
	var lj_conf = new LJ_Potential();
	var b0 = 40;
	var gtol = 1e-2;
	var rn = randint(0, 0xffffffff);
	var funcName = sortFunc.replace('sort', '');
	var start_t = Date.now();
	var V_all = [];
	var V_min = 1e36;
	var x_min = null;
	var points_min = this.points.slice(); // Copy array
	var i;
	for(i = 0; i < n; i++) {
		//# Use a different seed to get different maps. Deterministic.
		//seed = randint(0, 0xffffffff);
		//# If we don't do createMap4, the seed will be different, expensive but not as expensive as sorting/optimizing by a long shot.
		shuffle(this.points);
		//# The shuffle will cause these to go bad, so we have to recompute them
		var x0C = this.getX0Conns(true);
		var x0 = x0C[0];
		var conns = x0C[1];
		var nonConns = getNonConns(conns, true);
		//#print(x0)
		//#if verbose:
		//#	print("atom conns:", conns)
		//#end if

		//# Not necessary to get spherical coordinates of the original coordinates.
		//#output = [spherical2.cartesianToSpherical(v) for v in molecule.x0]
		//#print("Spherical coordinates of atoms:")
		//#print(output)
		
		//# This may be used eventually
		//#sphericalCoords, atom_name_hash = molecule.toSpherical(verbose)
		this[sortFunc](b0);
		var x_out = this.getX0Conns()[0];
		var V_configuration = V_SmallWideWorld(x_out, conns, b0, lj_conf, nonConns);
		//#print(funcName, "Potential:", V_configuration)
		V_all.push(V_configuration);
		if (V_configuration < V_min) {
			V_min = V_configuration;
			x_min = x_out.slice(); // Copy array, dunno why this is necessary.
			points_min = this.points.slice(); // Copy array
		} //#end if
		var sort_dt = (Date.now() - start_t);
		// Don't go past 9 seconds so that browsers won't break.
		if (sort_dt/1000. > 9) break;
	} //#next i
	debug(funcName + " dt: " + (sort_dt/1000.).toFixed(3) + " seconds, " + i + " iterations, Prev V:" + currentV.toFixed(3));
	debug(funcName + " Potential: min: " + V_min.toFixed(3) + " avg: " + ((V_all.reduce(adder))/(V_all.length)).toFixed(3) + " max: " + Math.max.apply(null, V_all).toFixed(3))
	if (x_min != null) {
		if((checkPast == false) || (V_min < currentV)) {
			//# FIXME: This makes my head spin, it can't actually work.
			this.points = points_min;
			var mat = reshape(Array.from(x_min), this.points.length, 2);
			this.setPos(mat);
			this.prettyV = V_min;
		} else {
			this.points = currentPoints;
			var mat = reshape(Array.from(currentX0), this.points.length, 2);
			this.setPos(mat);
		}
		//map1.setPos([np.array(xa)[0] for xa in m])
		this.centerBound();
	} //#end if
	this.prettying = true;
	return V_all;
} //#end def sortOrder(self, sortFunc = None, n = 100)

//#end class Map

/*
 * Vector2(x, y)
 * 2D Vector implementation in JavaScript.
 * JavaScript lacks operator override, so we have methods instead.
 */
function Vector2(x, y)
{
	if(x === undefined) { x = 380.0; }
	if(y === undefined) { x = 180.0; }
	this.x = new Number(x);
	this.y = new Number(y);
}
new Vector2();

Vector2.prototype.toString = function() {
	return (this.x.toFixed(3)) + ', ' + (this.y.toFixed(3))
}; //#end Vector2.prototype.toString

Vector2.prototype.add = function(vec) {
	return new Vector2(this.x + vec.x, this.y + vec.y);
};
Vector2.prototype.sub = function(vec) {
	return new Vector2(this.x - vec.x, this.y - vec.y)
};
Vector2.prototype.mul = function(flt) {
	return new Vector2(this.x * flt, this.y * flt);
};

Vector2.prototype.div = function(flt) {
	//#if (flt == 0) {
	//#	return Vector2(0, 0)
	//debug("(" + this.x + ", " + this.y + ") / " + flt);
	return new Vector2(this.x / flt, this.y / flt);
};

/*
 * Vector2.dot(other)
 * Dot product of two 2D vectors.
 */
Vector2.prototype.dot = function(other) {
	return (this.x * other.x) + (this.y * other.y);
};

//#Vector2.prototype.__iadd__ = function(vec) {:
//#	v = new Vector2(this.x + vec.x, this.y + vec.y)
//#	return v

//#Vector2.prototype.__idiv__ = function(flt) {:
//#	v = new Vector2(this.x / flt, this.y / flt)
//#	return v

Vector2.prototype.setPos = function(v) {
	this.x = v.x;
	this.y = v.y;
};

Vector2.prototype.length = function() {
	return Math.sqrt(this.x * this.x + this.y * this.y);
};

Vector2.prototype.lensq = function() {
	return (this.x * this.x + this.y * this.y);
};
//#end class Vector2

// Point2 Subclasses Vector2
function Point2(name, conns, x, y, color, url, size)
{
	// Deal with a bunch of optional params.
	if(x === undefined) { x = 380.0; }
	if(y === undefined) { y = 180.0; }
	if(color === undefined) { color = '#ccc'; }
	if(url === undefined) { url = ''; }
	if(size === undefined) { size = 4; }
	Vector2.call(this, x, y);
	//Vector2.__init__(x, y)
	this.name = name;
	this.url = url;
	this.conns = conns;
	this.color = color;
	this.size = size;
	this.visual = null;
	this.text = null;
	if (conns === null) this.conns = [];
}
Point2.prototype = Object.create(Vector2.prototype);
new Point2();

Point2.prototype.toString = function() {
	return this.name.toString() + ' ' + (this.x.toFixed(3)) + ', ' + (this.y.toFixed(3));
}; //#end Point2.prototype.__str__ = function() {

/*
 * Point2.addConn(conn, reConn, connObj)
 * Different from mapmake1.py due to the Conn object.
 * reConn is optional like in Python.
 * connObj is optional and should be left empty (undefined) so that the 
 * reconnection works.
 */
Point2.prototype.addConn = function(conn, reConn, connObj) {
	if(reConn === undefined) { reConn = true; }
	// Already connected, don't add duplicates.
	if(checkPointInConns(conn, this.conns)) { return; }
	var ourConnObj = new Conn(this, conn);
	if(connObj != undefined){
		// Connect the visuals so we don't draw twice.
		//ourConnObj.visualc = connObj.visualc;
	}
	//#print 'adding conn?', this.name, conn
	this.conns.push(ourConnObj);
	// Give them the same object, so when it renders, it doesn't render two 
	// objects, it renders the same object twice if you get what I mean.
	if(reConn) conn.addConn(this, false, ourConnObj);
};
Point2.prototype.setColor = function(value) {
	/* Set color to a hex value like //#ff0000 */
	this.color = value;
};

/*
 * Point2::draw() -- update the screen.
 * Only available in JavaScript for obvious reasons.
 */
Point2.prototype.draw = function(display)
{
	// Draw it.
	if(this.visual === null)
	{
		var svg = 'http://www.w3.org/2000/svg';
		this.visual = document.createElementNS(svg, 'circle');
		this.visual.setAttributeNS(null, 'stroke', '#999');
		this.visual.setAttributeNS(null, 'stroke-width', '2');
		this.visual.setAttributeNS(null, 'stroke-opacity', '1');
		this.visual.setAttributeNS(null, 'fill', this.color);
		this.visual.onmousedown = function() { return false; }
		this.text = document.createElementNS(svg, 'text');
		this.text.setAttributeNS(null, 'font-family', 'Droid Sans Mono;monospace');
		this.text.setAttributeNS(null, 'font-size', '12');
		this.text.setAttributeNS(null, 'style', "fill:#000");
		// TODO: make this editable.
		//this.text.contentEditable = true;
		this.text.onmousedown = function() { return false; }
		this.text.innerHTML = htmlentities(this.name.toString());
		display.appendChild(this.visual);
		display.appendChild(this.text);
	}

	this.visual.setAttributeNS(null,'cx', this.x);
	this.visual.setAttributeNS(null,'cy', this.y);
	this.visual.setAttributeNS(null,'r', 10);
	this.text.setAttributeNS(null,'x', this.x+14);
	this.text.setAttributeNS(null,'y', this.y+14);

	var i;
	for(i = 0; i < this.conns.length; i++)
	{
		this.conns[i].draw(display);
	}
}
//#end class Point2(Vector2)

/*
 * Class Conn
 * Only exists in JavaScript because of the design of JSON.
 * 
 */
function Conn(a, b)
{
	this.a = a;
	this.b = b;

	// Line which we draw.
	this.visual = null;
	this.visualc = {'visual':this.visual};
}
// Turn this into a class.
new Conn();

Conn.prototype.resolve = function(map)
{
	if(typeof this.a == "string")
	{
		this.a = map.getNamed(this.a);
		this.b = map.getNamed(this.b);
	}
}

/*
 * Conn::draw() -- update the screen.
 * 
 */
Conn.prototype.draw = function(display)
{
	// Draw it.
	if(this.visual === null)
	{
		// Don't add a line if our complement has already done it.
		if(this.visualc.visual !== null)
		{
			this.visual = this.visualc.visual;
		}
		else
		{
			var svg = 'http://www.w3.org/2000/svg';
			this.visual = document.createElementNS(svg, 'line');
			this.visual.setAttributeNS(null, 'stroke', '#000');
			this.visual.setAttributeNS(null, 'stroke-width', '1');
			this.visual.setAttributeNS(null, 'stroke-opacity', '1');
			this.visual.onmousedown = function() { return false; }
			display.appendChild(this.visual);
			// Connect the complementary's connection to this.
			this.visualc.visual = this.visual;
		}
	}
	// TODO: This shouldn't fire when this.visual == this.visualc.visual...
	this.visual.setAttributeNS(null, 'x1', this.a.x);
	this.visual.setAttributeNS(null, 'y1', this.a.y);
	this.visual.setAttributeNS(null, 'x2', this.b.x);
	this.visual.setAttributeNS(null, 'y2', this.b.y);
}
//#end class Conn

//class MapGroup:
function MapGroup() {
	this.points = [];
	this.left_bracket = null;
	this.right_bracket = null;
	this.finished = false;
}
// Turn this into a class.
new MapGroup();

/*
 * MapGroup.leaf(p)
	Make this group a single node leaf.
 */
MapGroup.prototype.leaf = function(p) {
	this.finished = true;
	this.points = [p];
	this.left_bracket = null;
	this.right_bracket = null;
} //#end def leaf(p)

/*
 * MapGroup.leaf(p)
	Append a point to points.
 */
MapGroup.prototype.addPoint = function(p) {
	this.points.push(p);
} //#end def addPoint(p)

/*
 * MapGroup.cull()
	Remove any nodes that aren't part of a loop.
	For example: a-b-c-d-e-f-g, d-a. Remove e, f, and g.
	Return culled nodes. Remove nodes from self.points.
	FIXME: It doesn't work when a whole group is consumed by another group.
 */
MapGroup.prototype.cull = function() {
	//# Obviously unnecessary to cull groups of 0, 1, or 2.
	if (this.points.length < 3) return [];
	//# TODO: Build a cache so that we can use it to be quicker.
	var cache = [];
	var to_cull = [];
	var foundCull = true;
	while (foundCull) {
		foundCull = false;
		//for p in this.points:
		for (i = 0; i < this.points.length; i++) {
			var p = this.points[i];
			var conns_in_G = 0;
			//for q in p.conns:
			for(j = 0; j < p.conns.length; j++) {
				var q = p.conns[j];
				if (q.a == p) {
					q = q.b;
				} else {
					q = q.a;
				}
				if (this.points.indexOf(q) >= 0) {
					conns_in_G++;
				} //#end if
			} //#next q
			if (conns_in_G < 2) {
				to_cull.push(p);
				foundCull = true;
				break;
			}//#end if
		} //#next p
		if (foundCull) {
			//this.points.remove(to_cull[-1]);
			this.points.splice(this.points.indexOf(to_cull[to_cull.length-1]), 1);
		} //#end if
	} //#loop
	return to_cull;
} //#end def cull()

/*
 * MapGroup.minX()
	A simple method to find the x value of the minimum x value among points.
	This allows necessary sorting.
	Example:
	groups.sort(key=lambda g:g.minX())
*/
MapGroup.prototype.minX = function() {
	//return min([p.x for p in self.points]);
	return this.points.reduce(function(previousValue, currentValue, currentIndex, array) {
		return Math.min(previousValue, currentValue);
	});
} //#end def minX()

//#end class MapGroup

function organize(map)
{
	var f = 0.71;
	var scale = (map.width + map.height) / 2.0;
	//map.randomize(Math.floor(scale / (0.65*f)));
	for(i = 0; i < 12; i++) {
		map.sort2(scale / (7.5*f), 0.7);
		if(i % 30 == 0) {
			map.sort3();
		} //#end if
		if(i % 32 == 0) {
			map.sort4();
		} //#end if
		if (i % 25 == 0) {
			map.randomize(Math.floor(scale / (10.0*f)));
		} //#end if
		//#a.sort1()
		//#a.sort1()
		//#a.sort1()
		//#a.sort1()
	/*
	*/
	} //#next i
	for(i = 0; i < 12; i++) {
		map.sort4();
	} //#next i
	//#m.sort3()
	map.center();
}

function createMap(seed_id) {
	if(seed_id === undefined) { seed_id = 3243; }
	// FIXME: No random seed in JavaScript, have to implement your own RNG. 
	// No, really.
	//Math.seed(seed_id);
	var a = new Map();
	var b = [];
	var i;
	for(i = 0; i < 8; i++) {
		b.push(new Point(String.fromCharCode(String.charCodeAt('b') + i)))
	} //#next i
	for(i = 0; i < 7; i++) {
		b[i].addConn(b[i+1]);
	} //#next i
	var addConn1 = randint(0,b.length-1);
	var addConn2 = (addConn1 + randint(0,b.length)) % b.length;
	if (addConn1 == addConn2) {
		//# Possible
		addConn2 = (addConn1 ^ 7) % b.length;
	}
	if (addConn1 == addConn2) {
		//# Unlikely
		addConn1 = 0;
	}
	if (addConn1 == addConn2) {
		//# Impossible
		addConn2 = 1;
	}
	//#print b, addConn1, addConn2
	b[addConn1].addConn(b[addConn2]);
	a.addPoints(b);
	for(i = 0; i < 4; i++) {
		a.randomize();
		a.sort1();
		//#a.sort1()
		//#a.sort1()
		//#a.sort1()
		//#a.sort1()
	} //#next i
	return a;
} //#end def createMap([seed_id])

/**
* reshape(mat, m, n)
* Take a flat array and turn it into a matrix.
* TODO: Be more flexible about what input matricies are allowed.
*/
function reshape(mat, m, n)
{
	var newMat = [];
	while(mat.length) { newMat.push(mat.splice(0, n)); }
	return newMat;
}

/**
 * gridToPos(grid, width=40)
 * Simple function does the positioning of nodes in grid.
 * Assumes a proper grid of points.
TODO: When multiple nodes are in one cell, they should be spread out as 
much as possible.
 */
function gridToPos(grid, width) {
	if (width === undefined) { width = 40; }
	var y, x;
	//for y in range(len(grid)):
	for (y = 0; y < grid.length; y++) {
		//for x in range(len(grid[0])):
		for (x = 0; x < grid[0].length; x++) {
			if (grid[y][x]) {
				if (Array.isArray(grid[y][x])) {
					var p_i;
					//for p in grid[y][x]:
					for(p_i = 0; p_i < grid[y][x].length; p_i++) {
						grid[y][x][p_i].setPos(new Vector2(x * width, y * width));
					} //#end if
				} else {
					//#print(x, y, grid[y][x].name)
					grid[y][x].setPos(new Vector2(x * width, y * width));
				} //#end if
			} //#end if
		} //#next x
	} //#next y
} //#end def gridToPos(grid)

/**
 * V_SmallWideWorld_Cache(x, conns, b0, lj_conf, nonConns, cache)
 * A massive speed up of sww_optimize1.V_SmallWideWorld without any loss of accuracy.
	TODO: speed test.
 */
function V_SmallWideWorld_Cache(x, conns, b0, lj_conf, nonConns, cache) {
	var Potential = cache.PotentialA;
	//debug("Pot: " + Potential);
	//# Harmonic potential with length b0.
	//for i in [cache.variant]:
	var i = cache.variant;
	var pt_i = new Vector2(x[2*i], x[2*i+1]);
	var verts = [pt_i];
	var dps = [];
	var dp_lens = [];
	var k;
	//for j in conns[i]:
	for(k = 0; k < conns[i].length; k++) {
		var j = conns[i][k];
		var pt_j = new Vector2(x[2*j], x[2*j+1]);
		var dp = pt_i.sub(pt_j);
		var dp_len = dp.length();
		Potential += V_Harmonic(dp_len, b0);
		verts.push(pt_j);
		dps.push(dp);
		dp_lens.push(dp_len);
	} //#next j
	//debug("Pot2: " + Potential);
	if (verts.length > 2) {
		//# Angular Harmonic calculation
		var theta0 = tau / conns[i].length;
		//# Angular Harmonic calculation
		var theta0 = 2 * Math.PI / conns[i].length;
		//# theta is computed for axb, bxc, cxa
		var thetas = [];
		for (k = 0; k < conns[i].length; k++) {
			var ax = dps[k];
			thetas.push(Number_mod(Math.atan2(ax.y, ax.x), 2*Math.PI));
		}
		thetas.sort(function(a, b) { return a - b; });
		for (k = 0; k < thetas.length; k++) {
			var b = (k+1) % thetas.length;
			var theta = Number_mod(thetas[b] - thetas[k], 2*Math.PI);
			Potential += V_Harmonic_Angle(theta, theta0);
		} //#next i
	} //#end if
	//#next i
	// FIXME: This is an experiment to see if non-connected potentials are worthwile for grids.
	return Potential;
	/*//debug("Pot3: " + Potential);
	//# Lennard-Jones for non-connected nodes.
	//for i in [cache.variant]:
	pt_i = new Vector2(x[2*i], x[2*i+1]);
	//for j in nonConns[i]:
	for (k = 0; k < nonConns[i].length; k++) {
		var j = nonConns[i][k];
		var pt_j = new Vector2(x[2*j], x[2*j+1]);
		var dp = pt_i.sub(pt_j);
		var dp_len = dp.length();
		Potential += lj_conf.V_LJ(dp_len);
	} //#next j
	//#next i
	//debug("Pot4: " + Potential);
	return Potential;*/
} //#end def V_SmallWideWorld_Cache(x, conns, b0, lj_conf, nonConns, cache)

/**
 * getOptionsList(pos_i, pos_j, brute=False)
 * Pick spots to move i based on position of i and j.
 */
function getOptionsList(pos_i, pos_j, brute) {
	if (brute === undefined) { brute = false; }
	var ret = [];
	//# TODO: Deal with i and j being in all possible configurations dx pos, neg, dy pos, neg.
	var dx = pos_j[0] - pos_i[0];
	var dy = pos_j[1] - pos_i[1];
	var stepk = 1;
	var stepm = 1;
	var startk = 0;
	var startm = 0;
	if (dx < 0) {
		stepk = -1;
	} //#end if
	if (dy < 0) {
		stepm = -1;
		dy -= 1;
	} else {
		dy += 1;
	} //#end if
	//#print('km', startk, dx, stepk, startm, dy, stepm)
	if (brute || (Math.abs(dx * dy) <= 6)) {
		var k, m;
		if ((stepk > 0) && (stepm > 0)) {
			//for k in range(startk, dx, stepk):
			for(k = startk; k < dx; k += stepk) {
				//for m in range(startm, dy, stepm):
				for(m = startm; m < dy; m += stepm) {
					ret.push([pos_i[0] + k, pos_i[1] + m]);
				} //#next m
			} //#next k
		} else if ((stepk > 0) && (stepm < 0)) {
			//for k in range(startk, dx, stepk):
			for(k = startk; k < dx; k += stepk) {
				//for m in range(startm, dy, stepm):
				for(m = startm; m > dy; m += stepm) {
					ret.push([pos_i[0] + k, pos_i[1] + m]);
				} //#next m
			} //#next k
		} else if (stepm > 0) {
			//for k in range(startk, dx, stepk):
			for(k = startk; k > dx; k += stepk) {
				//for m in range(startm, dy, stepm):
				for(m = startm; m < dy; m += stepm) {
					ret.push([pos_i[0] + k, pos_i[1] + m]);
				} //#next m
			} //#next k
		} else {
			//for k in range(startk, dx, stepk):
			for(k = startk; k > dx; k += stepk) {
				//for m in range(startm, dy, stepm):
				for(m = startm; m > dy; m += stepm) {
					ret.push([pos_i[0] + k, pos_i[1] + m]);
				} //#next m
			} //#next k
		}
		//# Reject the first one which is pos_i
		return ret.slice(1);
	} //#end if
	if (dx != 0) {
		ret.push([pos_i[0] + stepk, pos_i[1]]);
		ret.push([pos_j[0] - stepk, pos_j[1]]);
	} //#end if
	if (dy != 0) {
		ret.push([pos_i[0], pos_i[1] + stepm]);
		ret.push([pos_j[0], pos_j[1] - stepm]);
		if (dx != 0) {
			ret.push([pos_i[0] + stepk, pos_i[1] + stepm]);
			ret.push([pos_j[0] - stepk, pos_j[1] - stepm]);
		}
	} //#end if
	return ret;
} //#end def getOptionsList(pos_i, pos_j, brute=False)

/*
 * getNonConns(conns, i_lt_j=False)
	Inverse of conns. If a is not connected to b, b should be in conns[a].
	If you set i_lt_j == true, it will filter where i >= j.
	Example:
	conns = [[1], [0, 2], [1, 3, 4], [2], [2]]
	i_lt_j = false
	nonConns = [[2, 3, 4], [3, 4], [0], [0, 1, 4], [0, 1, 3]]
	i_lt_j = true
	nonConns = [[2, 3, 4], [3, 4], [], [4], []]
 */
function getNonConns(conns, i_lt_j) {
	if (i_lt_j === undefined) { i_lt_j = false; }
	var nonConns = [];
	var i;
	for (i = 0; i < conns.length; i++) {
		var nc = [];
		for (j = 0; j < conns.length; j++) {
			//# Don't check against ourself.
			if (i == j) continue;
			//# If we are connected, not not connected
			if ((conns[i].indexOf(j) >= 0) || (conns[j].indexOf(i) >= 0)) continue;
			nc.push(j);
		} //#next j
		// Ensure that we only have i < j
		if (i_lt_j) {
			nc = nc.filter(function (j) { return i < j; });
		}
		nonConns.push(nc);
	} //#next i
	return nonConns;
} //#end function getNonConns(conns, i_lt_j=False)

/*
 * This gem from StackOverflow because JavaScript can't do negative modulus correctly.
 */
function Number_mod(v1, n1) {
    return ((v1%n1)+n1)%n1;
};
