#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
SWW Minimize 1
by Javantea
Sept 30, 2015
Based on N Atom molecule using Basin-Hopping Method (aka L-BFGS-B)
by Javantea
July 23, 2014
Based on metropolis_min1.py by Javantea, July 17-22, 2014
Based on s_4atom.py by Javantea, June 24, 2014

# TODO: Improve Lennard-Jones for non-connected nodes.
"""
from __future__ import print_function
try:
	from matplotlib import pyplot
except ImportError as ie:
	print("Matplotlib is an optional dependency if you want graphical validation of LJ potential.")
	pyplot = None
#end try
import numpy as np
from scipy.optimize import minimize, basinhopping
import mapmake1
import math
Vector2 = mapmake1.Vector2
getNonConns = mapmake1.getNonConns
LJ_Potential = sww_optimize1.LJ_Potential

def nameIncremental(i):
	alphabet = 'abcdefghijklmnopqrstuvwxyz'
	output = ''
	while i >= 26:
		i, p = divmod(i, 26)
		output += alphabet[p]
	#loop
	output += alphabet[i]
	return output
#end def nameIncremental(i)

def createMap2(nodes=8, conns=7, seed_id=3243):
	connsAvailable = sum(range(nodes)) - (nodes - 1)
	if conns > connsAvailable: raise ValueError("Too many connections, conns >= sum(range(nodes)) - (nodes - 1)")
	mapmake1.seed(seed_id)
	a = mapmake1.Map()
	b = []
	for i in range(nodes):
		b.append(mapmake1.Point(nameIncremental(i)))
	#next i
	# Connect them all end to end
	for i in range(nodes-1):
		b[i].addConn(b[i+1])
	#next i
	for i in range(conns):
		addConn1 = mapmake1.randint(0,len(b)-1)
		# Fully connected!
		while len(b[addConn1].conns) >= (nodes-1):
			addConn1 = mapmake1.randint(0,len(b)-1)
		#end if
		addConn2 = addConn1
		while addConn2 == addConn1 or b[addConn2] in b[addConn1].conns:
			addConn2 = mapmake1.randint(0,len(b)-1)
		#loop
		#print b, addConn1, addConn2
		b[addConn1].addConn(b[addConn2])
	#next i
	a.addPoints(b)
	a.randomize()
	a.center()
	return a
#end def createMap2([nodes], [conns], [seed_id])

def OptimizeBasinHopping(map1, lj_conf, b0 = 40, steps=60, stepsize=50, disp=True, callback=None):
	"""
	Optimize using SciPy BasinHopping and L-BFGS-B
	One step is fairly fast and usually helps, so why not try a step?
	Optimizes in place.
	"""
	x0, conns = map1.getX0Conns()
	nonConns = getNonConns(conns, True)
	#res = minimize(V_molecular_mult, molecule.x0, args=(molecule.conns,), method='L-BFGS-B',
	#	options={'disp': True}) #, 'T0': 0.1, 'schedule':verbose_fast_sa
	minimizer_kwargs = {'args':(conns, b0, lj_conf, nonConns), 'method':'L-BFGS-B'}
	res = basinhopping(V_SmallWideWorld, x0, minimizer_kwargs=minimizer_kwargs, 
			disp=disp, niter=steps, stepsize=stepsize, callback=callback)

	m = np.matrixlib.matrix(res.x)
	m = m.reshape(len(map1.points), 2)
	for i, xa in enumerate(m):
		x = np.array(xa)[0]
		#print('debug2', x)
		map1.points[i].x = x[0]
		map1.points[i].y = x[1]
	#next i, x
	map1.centerBound()
	return
#end def OptimizeBasinHopping(map1, lj_conf, steps=60, stepsize=50, disp=True, callback=None)

def drawLJ():
	from matplotlib import pyplot
	lj_conf = LJ_Potential()
	data = [lj_conf.V_LJ(x) for x in range(33, 70)]
	xdata = [x for x in range(33, 70)]
	#print('\n'.join(['%3.3f %3.3f' % ((x*0.01) + 1, E) for (x, E) in enumerate(data)]))
	
	if pyplot:
		# TODO: x axis.
		fig = pyplot.figure()
		ax = fig.add_subplot(111)
		ax.set_xlabel('distance (pixels)')
		ax.set_ylabel('Potential (arbitrary magnitude)')
		ax.plot(xdata, data)
		#ax.axis([0.7, 2, 0, None])
		pyplot.show()
	#end if
#end def drawLJ()

def drawHarmonic():
	from matplotlib import pyplot
	b0 = 40
	data = [V_Harmonic(x, b0) for x in range(0, 80)]
	xdata = [x for x in range(0, 80)]
	#print('\n'.join(['%3.3f %3.3f' % ((x*0.01) + 1, E) for (x, E) in enumerate(data)]))
	
	if pyplot:
		# TODO: x axis.
		fig = pyplot.figure()
		ax = fig.add_subplot(111)
		ax.set_xlabel('distance (pixels)')
		ax.set_ylabel('Potential (arbitrary magnitude)')
		ax.plot(xdata, data)
		#ax.axis([0.7, 2, 0, None])
		pyplot.show()
	#end if
#end def drawHarmonic()

class AcceptPotential(object):
	def __init__(self, V):
		self.V = V
	def __call__(self, x=None, f=None, accept=None):
		return (f <= self.V)
#end class AcceptPotential(object)

def main():
	lj_conf = LJ_Potential()
	"""
	#This is how to test the LJ and the Harmonic potentials.
	#aconf = LJConf(A, B)
	#data = [E_Harmonic(x*0.1, a) for x in range(7, 20)]
	data = [V_LJ(x*1.0) for x in range(30, 80)]
	#print('\n'.join(['%3.3f %3.3f' % ((x*0.01) + 1, E) for (x, E) in enumerate(data)]))
	
	if pyplot:
		# TODO: x axis.
		pyplot.plot(data)
		pyplot.show()
	#end if
	data2 = [V_Harmonic(x*1.0, 40) for x in range(30, 80)]
	#print('\n'.join(['%3.3f %3.3f' % ((x*0.01) + 1, E) for (x, E) in enumerate(data)]))
	
	if pyplot:
		# TODO: x axis.
		pyplot.plot(data2)
		pyplot.show()
	#end if
	"""
	import sys
	import os
	
	nodes = 20
	conns = 4
	steps = 40
	x1 = False
	lin = False
	if '-x1' in sys.argv:
		x1 = True
		sys.argv.remove('-x1')
	#end if
	if '-lin' in sys.argv:
		lin = True
		sys.argv.remove('-lin')
	#end if
	if len(sys.argv) > 1:
		nodes = int(sys.argv[1])
		if len(sys.argv) > 2:
			conns = int(sys.argv[2])
			if len(sys.argv) > 3:
				steps = int(sys.argv[3])
			#end if
		#end if
	#end if
	
	map1 = createMap2(nodes, conns, 9999)
	map1.desc = "SWW Optimize map with {} nodes and {} conns [results]".format(nodes, conns)
	dri = map1.points[0] - map1.points[1]
	x0, conns = map1.getX0Conns()
	if x1 or lin:
		if x1:
			#x0 = [510.81305301, 385.65147798, 469.75645727, 392.61197485, 427.16541225, 396.01954412, 382.81828934, 393.89912569, 345.00536839, 411.11936337, 346.75667239, 369.61277072, 328.12245099, 329.2921495,  315.22425258, 288.763412, 306.39832123, 246.89207769, 293.30697898, 206.19322215, 276.1015715,  167.04972025, 254.96226834, 129.8842716,  237.75194095,  90.73870705, 224.65026922,  50.06737728, 207.25813946,  10.80185702, 193.9085541,  -28.59705128]
			x0 = [ 147.56420427,  43.79795302, 189.19732683,  44.68727509, 231.66791435,  49.36023504, 274.82081652,  59.80144291, 315.20266694,  50.02125504, 305.65524673,  90.4528723,  316.34993008, 133.56425343, 321.36519958, 175.79901571, 330.62383549, 217.57657446, 344.10700811, 258.14674393, 361.66757177, 297.13241723, 375.10795718, 337.72261268, 384.29768979, 379.48027738, 389.142904, 421.96628266, 389.6194682,  464.6928152, 385.60133738, 507.44900344, 378.00892276, 548.3489705 ]
		else:
			# Perfectly horizontal spaces 40 apart.
			x0a = []
			j = 0
			curr_x = 10
			for x in x0:
				if j == 0:
					x0a.append(curr_x)
					curr_x += 40
					j = 1
				else:
					x0a.append(10)
					j = 0
				#end if
			#next x
			x0 = x0a
		#end if
		m = np.matrixlib.matrix(x0)
		m = m.reshape(len(x0)//2, 2)
		for i, xa in enumerate(m):
			x = np.array(xa)[0]
			#print('debug2', x)
			map1.points[i].x = x[0]
			map1.points[i].y = x[1]
		#next i, x
	#end if
	nonConns = getNonConns(conns, True)
	#print('Conns:', conns)

	# FIXME: Race condition is very likely due to multiple programs running at once.
	file_num = 0
	while os.path.exists('sww_optimize1out2%ia.svg' % file_num):
		file_num += 1
	#loop
	open('sww_optimize1out2%ia.svg' % file_num, 'w').write(str(map1))
	callback = AcceptPotential((81.956*nodes) -178.76)
	OptimizeBasinHopping(map1, lj_conf, 40, steps, 50, True, callback)
	
	print('sww_optimize1out2%i.svg' % file_num)
	print('sww_optimize1out2%i.json' % file_num)
	print("Configuration:")
	x_end = map1.getX0Conns()[0]
	print(x_end)
	Vmin = V_SmallWideWorld(x_end, conns, 40, lj_conf, nonConns)
	print("Minimum potential:", Vmin)
	dr = Vector2(*(x_end[:2])) - Vector2(*(x_end[2:4]))
	print("Distance:", dr.length())
	#x0 = []
	#for v in map1.points:
	#	x0.extend(v)
	#next v
	#x0 = np.array(x0)
	V0 = V_SmallWideWorld(x0, conns, 40, lj_conf, nonConns)
	print("Initial Potential:", V0)
	print("Initial Distance:", dri.length())
	#print('debug', np.array(m[0])[0], np.array(m[1])[0])
	s = str(map1)
	results = '\nConfiguration:\n{}\nMinimum potential:{}\nDistance: {}\nInitial potential:{}\nInitial Distance: {}\n'.format(x_end, Vmin, dr.length(), V0, dri.length())
	s = s.replace('[results]', results)
	open('sww_optimize1out2%i.svg' % file_num, 'w').write(s)
	open('sww_optimize1out2%i.json' % file_num, 'w').write(map1.saveJson())
	#print("Minimum potential breakdown:")
	#V_SmallWideWorld_verbose(res.x, conns, 40, lj_conf)
#end def main()

if __name__ == '__main__':
    main()
#end if
