#!/usr/bin/env python3
"""
A2 Intelligent Minimization
by Javantea
Feb 12, 2016

Original idea years ago, since we know the potential between nodes, it's easy for us to decide which node to move. Therefore, we can use the x-y coordinates of that node only to minimize a graph. The first proof of concept is as below (or boxleaf1a.json):

d--a--------e
|  |
b--c

As you can plainly see, there is a 14k potential caused by the a--e bond. By examining a and e's contribution  to the system, we can choose to move e and not move a (since a has low potential already). Initial potential: 14817. boxleaf1.json shows the end goal generated by SWW lattice sort, potential 25.092.

Without any help, L-BFGS-B is able to get 25.9 potential which is strange, but apparently what it thinks is right. sww_energy1.py calculates the boxleaf1.json solution to be 26.439247496203777 which is higher, which explains the L-BFGS-B's choice. Apparently it's correct.

The second graph to verify the validity of this claim should be (line3a.json):

a--b-c

This has a potential of 1051.0 compared to the local minima solution found by SWW linear sort: 19.736. Our end goal without BFGS should be the same as linear sort (line3.json), but if we can get lower by becoming an equilateral triangle, I would prefer that. With L-BFGS-B (with only two variables, it should be very quickly solved), it should be fast, but not instantaneous which is what we would like.

time python3 a2.py -f line3a.json 
Initial Potential: 1050.9707264944057
Max Bond: 515.6966860376589 1 1
Max Unconn: -0.012004726892973841 0 0
Minimizing bond b 430.000, 283.000 c 447.290, 283.190
Switched, so we will use c
      fun: 1.6959721217835599
 hess_inv: <2x2 LbfgsInvHessProduct with dtype=float64>
      jac: array([ -1.77635684e-07,  -2.22044605e-08])
  message: b'CONVERGENCE: NORM_OF_PROJECTED_GRADIENT_<=_PGTOL'
     nfev: 171
      nit: 40
   status: 0
  success: True
        x: array([ 405.34100339,  314.50901449])
Local Minimized: 1.69597212178356
line3amin.json

real    0m0.171s
user    0m0.158s
sys     0m0.013s

This is fast and good which is what we want. But it could be faster without L-BFGS-B which is what we want. Surprisingly, L-BFGS-B is almost instantaneous when used on two variables which means we're done with this algorithm.
"""
from __future__ import print_function
import sys
import os
#sys.path.append('..')
import sww_optimize1
import numpy as np
from scipy.optimize import minimize
mapmake1 = sww_optimize1.mapmake1
V_Harmonic = sww_optimize1.V_Harmonic
V_Harmonic_Angle = sww_optimize1.V_Harmonic_Angle
Vector2 = mapmake1.Vector2
ConnList = sww_optimize1.ConnList

def usage():
	print(main.__doc__)
	return 1
#end def usage()

def V_SmallWideWorld_detail(x, conns, b0, lj_conf, nonConns):
	"""
	Based on V_SmallWideWorld, records values for connections and non-conns.
	"""
	Potential = 0.0
	Bond_potentials = []
	Bond_angle_potentials = []
	# Harmonic potential with length b0.
	for i in range(len(conns)):
		pt_i = Vector2(*x[2*i:2*i+2])
		verts = [pt_i]
		dps = []
		dp_lens = []
		Bond_potentials_i = []
		for j in conns[i]:
			pt_j = Vector2(*x[2*j:2*j+2])
			dp = pt_i - pt_j
			dp_len = dp.length()
			v = V_Harmonic(dp_len, b0)
			Bond_potentials_i.append(v)
			Potential += v
			verts.append(pt_j)
			dps.append(dp)
			dp_lens.append(dp_len)
		#next j
		Bond_potentials.append(Bond_potentials_i)
		if len(verts) > 2:
			# Angular Harmonic calculation
			theta0 = 2 * np.pi / len(conns[i])
			# theta is computed for axb, bxc, cxa
			# FIXME: There was a bug here where I was using i. Now I'm confused.
			Bond_angle_potentials_i = []
			for k in range(len(conns[i])):
				b = (k+1) % len(dps)
				# Theta between axb is arccos((ax·xb)/(|ax|*|xb|))
				ax = dps[k]
				xb = dps[b]
				#print('xb', '(%3.3f,%3.3f)' % (xb.x, xb.y), '(%3.3f,%3.3f)' % (dps[b].x, dps[b].y), xb.length(), dp_lens[b], dps[b].length())
				top = ax.dot(xb)
				bot = dp_lens[k]*dp_lens[b]
				if bot == 0:
					theta = 0
				else:
					intermediate = top / bot
					# A very stupid floating point bug.
					if intermediate < -1.0:
						theta = np.pi
					elif intermediate > 1.0:
						theta = 0.0
					else:
						theta = np.arccos(intermediate)
					#end if
				#end if
				v = V_Harmonic_Angle(theta, theta0)
				Bond_angle_potentials_i.append(v)
				Potential += v
			#next i
			Bond_angle_potentials.append(Bond_angle_potentials_i)
		#end if
	#next i
	Unconn_potentials = []
	# Lennard-Jones for non-connected nodes.
	for i in range(len(nonConns)):
		pt_i = Vector2(*x[2*i:2*i+2])
		Unconn_potentials_i = []
		for j in nonConns[i]:
			pt_j = Vector2(*x[2*j:2*j+2])
			dp = pt_i - pt_j
			dp_len = dp.length()
			v = lj_conf.V_LJ(dp_len)
			Unconn_potentials_i.append(v)
			Potential += v
		#next j
		Unconn_potentials.append(Unconn_potentials_i)
	#next i
	return (Potential, Bond_potentials, Bond_angle_potentials, Unconn_potentials)
#end def V_SmallWideWorld_detail(x, conns, b0, lj_conf)

def find_max_V_SmallWideWorld(x, conns, b0, lj_conf, nonConns):
	"""
	Take a list of potentials from V_SmallWideWorld_detail and find the worst of bonded and non-bonded. 
	Return index into self.points/conns and a second index indicating which point it is conflicting with.
	TODO: decide which of the two is the problem child.
	"""
	noBond = False
	MinBond = V_Harmonic(b0, b0)
	MinBondAngle = V_Harmonic_Angle(180, 180)
	# r0 is the value for with the LJ potential is optimized.
	MinUnconn = lj_conf.V_LJ(lj_conf.r0)
	(Potential, Bond_potentials, Bond_angle_potentials, Unconn_potentials) = V_SmallWideWorld_detail(x, conns, b0, lj_conf, nonConns)
	maxBondFound = MinBond
	maxBondFound_i = 0
	maxBondFound_index = 0
	for i, BondPlist in enumerate(Bond_potentials):
		if len(BondPlist) == 0: continue
		# One liner to find index and max value
		maxBondP_index, maxBondP = max(enumerate(BondPlist), key=lambda x:x[1])
		if maxBondP > maxBondFound:
			maxBondFound = maxBondP
			maxBondFound_index = maxBondP_index
			maxBondFound_i = i
		#end if
	#next BondPlist
	if maxBondFound_index >= len(conns[maxBondFound_i]):
		print("Note: No bonds found > Min Bond Potential")
		noBond = True
	#end if
	print("Max Bond:", maxBondFound, maxBondFound_i, maxBondFound_index)
	maxUnconnFound = MinUnconn
	maxUnconnFound_i = 0
	maxUnconnFound_index = 0
	for i, UnconnPlist in enumerate(Unconn_potentials):
		if len(UnconnPlist) == 0: continue
		# One liner to find index and max value
		maxUnconnP_index, maxUnconnP = max(enumerate(UnconnPlist), key=lambda x:x[1])
		if maxUnconnP > maxUnconnFound:
			maxUnconnFound = maxUnconnP
			maxUnconnFound_index = maxUnconnP_index
			maxUnconnFound_i = i
		#end if
	#next UnconnPlist
	print("Max Unconn:", maxUnconnFound, maxUnconnFound_i, maxUnconnFound_index)
	if maxUnconnFound_index >= len(nonConns[maxUnconnFound_i]):
		print("Note: fully connected graph?")
		maxUnconn = None
	else:
		a, b = maxUnconnFound_i, nonConns[maxUnconnFound_i][maxUnconnFound_index]
		maxUnconn = (maxUnconnFound, a, b)
		#  TODO: Which is worse?
		#if sum(UnconnPlist[a]) < sum(UnconnPlist[b]):
		#	maxUnconn = (maxUnconnFound, b, a)
		#else:
		#	if len(
	#end if
	if noBond:
		maxBond = None
	else:
		maxBond = (maxBondFound_i, maxBondFound_i, conns[maxBondFound_i][maxBondFound_index])
	#end if
	return maxBond, maxUnconn
#end def find_max_V_SmallWideWorld(x, conns, b0, lj_conf, nonConns)


def V_SmallWideWorld_single(x, x_other, index, conns, b0, lj_conf, nonConns):
	"""
	Based on V_SmallWideWorld
	x is only one vertex.
	x_other is x0 as in V_SmallWideWorld.
	The index is the point which we are affecting in x_other with x.
	Hopefully that makes sense.
	TODO: Cache, a major speed improvement but a lot of computation required!
	NOTE: V_SmallWideWorld_Cache may be close to what we want but V_SmallWideWorld is faster.
	"""
	x_other[2*index] = x[0]
	x_other[2*index+1] = x[1]
	
	# The rest is the same.
	Potential = 0.0
	# Harmonic potential with length b0.
	for i in range(len(conns)):
		pt_i = Vector2(*x_other[2*i:2*i+2])
		verts = [pt_i]
		dps = []
		dp_lens = []
		for j in conns[i]:
			pt_j = Vector2(*x_other[2*j:2*j+2])
			dp = pt_i - pt_j
			dp_len = dp.length()
			Potential += V_Harmonic(dp_len, b0)
			verts.append(pt_j)
			dps.append(dp)
			dp_lens.append(dp_len)
		#next j
		if len(verts) > 2:
			# Angular Harmonic calculation
			theta0 = 2 * np.pi / len(conns[i])
			# theta is computed for axb, bxc, cxa
			# FIXME: There was a bug here where I was using i. Now I'm confused.
			for k in range(len(conns[i])):
				b = (k+1) % len(dps)
				# Theta between axb is arccos((ax·xb)/(|ax|*|xb|))
				ax = dps[k]
				xb = dps[b]
				#print('xb', '(%3.3f,%3.3f)' % (xb.x, xb.y), '(%3.3f,%3.3f)' % (dps[b].x, dps[b].y), xb.length(), dp_lens[b], dps[b].length())
				top = ax.dot(xb)
				bot = dp_lens[k]*dp_lens[b]
				if bot == 0:
					theta = 0
				else:
					intermediate = top / bot
					# A very stupid floating point bug.
					if intermediate < -1.0:
						theta = np.pi
					elif intermediate > 1.0:
						theta = 0.0
					else:
						theta = np.arccos(intermediate)
					#end if
				#end if
				Potential += V_Harmonic_Angle(theta, theta0)
			#next i
		#end if
	#next i
	# Lennard-Jones for non-connected nodes.
	for i in range(len(nonConns)):
		pt_i = Vector2(*x_other[2*i:2*i+2])
		for j in nonConns[i]:
			pt_j = Vector2(*x_other[2*j:2*j+2])
			dp = pt_i - pt_j
			dp_len = dp.length()
			Potential += lj_conf.V_LJ(dp_len)
		#next j
	#next i
	return Potential
#end def V_SmallWideWorld_single(x, conns, b0, lj_conf)

def main():
	"""Small Wide World Algorithm A2
This program creates or optimizes an already created graph using intelligent control of L-BFGS-B local optimization.
Usage:
a2.py [-f filename] [nodes] [conns] [branches]
-f filename | Load a JSON file.
nodes       | If you aren't loading from file, how many nodes on the 
              graph. Default: 20
conns       | How many random connections to make between nodes on the 
              backbone. Default: 0
branches    | How many random branches to add to the backbone. Default: 1
"""
	nodes = 20
	conns = 0
	branches = 1
	seed_id = 3243
	gtol = 1e-2
	# FIXME: use getopt or something
	# Skip minimization, try out our algorithm.
	mini1 = False
	if '-f' in sys.argv:
		if len(sys.argv) <= sys.argv.index('-f'):
			return usage()
		#end if
		base_filename = sys.argv[sys.argv.index('-f')+1]
		map1 = mapmake1.Map()
		map1.loadJson(open(base_filename, 'r').read())
		nodes = len(map1.points)
	else:
		# FIXME: Race condition is very likely due to multiple programs running at once.
		file_num = 0
		while os.path.exists('a2_{0}.json'.format(file_num)):
			file_num += 1
		#loop
		base_filename = 'a2_{0}.json'.format(file_num)
		if len(sys.argv) > 1:
			nodes = int(sys.argv[1])
		#end if
		map1 = sww_optimize1.createMap4(nodes, conns, branches, seed_id, runSorts=False)
		#map1.sortLinear()
		if conns <= 0:
			map1.sortBranch()
		else:
			map1.sortLoop()
		#end if
		map1.centerBound()
	#end if
	
	open(base_filename,'w').write(map1.saveJson())
	# It's important for it not to be linear if it's going to be minimized
	# because linear is a sort of ridge.
	#if mini1:
	#	map1.randomize(8)
	#end if

	LJ_Potential = sww_optimize1.LJ_Potential
	V_SmallWideWorld = sww_optimize1.V_SmallWideWorld
	getNonConns = sww_optimize1.getNonConns
	x0, conns = map1.getX0Conns(True)
	conns_cpp = ConnList(conns)
	nonConns = getNonConns(conns, True)
	nonConns_cpp = ConnList(nonConns)
	b0 = 40
	lj_conf = LJ_Potential()
	V_0 = V_SmallWideWorld(x0, conns_cpp, b0, lj_conf, nonConns_cpp)
	print("Initial Potential:", V_0)
	
	maxBond, maxUnconn = find_max_V_SmallWideWorld(x0, conns, b0, lj_conf, nonConns)
	# If bond is worse, minimize bond.
	# If Unconn is worse, minimize unconn.
	# TODO: We could do both if we wanted to.
	minimizeBond = False
	minimizeUnconn = False
	if maxBond:
		if maxUnconn:
			if maxBond[0] > maxUnconn[0]:
				minimizeBond = True
			else:
				minimizeUnconn = True
			#end if
		else:
			minimizeBond = True
		#end if
	elif maxUnconn:
		minimizeUnconn = True
	#end if
	if minimizeBond:
		print("Minimizing bond", map1.points[maxBond[1]], map1.points[maxBond[2]])
		# Bond is worse than Unconn, need to shorten a bond.
		pt_a, pt_b = maxBond[1:3]
		# If one is connected to two and the other is connected to one... choose the one with less bonds.
		#print("Choice:", pt_a, pt_b, len(map1.points[pt_a].conns), len(map1.points[pt_b].conns), [d.name for d in map1.points], [[c.name for c in p.conns] for p in map1.points])
		if len(map1.points[pt_a].conns) > len(map1.points[pt_b].conns):
			pt_a = pt_b
			pt_b = maxUnconn[1]
			print("Switched, so we will use", map1.points[pt_a].name)
		#end if
		minimizer_opts = {"method": "L-BFGS-B", 'args':(x0, pt_a, conns, b0, lj_conf, nonConns), 'options':{'gtol':gtol, 'eps':1e-4}}
		x0_piece = x0[2 * pt_a:2 * pt_a + 2]
		res = minimize(V_SmallWideWorld_single, x0_piece, **minimizer_opts)
		print(res)
		x1 = x0[:]
		x1[2*pt_a] = res.x[0]
		x1[2*pt_a+1] = res.x[1]
		Vmin = V_SmallWideWorld(x1, conns_cpp, b0, lj_conf, nonConns_cpp)
		print('Local Minimized:', Vmin)
		m = np.matrixlib.matrix(x1)
		m = m.reshape(len(map1.points), 2)
		map1.setPos([np.array(xa)[0] for xa in m])
		map1.centerBound()
		open(base_filename.replace('.json', 'min.json'), 'w').write(map1.saveJson())
		open(base_filename.replace('.json', 'min.svg'), 'w').write(map1.drawSvg())
		print(base_filename.replace('.json', 'min.json'))
	#end if
	if minimizeUnconn:
		# Unconn is worse than bond, Move two overlapping nodes away or two far away nodes closer.
		print("Minimizing Unconn", map1.points[maxUnconn[1]].name, map1.points[maxUnconn[2]].name)
		pt_a, pt_b = maxUnconn[1:3]
		# If one is connected to two and the other is connected to one... choose the one with less bonds.
		if len(map1.points[pt_a].conns) > len(map1.points[pt_b].conns):
			pt_a = pt_b
			pt_b = maxUnconn[1]
			print("Switched, so we will use", map1.points[pt_a].name)
		#end if
		minimizer_opts = {"method": "L-BFGS-B", 'args':(x0, pt_a, conns, b0, lj_conf, nonConns)}
		x0_piece = x0[2 * pt_a:2 * pt_a + 2]
		# TODO: We don't want overlap but this could do worse.
		x0_piece[0] += 0.1
		x0_piece[1] += 0.1
		print('x0_piece:', x0_piece)
		res = minimize(V_SmallWideWorld_single, x0_piece, **minimizer_opts)
		print(res)
		x1 = x0[:]
		x1[2*pt_a] = res.x[0]
		x1[2*pt_a+1] = res.x[1]
		Vmin = V_SmallWideWorld(x1, conns_cpp, b0, lj_conf, nonConns_cpp)
		print('Local Minimized:', Vmin)
		m = np.matrixlib.matrix(x1)
		m = m.reshape(len(map1.points), 2)
		map1.setPos([np.array(xa)[0] for xa in m])
		map1.centerBound()
		open(base_filename.replace('.json', 'min.json'), 'w').write(map1.saveJson())
		open(base_filename.replace('.json', 'min.svg'), 'w').write(map1.drawSvg())
		print(base_filename.replace('.json', 'min.json'))
	#end if
	
#end def main()

if __name__ == '__main__':
	main()
#end if
